<?php

class KB_color extends KB_class {

	private $core;

	function __construct($core) {
		$this->core = $core;
	}

	// PUBLIC

	public function create ($h = "", $s = "", $l = "") {

		return $this->adjust("hsl(0,0%,0%)", $h, $s, $l);

	}


	public function adjust($input = "", $h = "", $s = "", $l = "") {

		$color = $this->color($input);

		if (!is_array($color))
			return $input;

		$h = $this->value($h);
		$s = $this->value($s);
		$l = $this->value($l);

		if (is_int($h))
			$color[0] = $h;
		else
			$color[0] += either(intval($h),0);

		$color[0] %= 360;

		if (!are_equal($color[1], 0)) {

			if (is_int($s))
				$color[1] =  max($s,0);

			else
				$color[1] += either(max(intval($s),0),0) % 360;

		}

		$color[1] =  min(100, $color[1]);

		if (is_int($l))
			$color[2] =  $l;

		else
			$color[2] += either(intval($l),0) % 360;

		$color[2] =  min(100, $color[2]);

		return $this->render($input, $color);

	}


	public function hue($input = "", $h = "") {
		return $this->adjust($input, $h, "+0", "+0");
	}


	public function saturation($input = "", $s = "") {
		return $this->adjust($input, "+0", $s, "+0");
	}


	public function luminance($input = "", $l = "") {
		return $this->adjust($input, "+0", "+0", $l);
	}


	public function convert($input = "", $target = "") {

		if (any_empty($input, $target))
			return $input;

		$color = $this->color($input);

		if (!is_array($color))
			return $input;

		return $this->render($target, $color);

	}


	// ALIAS

	public function hsl($input = "", $h = "", $s = "", $l = "") {
		return $this->adjust($input, $h, $s, $l);
	}

	public function h($input = "", $h = "") {
		return $this->hue($input, $h);
	}


	public function s($input = "", $s = "") {
		return $this->saturation($input, $s);
	}

	public function l($input = "", $l = "") {
		return $this->luminance($input, $l);
	}

	// PRIVATE

	private function color($input = "") {

		if (is_empty($input))
			return array(0,0,0);

		$input = $this->clean_input($input);

		switch ($this->type($input)) {

			case 'hex':
				return $this->color("rgb(".hexdec(substr($input,1,2)).",".hexdec(substr($input,3,2)).",".hexdec(substr($input,5,2)).")");
			case 'rgb':
			case 'rgba':
				return $this->parse_rgb(array(
					preg_replace("/^[\w]+\(([\d]+),([\d]+),([\d]+)(.+)$/","$1",$input),
					preg_replace("/^[\w]+\(([\d]+),([\d]+),([\d]+)(.+)$/","$2",$input),
					preg_replace("/^[\w]+\(([\d]+),([\d]+),([\d]+)(.+)$/","$3",$input),
				));
			case 'hsl':
			case 'hsla':
				return array(
					preg_replace("/^[\w]+\(([\d]+),([\d]+)%,([\d]+)%(.+)$/","$1",$input),
					preg_replace("/^[\w]+\(([\d]+),([\d]+)%,([\d]+)%(.+)$/","$2",$input),
					preg_replace("/^[\w]+\(([\d]+),([\d]+)%,([\d]+)%(.+)$/","$3",$input),
				);

			default:
				return false;
		}

	}

	private function render($input = "", $target = "") {

		$target_rgb = $this->render_rgb($target);

		$input = $this->clean_input($input);

		switch ($this->type($input)) {

			case 'hex':
				return "#".str_pad(dechex($target_rgb[0]),2,"0",STR_PAD_LEFT).str_pad(dechex($target_rgb[1]),2,"0",STR_PAD_LEFT).str_pad(dechex($target_rgb[2]),2,"0",STR_PAD_LEFT);

			case 'rgb':
				return "rgb(".$target_rgb[0].",".$target_rgb[1].",".$target_rgb[2].")";

			case 'rgba':
				return "rgba(".$target_rgb[0].",".$target_rgb[1].",".$target_rgb[2].",".preg_replace("/^.+,([.]?[\d]+)\)$/","$1",$input).")";

			case 'hsl':
				return "hsl(".$target[0].",".$target[1]."%,".$target[2]."%)";

			case 'hsla':
				return "hsla(".$target[0].",".$target[1]."%,".$target[2]."%,".preg_replace("/^.+,([.]?[\d]+)\)$/","$1",$input).")";

			default:
				return false;
		}
	}

	private function type($input = "") {

		$input = $this->clean_input($input);

		if (any_empty($input, is_string($input)))
			return false;

		if ($this->core->array->match($input, array("hex","rgb","rgba","hsl","hsla")))
			return $input;

		if (preg_match("/^#[\w\d]{6}$/",$input))
			return "hex";

		if (preg_match("/^rgb\([\d]{1,3},[\d]{1,3},[\d]{1,3}\)$/",$input))
			return "rgb";

		if (preg_match("/^rgba\([\d]{1,3},[\d]{1,3},[\d]{1,3},[.]?[\d]+\)$/",$input))
			return "rgba";

		if (preg_match("/^hsl\([\d]{1,3},[\d]{1,3}%,[\d]{1,3}%\)$/",$input))
			return "hsl";

		if (preg_match("/^hsla\([\d]{1,3},[\d]{1,3}%,[\d]{1,3}%,[.]?[\d]+\)$/",$input))
			return "hsla";

		return false;

	}

	private function clean_input($input) {

		if (any_empty($input, is_string($input)))
			return false;

		return preg_replace("/[^\d\w#,()%]/","",$input);

	}

	private function value($input = "") {

		if (is_empty($input))
			return 0;

		if (are_empty(is_string($input), is_numeric($input)))
			return 0;

		if (is_numeric($input))
			return $input;

		$output = preg_replace("/[^\d+-]/","",$input);

		if (is_numeric($output))
			return intval($output);

		return $output;

	}

	private function parse_rgb($color) {

		$r = $color[0] / 255;
		$g = $color[1] / 255;
		$b = $color[2] / 255;

		$max = max( $r, $g, $b );
		$min = min( $r, $g, $b );

		$h = 0;
		$s = 0;

		$l = ( $max + $min ) / 2;
		$d = $max - $min;

  	if ($d == 0 ) {
      $h = $s = 0; // achromatic
  	}

		else {
    	$s = $d / ( 1 - abs( 2 * $l - 1 ) );
			switch( $max ){
        case $r:
        	$h = 60 * fmod( ( ( $g - $b ) / $d ), 6 );
          if ($b > $g)
            $h += 360;
          break;

				case $g:
        	$h = 60 * ( ( $b - $r ) / $d + 2 );
        	break;

				case $b:
        	$h = 60 * ( ( $r - $g ) / $d + 4 );
        	break;
      }
		}

		return array(round($h), round($s*100), round($l*100));
	}

	private function render_rgb($color) {

		$h = $color[0];
		$s = $color[1]/100;
		$l = $color[2]/100;

		$r;
    $g;
    $b;
		$c = ( 1 - abs( 2 * $l - 1 ) ) * $s;
		$x = $c * ( 1 - abs( fmod( ( $h / 60 ), 2 ) - 1 ) );
		$m = $l - ( $c / 2 );

		if ( $h < 60 ) {
			$r = $c;
			$g = $x;
			$b = 0;
		} else if ( $h < 120 ) {
			$r = $x;
			$g = $c;
			$b = 0;
		} else if ( $h < 180 ) {
			$r = 0;
			$g = $c;
			$b = $x;
		} else if ( $h < 240 ) {
			$r = 0;
			$g = $x;
			$b = $c;
		} else if ( $h < 300 ) {
			$r = $x;
			$g = 0;
			$b = $c;
		} else {
			$r = $c;
			$g = 0;
			$b = $x;
		}

		$r = ( $r + $m ) * 255;
		$g = ( $g + $m ) * 255;
		$b = ( $b + $m  ) * 255;

    return array( floor( $r ), floor( $g ), floor( $b ) );
	}

}