<?php


class KB_data extends KB_class_extended {

	// Public

	// public $order = array();
	public $group = false;
	public $position = false;
	public $statistic = false;
	public $properties = false;
	public $auto_cache = false;
	public $auto_create = false;
	public $create_sql = false;
	// public $correlation = NULL;
	public $foreign = false;
	public $cache = false;

	// Deprecated
	public $db_name = "";

	// Private
	private $base_dir = "";
	private $childs = array();
	private $elements;
	private $root = 'NULL';
	private $read_all = false;
	private $read_cache = false;
	private $group_by_parent_ID = false;
	private $filters = array();

	private $core, $expandable, $indexed, $correlation, $key;


	function __construct($core) {

		$this->core = $core;

		$this->config("db",array(
			"table" => "kb_data",
			"name" => ""
		));
		$this->expandable(false);
		$this->indexed(true);
		$this->correlation(false);
		$this->key("ID");
		$this->private_config("type", "data");

	}


	public function cache ($set = NULL) {

		if ($set !== NULL)
			$this->cache = (bool) $set;

		if (!$this->cache)
			return false;

		if (is_object($this->cache))
			return $this->cache;

		if (!$this->properties())
			return false;

		if (!$this->properties->count())
			return false;

		$this->cache = $this->core->data($this->db_table().'_cache');
		$this->cache->expandable(true);
		$this->cache->table_ID($this->table_ID());

		$this->fill_cache_table();

		if (isset($this->order))
			$this->cache->order = $this->order;

		if ($this->limit())
			$this->cache->limit($this->limit());

		$this->cache->indexed($this->indexed());
		$this->cache->position($this->position());
		$this->cache->foreign($this->foreign());
		$this->cache->correlation($this->correlation());
		$this->cache->filter($this->filter());
		$this->cache->group($this->group());

		return $this->cache;
	}


	public function column ($column = "") {

		if (any_empty(
			not_empty($column),
			is_string($column),
			not_empty(key_of($column, $this->columns()))
			))
			return [];

		if (!is_null($cache = $this->private_config("cache_function_column_".$column)))
			return $cache;

		$data = $this->core->db->select([$this->db_table()=>$column],$this->filter());
		// $data = $this->core->db->select([$this->db_table()=>$column]);
		$data = key_fetch($column, $data);

		$this->private_config("cache_function_column_".$column, $data);
		return $data;

	}


	public function column_values ($column = "") {

		if (!is_null($cache = $this->private_config("cache_function_column_values_".$column)))
			return $cache;

		$data = $this->column($column);
		$data = $this->core->array->unique($data);

		$this->private_config("cache_function_column_values_".$column, $data);
		return $data;

	}


	public function columns () {

		return $this->core->db->table->columns($this->db_table(), $this->db_name());

	}


	public function count ($select = []) {

		if (!$this->core->db->table->exists($this->db_table()))
			return false;

		return $this->core->db->rows(
			$this->db_table(),
			$this->core->array->merge(
				$this->filter(),
				$select
			)
		);
	}


	public function count_per ($key = false, $value = "") {

		if ($key === false)
			return 0;

		$data = $this->private_config("group_by_$key");

		if (is_empty($data)) {

			if ($this->core->db->table->exists($this->db_table())) {
				$data = $this->core->db->select($this->db_table());
				$data = $this->core->array->group_by($data,$key);
			}

			else {
				$data[$value] = array();
			}

			$this->private_config("group_by_$key", $data);
		}

		if (isset($data[$value]))
			return count($data[$value]);

		return 0;

	}


	public function correlates ($ID = false, $table = false, $entries = NULL) {

		if (any_empty($ID, $table))
			return array();

		if ($entries === NULL)
			return $this->core->correlation->get(
				$this,
				$ID,
				$table
			);

		return $this->core->correlation->set(
			$this,
			$ID,
			$table,
			$entries
		);

	}


	public function correlation($set = NULL) {

		if (is_null($set))
		 	return ifttt(is_null($this->correlation), $this->config("correlation"), $this->correlation);

		return $this->config("correlation", $set);
	}


	public function create ($array = []) {

		// Multiple new entries

		if (is_indexed($array)) {

			// Fast bulk insert

			if (
				not_empty($this->columns()) &&
				!$this->expandable() &&
				!$this->statistic() &&
				!$this->properties() &&
				!$this->cache()
			)
				return $this->sql_insert($array);

			// Slow one entry at a time

			$ID = [];

			foreach($array as $entry)
				$ID[] = $this->create($entry);

			return $ID;
		}

		// Single new entry

		// Read structure of data table

		$columns = $this->columns();

		// Autocreate Table

		if (is_empty($columns)) {

			if ($this->auto_create) {

				if ($this->create_table()) {
					$columns = $this->columns();
				}
				else {
					return false;
				}
			}
			else {
				return false;
			}
		}

		// Expand columns if allowed

		if (
				$this->expandable() &&
				count(array_keys($array)) > count($this->core->array->intersect(array_keys($array),array_keys($columns)))
			) {

			$additional_columns = $this->core->array->diff(array_keys($array),array_keys($columns));

			if ($this->statistic())
				$additional_columns = $this->core->array->diff($additional_columns, array("created","updated","created_by","updated_by"));

			foreach ($additional_columns as $additional_columns_key => $additional_column) {

				if (not_empty($additional_column_config = key_fetch($additional_column, $this->auto_create))) {
					$additional_columns[$additional_column] = $additional_column_config;
				 	unset($additional_columns[$additional_columns_key]);
				}
			}

			$this->core->db->table->expand($this->db_table(),$additional_columns);
			$columns = $this->columns();
		}


		// Limit entry to available columns

		$entry = keys_of(array_keys($columns),$array);

		if (is_empty($entry) && $this->key())
			$entry = array($this->key()=>NULL);

		// Create entry in data table

		$ID = $this->core->db->insert($this->db_table(),$entry, false, $this->db_name());

		if (is_empty($ID))
			return false;

		// Update position

		if ($this->position())
			$this->reposition(
				ifttt(
					$this->group(),
					array($this->group() => either(key_of($this->group(),$array), "NULL")),
					false
				)
			);


		// Set statistics

		if ($this->statistic()) {

			$array["created"] = either(key_of("created", $array), "NOW()");
			$array["updated"] = either(key_of("updated", $array), "NOW()");

			$array["created_by"] = either(key_of("created_by", $array), $this->core->user->ID());
			$array["updated_by"] = either(key_of("updated_by", $array), $this->core->user->ID());

			$this->core->statistic->create(array(
				"table_ID"	=> $this->table_ID(true),
				"entry_ID"	=> $ID,
				"user_ID"		=> $array["created_by"],
				"date"			=> $array["created"],
				"description"	=> "created"
			));

			$this->core->statistic->create(array(
				"table_ID"	=> $this->table_ID(true),
				"entry_ID"	=> $ID,
				"user_ID"		=> $array["updated_by"],
				"date"			=> $array["updated"],
				"description"	=> "updated"
			));

		}

		// Set properties

		if ($this->properties() && is_object($this->properties)) {

			$properties = $array;

			if ($this->statistic)
				unset(
					$properties["created"],
					$properties["updated"],
					$properties["created_by"],
					$properties["updated_by"]
				);

			// Remove columns already set
			$properties = $this->core->array->diff($properties, $entry);

			if (not_empty($properties)) {

				// Autocreate Table

				if (
					not_empty($this->properties->columns()) ||
					($this->auto_create && $this->create_table())
				) {

					// Add ID and insert properties
					$this->properties->create(
						$this->core->array->merge(array("ID"=>$ID),$properties)
					);
				}
			}
		}

		// Update cache table

		if ($this->cache())
			$this->cache->create($array+array("ID"=>$ID));


		$this->reset("cache");
		return $ID;

	}


	public function db_name ($name = "") {

		if (is_string($name) && not_empty($name)) {
			$this->config("db","name",$name);
		}

		return key_of("name",$this->config("db"));

	}


	public function db_table ($name = "", $auto_create = false) {

		if (is_empty($name))
			return either(
				$this->db_name, // <- Deprecated, but supported
				key_of("table",$this->config("db"))
			);

		if (is_string($name)) {

			$this->config("db","table",$name);

			if (is_bool($auto_create)) {
				$this->auto_create = $auto_create;
			}
			else if (is_array($auto_create)) {
				$this->auto_create = true;
				$this->db_column_options($auto_create);
			}
		}

		return either(
			$this->db_name, // <- Deprecated, but supported
			key_of("table",$this->config("db"))
		);
	}


	public function db_column_options ($set = NULL) {

		$options = either($this->private_config("db_column_options"), []);

		if (is_empty($set))
			return $options;

		if (is_string($set))
			$set = [$set];

		$presets = [
			"boolean" => "TINYINT(1) UNSIGNED NULL DEFAULT NULL",
			"currency" => "DECIMAL(18,2) NULL DEFAULT '0'",
			"date" => "DATE NULL DEFAULT NULL",
			"datetime" => "DATETIME NULL DEFAULT NULL",
			"false" => "TINYINT(1) UNSIGNED NOT NULL DEFAULT '0'",
			"ID" => "SMALLINT UNSIGNED NULL DEFAULT NULL",
			"now" => "DATETIME NULL DEFAULT NOW()",
			"text" => "TEXT NOT NULL DEFAULT ''",
			"today" => "DATE NULL DEFAULT CURRENT_DATE()",
			"true" => "TINYINT(1) UNSIGNED NOT NULL DEFAULT '1'",
			"varchar" => "VARCHAR(255) NOT NULL DEFAULT ''",
		];

		foreach($set as $key => $value) {

			if (is_numeric($key)) {
				unset($set[$key]);
				$key = trim($value);
				$value = "varchar";

			}

			$value = either(key_of($value, $presets), $value);

			$set[$key] = trim($value);
		}

		if (is_array($set)) {
			$options = $this->core->array->union(
				$options,
				$set
			);
			$this->private_config("db_column_options", $options);
		}

		return $options;

	}


	public function db_table_options ($set = NULL) {

		$options = either($this->private_config("db_table_options"), array());

		if (is_empty($set))
			return $options;

		if (is_string($set))
			$set = array($set);

		if (is_array($set)) {
			$options = $this->core->array->union(
				$options,
				$set
			);
			$this->private_config("db_table_options", $options);
		}

		return $options;

	}


	public function delete ($search = []) {

		if (is_empty($search))
			return true;

		// Backwards Compatibility

		if (is_numeric($search)) {

			$search = array(
				$this->key() => $search
			);
		}

		$entries = false;

		// Read search on request

		if (
			($this->correlation && $this->table_ID()) ||
			($this->position() && $this->group()) ||
			$this->statistic()
		) {
			$entries = $this->read_select($search);
		}


		// Delete Entry

		if (!$this->core->db->delete(
			$this->db_table(),
			$search,
			$this->db_name()
		))
			return false;


		// Correlation

		if ($this->correlation && $this->table_ID()) {

			$this->core->correlation->delete(
				array(
					"table_ID_1" => $this->table_ID(),
					"entry_ID_1" => key_fetch($this->key(),$entries),
				)
			);

			$this->core->correlation->delete(
				array(
					"table_ID_2" => $this->table_ID(),
					"entry_ID_2" => key_fetch($this->key(),$entries),
				)
			);
		}


		// Position

		if ($this->position()) {

			if ($this->group()) {

				$group_IDs = key_fetch($this->group(), $entries);
				$group_IDs = $this->core->array->unique($group_IDs);

				foreach ($group_IDs as $group_ID) {

					$this->reposition(array($this->group() => $group_ID));

					if ($this->cache())
						$this->cache->reposition(array($this->group() => $group_ID));

				}

			}

			else {
				$this->reposition();

				if ($this->cache())
					$this->cache->reposition();
			}
		}


		//Statistics

		if ($this->statistic) {

			$this->core->statistic->delete(
				array(
					"table_ID" => $this->table_ID(),
					"entry_ID" => key_fetch($this->key(), $entries),
				)
			);

		}

		return true;

	}


	public function delete_tables () {

		if (not_empty($this->cache))
			$this->core->db->delete_table($this->db_table()."_cache");

		if (not_empty($this->properties->db_table()))
			$this->core->db->delete_table($this->properties->db_table());

		if (not_empty($this->db_table()))
			$this->core->db->delete_table($this->db_table());

		if (not_empty($this->group->properties->db_table()))
			$this->core->db->delete_table($this->group->properties->db_table());

		if (not_empty($this->group->db_table()))
			$this->core->db->delete_table($this->group->db_table());

		if (not_empty($this->statistic)) {
			$core_tables_o = $this->core->data_o();
			$core_tables_o->db_table("kb_tables");
			$core_tables_o->delete(array("name"=>$this->db_table()));
		}
		return true;
	}


	public function dropdown_position ($ID = false) {

		$positions_array = array();

		if ($ID) {

			$element = $this->entry($ID);

			if ($this->group()) {
				$positions = $this->read(array($this->group()=>$element[$this->group()]));
			}
			else {
				$positions = $this->read();
			}

			foreach ($positions as $position) {

				if ($element["position"] == $position['position'])
					$positions_array[""] = 'keep position';

				else if ($element["position"] > $position['position'])
					$positions_array[$position['position']] = 'before '.$position['name'];

				else
					$positions_array[$position['position']] = 'after '.$position['name'];
			}
		}
		else {
			$positions_array = array("NULL"=>"last position");
		}
		return $positions_array;
	}


	public function duplicate ($search = [], $replace = []) {

		if (is_empty($search))
			return false;

		// Backwards Compatibility

		if (is_string($replace) && not_empty($replace)) {

			deprecated("Replace should be an array when using replace()");

			$replace = array(
				"name" => $replace
			);
		}

		$entry = $this->entry($search);

		if (is_empty($entry))
			return false;

		$replace = keys_of(array_keys($entry), $replace);

		$entry = $this->core->array->merge($entry, $replace);

		if ($this->key())
			unset($entry[$this->key()]);

		if($this->position())
			unset($entry["position"]);

		if($this->statistic())
			unset(
				$entry["updated_by"],
				$entry["updated"],
				$entry["created"],
				$entry["created_by"]
			);

		return $this->create($entry);
	}


	public function entry ($filter = false, $options = false) {

		if (is_numeric($filter) && !$this->indexed())
			return $this->read_from_cache($filter);

		if (is_numeric($filter) && $this->key())
			$filter = [$this->key() => $filter];

		if (is_string($filter))
			return [];

		$options = $this->core->array->union(
			[
				"limit" => "1"
			],
			$options
		);

		$data = $this->read($filter, $options);
		$data = current($data);
		return $data;
	}


	public function entry_select ($where = []) {

		if (is_empty($where))
			return array();

		if (is_string($where) || is_int($where))
			$where = array($this->key() => $where);

		return $this->core->db->select_first($this->db_table(), $where);

	}


	public function exclude ($add = NULL) {

		if (!is_array($add) || is_empty($add))
			return either($this->private_config("exclude"), array());

		foreach ($add as $column => $value) {

			if (key_of($column, $this->columns()))
				continue;

			if ($this->statistic() && $this->core->array->match($column, array("created", "updated")))
				continue;

			if (is_empty($column) && is_array($value))
				continue;

			unset($add[$column]);

		}

		if (are_equal($this->private_config("exclude"),$add))
			return $add;


		$filters = $this->private_config("exclude");
		$filters = remove_keys_of(array_keys($add), $filters);

		$this->private_config("exclude",
			$this->core->array->union(
				$filters,
				$add
			)
		);

		$this->clear("cache");

		if ($this->cache())
			$this->cache->exclude($add);

		return either($this->private_config("exclude"), array());

	}


	public function expandable ($set = NULL) {

		if ($set === NULL)
			return $this->expandable;

		$this->expandable = $set;

	}


	public function filter ($add = NULL, $permanent = false) {

		if ($add === false) {

			$filter = $this->filter();

			if (is_empty($filter))
				return;

			foreach ($filter as $key => $value)
				$filter[$key] = false;

			$this->private_config("filter",$filter);
			return [];
		}

		if (!is_array($add) || is_empty($add))
			return either($this->private_config("filter"), array());

		foreach ($add as $column => $value) {

			if (key_of($column, $this->columns()))
				continue;

			if ($this->statistic() && $this->core->array->match($column, array("created", "updated")))
				continue;

			if (is_empty($column) && is_array($value))
				continue;

			if ($this->properties() && $this->key() && not_empty(key_of($this->key(), $this->columns()))) {

				$current_filter = key_of($this->key(), $this->filter());
				$new_filter = key_fetch($this->properties->key(), $this->properties->read(["property"=>$column,"value"=>$value]));

				if (not_empty($current_filter))
					$new_filter = $this->core->array->intersect($current_filter, $new_filter);

				// continue;
				// $this->properties->filter(["property"=>$column,"value"=>$value]);
				$this->filter([
					$this->key() => $new_filter
				]);

			}

			unset($add[$column]);

		}

		if (!$permanent && are_equal($this->private_config("filter"),$add))
			return $add;

		if ($permanent) {

			$filters = $this->private_config("permanent_filter");

			$this->private_config("permanent_filter",
				$this->core->array->union(
					$filters,
					$add
				)
			);

		}

		else {

			$filters = $this->private_config("filter");
			$filters = remove_keys_of(array_keys($add), $filters);

			$this->private_config("filter",
				$this->core->array->union(
					$filters,
					$add
				)
			);
		}

		$this->clear("cache");

		if ($this->cache())
			$this->cache->filter($add, $permanent);

		return either($this->private_config("filter"), array());

	}


	public function foreign ($set = NULL) {

		if ($set === NULL)
			return $this->foreign;

		if (any_empty($set, is_array($set))) {
			// $this->foreign = false;
			return $this->foreign;
		}

		$this->clear("cache");

		if ($set === false) {
			$this->foreign = false;
			return $this->foreign;
		}

		if ($this->cache()) {
			$this->cache->foreign($set);
		}

		$this->foreign = either($this->foreign, []);

		foreach ($set as $key => $value) {

			if (is_string($value)) {

				if ($object = $this->private_config("foreign_cache_".$value)) {
					$this->foreign[$key] = $object;
				}

				else {

					$object = $this->core->data($value, false);

					if($this->core->db->table_exists($value."_properties"))
						$object->properties(true);

					$this->foreign[$key] = $object;
					$this->private_config("foreign_cache_".$value,$object);

				}
			}

			else if (is_array($value)) {

				$this->foreign[$key] = $this->core->data(key($value), false);

				if($this->core->db->table_exists(key($value)."_properties"))
					$this->foreign[$key]->properties(true);

				$this->foreign[$key]->foreign(current($value));
			}
		}

		return $this->foreign;

	}


	public function foreign_values ($entry = []) {

		if (any_empty($entry, is_array($entry)))
			return $entry;

		foreach ($entry as $key => $value) {

			if (is_empty($object = key_of($key, $this->foreign())))
				continue;

			if (is_string($object)) {
				$object = key_of($key, $this->foreign(array($key=>$object)));
			}

			if (not_empty($value))
				$entry[$key] = $object->entry($value);

		}
		return $entry;
	}


	public function group ($create_object = false) {

		if ($create_object === false) {

			if (is_object($this->group))
				return "group_ID";

			return $this->group;

		}

		if (is_string($create_object) && not_empty($create_object)) {
			$this->group = $create_object;
			return $this->group;
		}

		$this->group = $this->core->data_group();

		if (is_array($create_object))
			$this->group->db_table($this, $create_object);
		else
			$this->group->db_table($this, true);

		$this->group->db_name($this->db_name());

		return $this->group;

	}


	public function limit ($start = NULL, $count = NULL) {

		$limit = $this->private_config("limit");

		// Get
		if ($start === NULL && $count === NULL)
			return $limit;

		// Set

		if (is_indexed($start))
			return $this->limit(key_of(0, $start), key_of(1, $start));

		if ($start === false && is_null($count))
			return $this->private_config("limit", false);

		// $start = either(ifttt(is_numeric($start),$start), 0);
		// $count = either(ifttt(is_numeric($count),$count), $this->pagination, "18446744073709551615");

		if (!is_numeric($start))
			return $limit;

		$limit = $start;

		if (is_numeric($count))
			$limit = array($start,$count);

		$this->private_config("limit", $limit);

		if ($this->cache())
			$this->cache->limit($limit);

		return $limit;
	}


	public function merge ($array = [], $key = false) {

		if (is_empty($array))
			return false;

		if ($key === false)
			$key = $this->key();

		if ($key === false)
			return $this->create($array);

		if (is_indexed($array)) {
			foreach ($array as $entry) {
				$return[] = $this->merge($entry, $key);
			}
			return $return;
		}

		if (is_empty(key_of($key, $array)))
			return $this->create($array);

		$entry = $this->read(key_of($key, $array));

		if (is_empty($entry))
			return $this->create($array);

		return $this->update($array);

	}


	public function order ($set = NULL) {

		$order = array();

		if (is_string($set) && not_empty($set))
			$set = array($set);

		if (is_array($set) && not_empty($set)) {

			if (are_equal($this->private_config("order"),$set))
				return $set;

			foreach ($set as $key => $value) {

				if (is_numeric($key)) {
					$key = $value;
					$value = "";
				}

				if (
					key_of($key, $this->columns()) ||
					($this->statistic() && $this->core->array->match($key, array("created", "updated")))
					)
					$order[$key] = $value;

			}

			$this->private_config("order",$order);
			$this->reset("cache");

			if ($this->cache())
				$this->cache->order($set);

		}

		return either($this->private_config("order"), array());

	}


	public function position ($set = NULL) {

		if ($set === NULL)
			return $this->position;

		$this->position = $set;

		if ($set)
			$this->order(array("position" => "ASC"));

		else
			$this->order(
				remove_keys_of("position",$this->order())
			);
	}


	public function properties ($create_object = false) {

		if ($create_object) {

			$this->indexed(false);

			$this->properties = $this->core->data_property();
			$this->properties->db_table($this);
			$this->properties->db_name($this->db_name());
		}

		if ($this->properties === false)
			return false;

		return true;

	}


	public function read ($filter = false, $options = false, $cache = true) {

		if ($this->cache())
			return $this->cache->read($filter, $options, $cache);

		if ($cache) {

			$filters = $this->core->array->union(
				$this->filter(),
				(array) $filter
			);

			$hash = md5(serialize(array($filters, $this->private_config("permanent_filter"), $this->limit(), $options, array_keys((array)$this->foreign))));
			// $hash = serialize(array($filter, $options));

			if (not_empty($cache = $this->config("cache_".$hash)))
				return $cache;

			$cache = $this->read($filter, $options, false);
			$this->config("cache_".$hash, $cache);
			return $cache;
		}

		if (not_empty($filter) && is_numeric($filter))
			return $this->read([$this->key() => $filter], $options);

		$search = either($this->search(), []);

		if (not_empty($filter))
			foreach ((array) $filter as $key => $value) {
				if (has_regex($value)) {
					$search[$key] = $value;
					$filter = remove_keys_of($key, $filter);
				}
			}

		if (not_empty($search)) {
			$limit = $this->limit();
			$this->limit(false);
		}

		if ($this->properties())
			$entries = $this->read_with_properties($filter, $options);

		else
			$entries = $this->read_select($filter, $options);

		if ($this->foreign() && !key_of("ignore_foreign",$options)) {
			foreach ($entries as $key => $entry)
				$entries[$key] = $this->foreign_values($entry);
		}

		if (not_empty($search)) {
			$entries = array_find($search, $entries);

			if (not_empty($limit)) {
				$entries = array_slice($entries, key_of(0,$limit), key_of(1,$limit));
				$this->limit($limit);
			}
		}

		if (!$this->indexed() && $this->key() && !$this->properties())
			$entries = $this->core->array->index($entries, $this->key());

		return $entries;

	}


	public function read_all ($options = false) {

		if ($this->read_all !== false && $options === false)
			return $this->read_all;

		if ($this->cache())
			$read_all = $this->cache->read_from_db($options);

		else
			$read_all = $this->read_from_db($options);

		$read_all = $this->correlation_data($read_all);

		// Cache
		if($options === false)
			$this->read_all = $read_all;

		return $read_all;

	}


	public function read_all_sorted_by ($by = "name") {

		if (!isset($this->read_all))
			$this->read_all = $this->read_from_db();

		$array = $this->read_all;

		$array = $this->core->array->sort_by($array,$by);

		return $array;
	}


	public function read_children ($ID = false) {

	 	if ($ID === false)
	 		return [];

		return either(
			(array) key_fetch($this->key(), $this->children($ID)),
			[]
		);
	}


	public function read_column ($column = false, $options = [], $select = NULL) {

		$options = $this->core->array->merge([
			"default" => false,
			"exclude" => false,
			"spacing" => "&nbsp;&nbsp;&nbsp;&nbsp;",
			"ID" => "",
			"depth" => -1,
			"group" => false,
		],$options);

		$hash = md5(serialize(array(
			$column,
			array_keys((array)$this->private_config("permanent_filter")),
			// remove_keys_of(["group"],$options),
			$select,
			array_keys((array)$this->foreign())
		)));



		$cache = [];

		if ($this->group()) {

			if (is_empty($data = $this->private_config("cache_".$hash))) {

				$data = $this->read($select);
				$data = $this->core->array->sort_by($data,$this->group());
				$data = $this->core->array->group_by($data,$this->group());
				$this->private_config("cache_".$hash, $data);

			}

			$options["ID"] = either(key_of("ID", $options), key($data));

			return $this->array_list($data, $column, $options, $this->group());
		}

		if (is_empty($data = $this->private_config("cache_".$hash))) {
			$data = $this->read($select);
			$this->private_config("cache_".$hash, $data);
		}

		return $this->array_list($data, $column, $options);

	}


	public function read_parents ($ID = false) {

		if (any_empty($ID, is_numeric($ID)))
			return [];

		$read_all = $this->read_all();
		$read_all = array_values($read_all);

		$parent_IDs = array();

		while ($entry = fetch(array("ID"=>$ID),$read_all)) {
			$ID = key_fetch($this->group(),$entry);
			if ($ID)
				$parent_IDs[] = $ID;
		}

		return $parent_IDs;

	}


	public function read_per ($key = false, $value = false, $limit = false) {

		if (is_empty($key))
			return array();

		if (!is_array($key)) {
			$key = array($key=>$value);
		}

		else {
			$limit = $value;
		}

		$options = array(
			"limit" => $limit,
		);

		$return = $this->read($key, $options);

		if (is_empty($return))
			return array();

		if ($limit == 1)
			return current($return);

		else
			return $return;

	}


	public function remove_order ($find = false) {

		if ($find) {

			if (is_empty($this->order))
				return true;

			if (is_array($this->order)) {

				foreach ($this->order as $key => $order) {

					if($order === $find) {

						unset($this->order[$key]);
						return $this->order;
					}

				}
				return true;
			}

			return false;
		}

		$this->order = array();
		return true;
	}


	public function reposition ($group = false, $entry = []) {

		if (not_empty($entry) && is_indexed($entries = $entry, true)) {

			foreach ($entries as $entry)
				if (!$this->reposition($group, $entry))
					return false;

			return true;
		}

		return $this->core->db->reorder($this->db_table(), "position", $group, $entry);
	}


	public function reset() {

		$args = func_get_args();

		if (is_empty($args))
			return false;

		if (in_array("cache", $args)) {
			$this->read_all = false;
			$this->read_cache = false;
			$this->group_by_parent_ID = false;
		}

		if (in_array("foreign", $args)) {
			$this->foreign = false;
			$this->read_all = false;
			$this->read_cache = false;
			$this->group_by_parent_ID = false;
		}

		if (in_array("correlation", $args)) {
			$this->correlation = false;
		}

		return true;
	}


	public function search ($column = NULL, $string = false) {

		$search = $this->private_config("search");

		if (is_empty($column))
			return either($search, []);

		if (has_regex($string)) {

			if (is_indexed($column)) {

				$search = remove_keys_of("", $search);

				$add_search = array();

				foreach ($columns = $column as $column)
					$add_search[][$column] = $string;

				$search[] = $add_search;
				$this->private_config("search",$search);

				return $search;

			}

			if (!is_string($column))
				return false;

			$search = remove_keys_of($column, $search);
			$search[$column] = $string;
			$this->private_config("search",$search);

			return $search;
		}

		if (not_empty($string))
			$string = "%".$string."%";

		if (is_indexed($column)) {

			$add_search = array();

			foreach ($columns = $column as $column)
				$add_search[$column] = $string;

			return $this->filter(array("" => $add_search));
		}

		if (!is_string($column))
			return false;

		return $this->filter(array($column => $string));
	}


	public function sort ($column = "", $where = []) {

		if (is_empty($column))
			return false;

		if ($this->position() === false)
			return false;

		if (key_of($column, $this->columns()) === NULL)
			return false;

		return $this->core->db->table->sort(
			$this->db_table(),
			"position",
			$column,
			$where
		);
	}


	public function statistic ($set = NULL) {

		if ($set === NULL)
			return $this->statistic;

		$this->statistic = $set;

	}


	public function table_ID ($set = NULL) {

		if (is_numeric($set))
			$this->private_config("table_ID",$set);

		$table_ID = $this->private_config("table_ID");

		if (not_empty($table_ID))
			return $table_ID;

		if (is_empty($this->db_table()))
			return false;

		$tables = $this->tables();

		$ID = key_of($this->db_table(), $tables);

		if (is_empty($ID) && $set === true) {

			$ID = $this->core->tables->create(array("name"=>$this->db_table()));
			$tables[$this->db_table()] = $ID;
			$this->core->tables->config("cache", $tables);
		}

		$this->private_config("table_ID", $ID);

		return $ID;

	}


	public function tables () {

		$tables = $this->core->tables->config("cache");

		if (is_empty($tables)) {
			foreach($this->core->tables->read() as $table) {
				$tables[key_of("name", $table)] = key_of("ID", $table);
			}
			$this->core->tables->config("cache",$tables);
		}

		return $tables;
	}


	public function update ($array, $where = []) {

		if (any_empty($array, is_array($array)))
			return true;


		// Multiple updates

		if (is_indexed($array)) {
			foreach($array as $entry) {
				$status[] = $this->update($entry, $where);
			}
			return $status;
		}


		// Single update

		if (is_numeric($where) && $this->key())
			$where = [$this->key() => $where];

		if (is_indexed($where))
			$where = keys_of($where, $array);

		if (is_empty($where) && not_empty(key_of($this->key(), $array)))
			$where = [$this->key() => key_of($this->key(), $array)];

		if (is_empty($where))
			return false;

		$ID = key_of($this->key(), $where);

		if (is_empty($ID) || is_array($ID))
			return $this->sql_update($array, $where);

		// Read structure of data table

		$columns = $this->columns();


		// Update position

		if ($this->position)
			$this->reposition($this->group(),$array);

		// Expand columns if allowed

		if (
			// $this->expandable &&
			count(array_keys($array)) > count($this->core->array->intersect(array_keys($array),array_keys($columns)))
		) {

			$additional_columns = $this->core->array->diff(array_keys($array),array_keys($columns));

			if ($this->statistic())
				$additional_columns = $this->core->array->diff($additional_columns, array("created","updated","created_by","updated_by"));


			if (
				not_empty($additional_columns) &&
				not_empty($additional_columns_update = keys_of($additional_columns, $this->db_column_options()))
			) {

				$additional_columns = $this->core->array->diff(
					$additional_columns,
					array_keys($additional_columns_update)
				);

				$this->core->db->table->expand($this->db_table(),$additional_columns_update);
				$columns = $this->columns();

			}

			if (not_empty($additional_columns) && $this->expandable()) {

				$this->core->db->table->expand($this->db_table(),$additional_columns);
				$columns = $this->columns();

			}

		}

		// Limit entry to available columns

		$entry = keys_of(array_keys($columns),$array);

		// Remove values not applicable

		if ($this->position())
			unset($entry["position"]);

		$entry = remove_keys_of(array_keys($where), $entry);

		// Update entry in data table

		$table = ifttt(
			$this->db_name(),
			$this->db_name().".".$this->db_table(),
			$this->db_table()
		);



		if ($this->core->db->update($table, $entry, $where) === false)
			return false;

		// Set statistics

		if ($this->statistic()) {

			$array["updated"] = either(key_of("updated", $array), "NOW()");
			$array["updated_by"] = either(key_of("updated_by", $array), $this->core->user->ID());

			$return = $this->core->statistic->update(
				array(
					"user_ID"		=> $array["updated_by"],
					"date"			=> $array["updated"],
				),
				array(
					"table_ID" => $this->table_ID(true),
					"entry_ID" 	=> $ID,
					"description" => "updated",
				)
			);

			if (are_equal($return,0)) {

				$this->core->statistic->delete(array(
					"table_ID"	=> $this->table_ID(true),
					"entry_ID"	=> $ID,
					"description"	=> "created"
				));

				$this->core->statistic->create(array(
					"table_ID"	=> $this->table_ID(true),
					"entry_ID"	=> $ID,
					"user_ID"		=> either(key_of("created_by", $array), $this->core->user->ID()),
					"date"			=> either(key_of("created", $array), "NOW()"),
					"description"	=> "created"
				));

				$this->core->statistic->delete(array(
					"table_ID"	=> $this->table_ID(true),
					"entry_ID"	=> $ID,
					"description"	=> "updated"
				));

				$this->core->statistic->create(array(
					"table_ID"	=> $this->table_ID(true),
					"entry_ID"	=> $ID,
					"user_ID"		=> either(key_of("updated_by", $array), $this->core->user->ID()),
					"date"			=> either(key_of("updated", $array), "NOW()"),
					"description"	=> "updated"
				));
			}

		}

		// Set properties

		if ($this->properties() && is_object($this->properties)) {


			$properties = $array;

			// Remove values not applicable
			if ($this->statistic())
				unset(
					$properties["created"],
					$properties["updated"],
					$properties["created_by"],
					$properties["updated_by"]
				);

			if ($this->position())
				unset($properties["position"]);

			// Remove columns already set
			$properties = $this->core->array->diff($properties, $entry);

			if (not_empty($properties)) {
				// Add ID and update properties
				$this->properties->update(
					$this->core->array->merge(array("ID"=>$ID),$properties)
				);
			}
		}

		// Update Cache

		if ($this->cache())
			$this->cache->update($array, $where);

		$this->reset("cache");

		return true;
	}



	// Alias

	public function clear() {
		return call_user_func_array(array($this, "reset"), func_get_args());
	}


	private function create_array_sorted($by="parent_ID") {
		//Find all parent_ID and sort them
		$array = $this->read();

		$array = $this->core->array->group_by($array,$by);
		return $array;

	}

	protected function dropdown_elements($id = false , $depth = -1, $spacing = "", $default = false, $exclude = false, $column = "name", $group=false) {
		// Deprecated
		if(!$id) $id = $this->root;
		if($id == NULL || $id == "" || $id == 0 || $id == "0") $id = NULL;

		if($data = $this->create_array_sorted("parent_ID"))
		{
			$i=0;
			static $elements;

			if($depth == -1)
			{
				$elements = array();
			}
			if($depth == -1 && $default != false)
			{
				$elements[$this->root] = $default;
			}
			$depth++;
			while(isset($data[$id][$i]))
			{
				$margin=str_repeat($spacing,intval($depth+1));
				if($data[$id][$i]['ID'] != $exclude)
				{
					if(isset($data[$id][$i][$column]))
						// if($group)
							// $elements[$data[$id][$i]['parent_ID']][$data[$id][$i]['ID']] = $margin.$data[$id][$i][$column];
						// else
							$elements[$data[$id][$i]['ID']] = $margin.$data[$id][$i][$column];

					$next = $data[$id][$i]['ID'];
				}
				if(isset($next))
				{
					$this->dropdown_elements($next, $depth, $spacing, $default, $exclude, $column, $group);
				}
				$i++;
			}
		}
		else if($data = $this->create_array_sorted("group_ID")){

			$i=0;
			static $elements;

			if($depth == -1)
			{
				$elements = array();
			}
			if($depth == -1 && $default != false)
			{
				$elements[$this->root] = $default;
			}
			$depth++;
			foreach($data as $data_group => $data_entries){
				foreach($data_entries as $data_entry){
						$elements[$data_group][$data_entry["ID"]] = $data_entry["name"];
				}
			}
			// while(isset($data[$id][$i]))
			// {
			// 	$margin=str_repeat($spacing,intval($depth+1));
			// 	if($data[$id][$i]['ID'] != $exclude)
			// 	{
			// 		if(isset($data[$id][$i][$column]))
			// 			// if($group)
			// 				// $elements[$data[$id][$i]['group_ID']][$data[$id][$i]['ID']] = $margin.$data[$id][$i][$column];
			// 			// else
			// 				$elements[$data[$id][$i]['ID']] = $margin.$data[$id][$i][$column];
			//
			// 		$next = $data[$id][$i]['ID'];
			// 	}
			// 	if(isset($next))
			// 	{
			// 		$this->dropdown_elements($next, $depth, $spacing, $default, $exclude, $column, $group);
			// 	}
			// 	$i++;
			// }
		}
		else
		{
			if($default != false)
			{
				$elements[$this->root] = $default;
			}
			$array = $this->read_all();
			// asort($array);
			foreach($array as $key => $value)
			{
				$elements[$value["ID"]] = $value["name"];
			}
		}
		return isset($elements) ? $elements : false;
	}



	// Private

	protected function array_list($data = false, $column = false, $options = [], $case=false){

		if (any_empty($data, $column))
			return $data;

		switch ($case){

			case "parent_ID":

				$elements = array();

				$i=0;

				if($options["depth"] == -1 && $options["default"] != false){
					if(is_array($options["default"]))
						$elements[key($options["default"])] = current($options["default"]);
					else
						$elements["NULL"] = $options["default"];
				}

				$options["depth"]++;

				if (key_of(key_of("ID", $options), $data))
				foreach ($data[$options["ID"]] as $key => $value) {

					$margin = str_repeat($options["spacing"], intval($options["depth"]));

					if (is_empty(key_of("ID", $value)))
						continue;

					if(
						!in_list($value['ID'],$options["exclude"]) &&
						not_empty($return = key_fetch($column, $value))
						)
						$elements[$value['ID']] = $margin.$return;

					$ID =	$value['ID'];
					$child_options = $options;
					$child_options["ID"] = $ID;
					$elements = $elements + $this->array_list($data, $column, $child_options, $case);

				}

				return $elements;

			case "group_ID":

				$elements = array();

				if(($default = key_fetch("default",$options)) !== false)
					$elements[NULL]	= $default;

				if($options["group"]){
					$group = $options["group"]->read_names();
				}

				foreach($data as $data_group => $data_entries)
					foreach($data_entries as $data_entry)
							if($options["group"])
								$elements[key_of($data_group,$group)][$data_entry["ID"]] = key_of($column,$data_entry);
							else
								$elements[$data_entry["ID"]] = $data_entry[$column];

				return $elements;

			case false:

				$elements = array();

				if(key_of("default",$options) != false)
					$elements[NULL]	= key_of("default",$options);

				foreach($data as $data_group => $data_entry)
					$elements[$data_entry[$this->key()]] = $data_entry[$column];

				return $elements;

		}


	}

	private function correlation_data($read_all) {

		if(!$this->correlation() || !$this->columns() || !$this->table_ID())
			return $read_all;

		$tables = $this->tables();
		$tables_ID = array_combine(array_values($tables), array_keys($tables));
		$correlation_data = $this->core->correlation->read(array("table_ID_2"=>$this->table_ID()));
		$correlation_data = $this->core->array->group_by($correlation_data,"table_ID_1");

		foreach($correlation_data as $table_ID => $entries) {

			if($this->correlation() === true || fetch($tables[$table_ID],$this->correlation())){

				$temp_data = $this->core->data($tables_ID[$table_ID]);

				if ($this->core->db->table_exists($tables_ID[$table_ID]."_properties"))
					$temp_data->properties(true);


				if($this->core->db->table_exists($tables_ID[$table_ID]."_cache"))
					$temp_data->cache(true);


				if ($this->indexed())
					foreach($read_all as $key => $array){
						$read_all[$key][$tables_ID[$table_ID]] = $temp_data->read(array("ID"=>fetch("entry_ID_1",fetch(array("entry_ID_2"=>key_of($this->key(), $array)),$entries))));
					}
				else
					foreach(array_keys($read_all) as $key){
						$read_all[$key][$tables_ID[$table_ID]] = $temp_data->read(array("ID"=>fetch("entry_ID_1",fetch(array("entry_ID_2"=>$key),$entries))));
					}
			}
		}

		return $read_all;
	}

	private function create_table() {

		if (is_empty($this->db_table()))
			return false;

		// Create Group Table first

		if ($this->group() && is_object($this->group) && !$this->core->db->table->exists($this->group->db_table()))
			$this->group->create_table();

		$table = [
			$this->key()	=> "INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY",
		];

		$table = $this->core->array->merge($table, $this->db_column_options());

		if ($this->position())
			$table["position"] = "INT UNSIGNED NULL";

		if ($this->group()) {

			$table[$this->group()] = "INT UNSIGNED NULL";

			$this->db_table_options("KEY `".$this->group()."` (`".$this->group()."`)");

			if (are_equal($this->private_config("type"), "data") && is_object($this->group))
				$this->db_table_options("FOREIGN KEY (`".$this->group()."`) REFERENCES `".$this->group->db_table()."` (`".$this->key()."`) ON DELETE CASCADE ON UPDATE CASCADE");

			else if (are_equal($this->private_config("type"), "group") && $this->group())
				$this->db_table_options("FOREIGN KEY (`".$this->group()."`) REFERENCES `".$this->db_table()."` (`".$this->key()."`) ON DELETE CASCADE ON UPDATE CASCADE");

		}

		if (
			($this->core->db->table->exists($this->db_table()) === false) &&
			($this->core->db->table->create($this->db_table(),$table, $this->db_table_options()) === false)
		)
			return false;

		// Make an entry in kb_tables

		if (is_empty($this->core->tables->read(array("name"=>$this->db_table()))))
			$this->core->tables->create(array("name"=>$this->db_table()));

		// Properties table

		if ($this->properties()) {

			$table = array(
				$this->properties->key() => "INT UNSIGNED NOT NULL",
				"property" => "VARCHAR( 255 ) COLLATE utf8_unicode_ci DEFAULT NULL",
				"value" => "MEDIUMTEXT COLLATE utf8_unicode_ci DEFAULT NULL",
			);

			$options = array(
				"KEY `".$this->properties->key()."` (`".$this->properties->key()."`)",
				"FOREIGN KEY (`".$this->properties->key()."`) REFERENCES `".$this->db_table()."` (`".$this->key()."`) ON DELETE CASCADE ON UPDATE CASCADE",
			);

			if(
				($this->core->db->table->exists($this->db_table()."_properties") === false) &&
				($this->core->db->table->create($this->db_table()."_properties",$table,$options) === false)
			)
				return false;
		}

		return true;

	}

	private function fill_cache_table($array = false) {

		if ($this->core->db->table->exists($this->cache->db_table()))
			return true;

		if (empty($array))
			// $array = $this->read_from_db(array("ignore_filter"=>true, "ignore_foreign"=>true));
			$array = $this->read_from_db();

		if (empty($array))
			return true;

		// SETUP TABLE
		foreach($this->core->db->table->columns($this->db_table()) as $column => $setting)
			$fields[$column] = $setting["COLUMN_TYPE"]." ".$setting["EXTRA"];

		foreach($array as $entry) {
			foreach($entry as $key => $value) {
				if(!array_key_exists($key,$fields))
					$fields[$key] = 'text COLLATE utf8_unicode_ci ';
			}
		}

		foreach($fields as $column => $setting){
			$sql_fields[] = $column." ".$setting;
		}

		$sql_fields = implode(", ",$sql_fields);

		$this->core->db->table->config("db_tables".$this->db_name(), "");

		$sql = 'CREATE TABLE '.DB_PREFIX.$this->db_table().'_cache (';
		$sql .= $sql_fields;
		if(array_key_exists("ID", $fields))
			$sql .= ', PRIMARY KEY(ID)';

		$sql .= '	) ENGINE=InnoDB DEFAULT CHARSET=utf8;';

		$sql2 = 'ALTER TABLE `'.$this->db_table().'_cache` ADD FOREIGN KEY ( `ID` ) REFERENCES `'.$this->db_table().'` (`ID`) ON DELETE CASCADE ON UPDATE CASCADE ;';

		// INSERT CONTENT

		if ($this->core->db->sql_query($sql) && $this->core->db->sql_query($sql2)) {

			$this->core->db->table->config("db_columns",array());


			if(is_empty($array))
				return true;

			$i=0;
			$keys = array_combine(array_keys($fields), array_fill(1,count($fields),'""'));

			foreach($array as $entry) {
				foreach($entry as &$value){
					$value = $value === NULL ? "NULL" : '"'.addslashes($value).'"';
				}
				$entry = array_merge($keys,$entry);
				$values[] = '('.implode(',',$entry).')';
			}

			// avoid memory being exhausted
			$count = count($values);
			$length = 100;
			$n = ceil($count/$length);
			$i = 0;

			while($i < $n) {
				$start = $i * $length;
				$values_chunk = array_slice($values,$start,$length);
				$sql = 'INSERT INTO '.DB_PREFIX.$this->db_table().'_cache ('.
				implode(",",array_keys($keys)).') VALUES '.
				implode(", ",$values_chunk).";";
				$this->core->db->sql_query($sql);
				$i++;
			}

			$keys = NULL;
			unset($keys);
			$values = NULL;
			unset($values);
		}

	}

	private function get_deprecated_filters($prefix = "") {

		if(not_empty($prefix))
			$prefix = $prefix.".";

		$filters = array();

		if(isset($this->filters) && not_empty($this->filters))
			foreach($this->filters as $filter_value)
				$filters[] = $prefix.$filter_value;

		return $filters;
	}

	protected function ID_as_index($array = "") {

		$data = $this->private_config("ID_as_index");

		if (is_empty($data)) {

			if (any_empty($array, is_array($array)))
				$array = $this->read();

			$data = array();

			foreach ($array as $entry)
				$data[key_of($this->key(), $entry)] = $entry;

			$this->private_config("ID_as_index", $data);
		}

		return $data;
	}

	private function read_from_cache($ID = "") {

		$cache = $this->ID_as_index();

		return key_of($ID, $cache);

	}

	private function read_from_db($options = []) {

		$options = $this->core->array->merge(
			array(
				"ignore_filter" => false,
				"ignore_foreign" => false,
				"limit" => $this->limit(),
				"filter" => false,
				"search" => false,
			),
			$options
		);

		$read_all = array();

		if (any_empty($this->db_table(),$this->columns()))
			return $read_all;

		$sql_all = $this->read_select(
			key_of("filter", $options),
			keys_of(array("limit"), $options)
		);

		if($sql_all)
			// Old memory saving way, deprecated
			/*
			while($array = each($sql_all))
			{
				foreach($array["value"] as $key => $value)
				{
					// $value = is_null($value) ? "NULL" : $value;
					if($this->key() && isset($array["value"][$this->key()]))
						$read_all[$array["value"][$this->key()]][$key] = $value;
					else
						$read_all[$array["key"]][$key] = $value;
				}
				$sql_all[$array["key"]] = NULL;
				unset($sql_all[$array["key"]]);
			}
			*/
			foreach ($sql_all as $array_key => $array_value) {

				if (!$this->indexed() && $this->key() && !is_null(key_of($this->key(), $array_value)))
					$read_all[key_of($this->key(), $array_value)] = $array_value;
				else
					$read_all[$array_key] = $array_value;

				$sql_all[$array_key] = NULL;
				unset($sql_all[$array_key]);
			}

		$sql_all = NULL;

		if($this->key() && $this->properties && $this->db_table() && $this->key() == "ID" && $this->core->db->table_exists($this->db_table()."_properties")){

			$property_o = $this->core->data_property_o();
			$property_o->db_table($this);
			$properties = $property_o->read_all();

			if($properties) {

				$prop_keys = key_fetch("property",$properties);
				$prop_keys = $this->core->array->unique($prop_keys);
				$prop_keys = array_combine($prop_keys,array_fill(1,count($prop_keys),""));

				foreach ($read_all as $ID => $entry){
					$read_all[$ID] = $this->core->array->merge(
						$prop_keys,
						$entry
					);
				}
				// Old memory saving way, deprecated
				/*
				while(list($key,$property) = each($properties)){
					if(!empty($read_all[$property["ID"]])){
						$read_all[$property["ID"]][$property["property"]] = $property["value"];
					}
				}
				*/

				foreach ($properties as $key => $property) {
					if(!empty($read_all[$property["ID"]])){
						$read_all[$property["ID"]][$property["property"]] = $property["value"];
					}
				}

			}

			$properties = NULL;
			unset($properties);
		}

		$this->count = $this->core->db->found_rows();

		if($this->foreign() && !key_of("ignore_foreign",$options)){
			foreach($read_all as $key => $entry){
				$read_all[$key] = $this->foreign_values($entry);
			}
		}

		return $read_all;
	}

	private function read_with_properties ($filter = [], $options = []) {

		if (is_empty($this->columns()))
			return [];

		$filters = $this->core->array->union(
			$this->filter(),
			(array) $filter
		);

		$filters_table = keys_of(array_keys($this->columns()), $filters);
		$filters_properties = remove_keys_of(array_keys($this->columns()), $filters);

		$properties_set = $this->properties->column_values("property");

		$filters_properties = keys_of($properties_set, $filters_properties);

		foreach ($filters_properties as $key => $value) {
			$filters_properties["property"] = $this->core->array->merge(key_of("property",$filters_properties), (array) $key);
			$filters_properties["value"] = $this->core->array->merge(key_of("value",$filters_properties), (array) $value);
			unset($filters_properties[$key]);
		}

		if (not_empty($this->properties->filter())) {
			$IDs = $this->properties->column_values("ID");
			$filters_table = $this->core->array->merge(
				$filters_table,
				["ID"=>$IDs]
			);
			// $this->properties->filter(false); // WHY?
		}

		$entries = $this->read_select($filters_table, $options);

		if (is_empty($entries))
			return [];

		$IDs = key_fetch($this->key(), $entries);

		if (is_empty($IDs))
			return $entries;

		$filters_properties["ID"] = $IDs;

		$entries = $this->core->array->index($entries, $this->key());

		$properties = $this->properties->read_select($filters_properties);

		if (is_empty($properties))
			return $entries;

		// $entries = fetch(["ID"=>key_fetch("ID", $properties)], $entries);
		// $properties = $this->properties->read_select(["ID" => array_keys($entries)]);

		foreach ($properties as $property)
			$entries[key_of($this->key(), $property)][key_of("property", $property)] = key_of("value", $property);

		return $entries;

	}

	private function read_with_properties_backup ($filter = [], $options = []) {

		if (is_empty($this->columns()))
			return [];

		$filters = $this->core->array->union(
			(array) $filter,
			$this->filter()
		);

		$filters_table = keys_of(array_keys($this->columns()), $filters);
		$filters_properties = remove_keys_of(array_keys($this->columns()), $filters);

		$properties_set = $this->properties->column_values("property");

		$filters_properties = keys_of($properties_set, $filters_properties);

		foreach ($filters_properties as $key => $value) {
			$filters_properties["property"] = $this->core->array->merge(key_of("property",$filters_properties), (array) $key);
			$filters_properties["value"] = $this->core->array->merge(key_of("value",$filters_properties), (array) $value);
			unset($filters_properties[$key]);
		}

		$entries = $this->read_select($filters_table, $options);
		$properties = $this->properties->read_select($filters_properties);

		$entries = fetch(["ID"=>key_fetch("ID", $properties)], $entries);

		if (is_empty($entries))
			return [];

		$entries = $this->core->array->index($entries, $this->key());
		$properties = $this->properties->read_select(["ID" => array_keys($entries)]);

		foreach ($properties as $property)
			$entries[key_of($this->key(), $property)][key_of("property", $property)] = key_of("value", $property);

		if (!$this->foreign())
			return $entries;

		foreach ($entries as $key => $value)
			$entries[$key] = $this->foreign_values($value);

		return $entries;

	}

	private function read_select($where = [], $options = []) {

		$order = $this->order();

		if (not_empty($order) && is_string($order)) {

				if (!$this->core->array->match($order, array_keys($this->columns())))
					$order = "";
		}

		if (not_empty($order) && is_array($order)) {

			foreach ($order as $key => $value) {

				$column = ifttt(is_numeric($key), $value, $key);

				if ($this->core->array->match($column, array_keys($this->columns())))
					continue;

				if ($this->statistic() && $this->core->array->match($column, array("created", "updated")))
					continue;

				unset($order[$key]);

			}

		}

		$options = $this->core->array->merge(
			array(
				"order" => $order,
				"limit" => $this->limit()
			),
			$options
		);

		$where = $this->core->array->union(
			$this->filter(),
			$where
		);

		if (not_empty($where) && is_array($where)) {

			foreach ($where as $key => $value) {

				$column = ifttt(is_numeric($key), $value, $key);

				if (is_empty($key) && is_array($value))
					continue;

				if ($this->core->array->match($column, array_keys($this->columns())))
					continue;

				if ($this->statistic() && $this->core->array->match($column, array("created", "updated")))
					continue;

				// return array(); //?
				unset($where[$key]);

			}

		}

		if (!$this->core->db->table->exists($this->db_table(), $this->db_name()))
			return array();

		$table = ifttt(
			$this->db_name(),
			$this->db_name().".".$this->db_table(),
			$this->db_table()
		);

		if ($this->statistic()) {

			$table = array(
				$table,
				$this->core->json->encode(array($this->core->statistic->db_table()=>"created")) => array(
					"date" => "created",
					"user_ID" => "created_by",
					array(
						"table_ID" => $this->table_ID(),
						"entry_ID" => $this->db_table().".`ID`",
						"description" => "created",
					)
				),
				$this->core->json->encode(array($this->core->statistic->db_table()=>"updated")) => array(
					"date" => "updated",
					"user_ID" => "updated_by",
					array(
						"table_ID" => $this->table_ID(),
						"entry_ID" => $this->db_table().".`ID`",
						"description" => "updated",
					)
				)
			);
		}
		/*
		if (not_empty($permanent_filter = $this->private_config("permanent_filter")))
			$where = array(
				$where,
				$permanent_filter
			);
		*/

		if (not_empty($permanent_filter = $this->private_config("permanent_filter"))) {

			foreach ($where as $key => $value) {
				if (!is_null(key_of($key, $permanent_filter))) {
					$where[$key] = $this->core->array->intersect($value, key_of($key, $permanent_filter));
					$permanent_filter = remove_keys_of($key, $permanent_filter);
				}
			}

			$where = $this->core->array->union(
				$where,
				$permanent_filter
			);
		}

		if (not_empty($exclude = $this->private_config("exclude")))
			$where = array(
				"include" => $where,
				"exclude" => $exclude
			);

		return $this->core->db->select(
			$table,
			$where,
			key_of("order", $options),
			key_of("limit", $options)
		);
	}

	private function sql_insert($entries = []) {

		if (any_empty(not_empty($entries), is_indexed($entries)))
			return;

		$columns = [];

		foreach ($entries as $entry) {

			if (is_empty($columns))
				$columns = array_keys($entry);

			else
				$columns = $this->core->array->intersect($columns, array_keys($entry));

		}

		$columns = $this->core->array->intersect($columns, array_keys($this->columns()));

		foreach ($entries as $index => $entry) {
			$entries[$index] = keys_of($columns, $entry);
		}

		return $this->core->db->insert($this->db_table(), $columns, $entries, $this->db_name());

	}

	private function sql_update($update = [], $where = []) {

		if (any_empty(not_empty($update), is_array($update), !is_indexed($update)))
			return;

		$update = keys_of(array_keys($this->columns()), $update);
		$where = keys_of(array_keys($this->columns()), $where);

		$table = ifttt(
			$this->db_name(),
			$this->db_name().".".$this->db_table(),
			$this->db_table()
		);

		return $this->core->db->update($table, $update, $where);

	}



	// Should get deprecated

	public function read_names ($default = false, $exclude = false, $spacing = "", $ID = NULL, $depth = -1) {

		$options = array(
			"default" => $default,
			"exclude" => $exclude,
			"spacing" => $spacing,
			"ID" => $ID,
			"depth" => $depth,
			"group" => false,
		);
		$data = $this->read_all();
		return $this->array_list($data,"name",$options);

	}



	// Deprecated

	public function add_filter($filters = false, $values = false) {

		deprecated();

		if(!$filters)
			return false;

		if(is_string($filters) && $values === false) {

			$this->filters[] = $filters;
		}

		else if (is_indexed($filters) && $values === false) {

			foreach($filters as $filter)
				$this->add_filter($filter);
		}

		else if (is_array($filters) && !is_indexed($filters) && $values === false) {

			return $this->filter($filters);

		}

		else if (is_string($filters) && $values !== false) {

			return $this->filter(array($filters => $values));
		}

		$this->clear("cache");
	}

	public function add_order($order = "") {

		deprecated("Use order() instead of add_order()");

		return $this->order($order);
		/*
		if (!is_array($this->order))
			$this->order = array($this->order);

		if (!empty($order) && !is_array($order) && !in_array($order,$this->order)) {
			$this->order[] = $order;
			$this->reset("cache");
		}

		else if (!empty($order) && is_array($order)) {
			foreach($order as $f) {
				if (!in_array($f,$this->order)) {
					$this->order[] = $f;
					$this->reset("cache");
				}
			}
		}
		*/
	}

	public function add_search ($column = NULL, $search = NULL) {

		deprecated("Use search() instead of add_search()");
		return $this->search($column, $search);

	}

	public function db_columns(){

		deprecated();

		if(empty($this->auto_create))
			return false;

		if($this->auto_create === true)
			return true;

		$return = array();

		if(is_array($this->auto_create)){

			foreach($this->auto_create as $key => $value){

				if(is_numeric($key))
					$return[] = $value;

				else
					$return[] = $key;

				}
		}
		return $return;

	}

	public function filters ($prefix = "") {

		deprecated("filters() is deprecated, use filter()");
		return $this->get_deprecated_filters($prefix);

	}

	public function unset_variables() {
		deprecated("unset_variables() is deprecated");
		$this->read_all = false;
		$this->read_cache = false;
		$this->group_by_parent_ID = false;
	}

	public function read_all_sorted ($by = "parent_ID") {
		deprecated();
		// if(!isset($this->group_by_parent_ID))
		// {
			//	($this->create_array_sorted($by) == false) return false;
		// }
		return $this->create_array_sorted($by);
		// return $this->group_by_parent_ID;
	}

	public function read_all_structured ($parent_ID = false, $maxdepth = false, $depth = -1) {

		deprecated();

		$read_all = $this->ID_as_index();

		if($parent_ID === false && $this->group())
			$parent_ID = $this->core->array->get_highest_group_key($read_all,$this->group());

		if($parent_ID){
			$children = $this->read_children($parent_ID);
			$read_all = array_find(array("ID"=>$children),$read_all);
		}

		if($this->group()){
			$read_all = $this->core->array->group_by($read_all,$this->group());
			$read_all = $this->core->array->ungroup_by($read_all,"ID", $parent_ID);
		}

		if($maxdepth !== false)
			foreach ($read_all as $key => $entry)
				if($entry["depth"] > $maxdepth)
					unset($read_all[$key]);

		return $read_all;
	}

	public function read_dropdown ($default = false, $exclude = false, $spacing = "&nbsp;&nbsp;&nbsp;&nbsp;", $ID = NULL, $depth = -1) {

		deprecated();

		$options = array(
			"default" => $default,
			"exclude" => $exclude,
			"spacing" => $spacing,
			"ID" => either($ID,""),
			"depth" => $depth,
			"group" => false,
		);
		return $this->read_column("name",$options);
	}

	public function read_new() {
		deprecated("read_new() is deprecated");
		$this->unset_variables();
	}

	public function remove_filter($filter = false){

		deprecated();

		if ($filter) {

			if (is_indexed($filter)) {
				foreach($filter as $f)
					$this->remove_filter($f);
			}

			else if (is_array($filter)) {

			}

			else if (($key = array_search($filter, $this->filters)) !== false)
				unset($this->filters[$key]);
		}

		else{
			unset($this->filters);
		}

		$this->reset("cache");
	}

	public function set_limit ($start = false, $count = false) {

		deprecated("Use limit() instead of set_limit()");

		return $this->limit($start, $count);
	}

	/*
	public function read_backup ($filter = false, $options = false, $cache = true) {

		if ($this->cache())
			return $this->cache->read($filter, $options, $cache);

		if ($cache) {

			$filters = $this->core->array->union(
				(array) $filter,
				$this->filter()
			);

			$hash = md5(serialize(array($filters, $this->private_config("permanent_filter"),  $options, array_keys((array)$this->foreign))));
			// $hash = serialize(array($filter, $options));

			if (not_empty($cache = $this->config("cache_".$hash)))
				return $cache;

			$cache = $this->read($filter, $options, false);
			$this->config("cache_".$hash, $cache);
			return $cache;
		}

		if (not_empty($filter) && is_numeric($filter) && is_empty($options))
			return $this->read([$this->key() => $filter]);

		if ($this->indexed() && is_empty(remove_keys_of(array("filter", "order"), $options)) && !has_regex($filter)) {

			$entries = $this->correlation_data($this->read_select($filter, $options));

			if ($this->foreign() && !key_of("ignore_foreign",$options)) {
				foreach ($entries as $key => $entry)
					$entries[$key] = $this->foreign_values($entry);
			}

			if (not_empty($search = $this->search())) {
				$entries = array_find($search, $entries, "", "");
			}

			return $entries;

		}

		if ($this->properties() && $this->config("experimental_read_properties"))
			return $this->read_with_properties($filter);

		$data = $this->read_all(remove_keys_of("limit",$options));
		// if($this->read_cache && isset($this->read_cache[serialize($filter)]))
			// return $this->read_cache[serialize($filter)];

		if (not_empty($search = $this->search()))
			$data = array_find($search, $data, "", "");

		if ($filter === false)
			return $data;

		if (is_array($filter))
			$data = array_find($filter,$data);

		else if (is_string($filter) || is_int($filter)) {

			// Faster first
			if(isset($data[$filter]) && $this->indexed === false)
				$data = array($filter=>$data[$filter]);

			// Slower next
			else
				$data = array_find(array($this->key()=>$filter),$data, false);
		}

		// $this->read_cache[serialize($filter)] = $data;
		return $data;
	}

	*/
}


class KB_data_group extends KB_data{

	public $group = "parent_ID";
	public $akin = false;
	public $order = array();

	private $core;

	// Deprecated
	public $db_name = "";

	function __construct($core) {

		parent::__construct($core);
		$this->core = $core;

		$this->config("db",array(
			"table" => "kb_data_group",
			"name" => ""
		));

		$this->private_config("type", "group");

	}


	public function db_name($name = "") {

		if(is_object($name)){
			$this->db_name($name->db_name());
		}

		else if(is_string($name) && not_empty($name)) {
			$db = $this->config("db");
			$db["name"] = $name;
			$db = $this->config("db",$db);
		}

		return key_of("name",$this->config("db"));

	}


	public function db_table($name = "", $auto_create = false, $akin = false){

		if (is_empty($name))
			return either(
				$this->db_name, // <- Deprecated, but supported
				key_of("table",$this->config("db"))
			);

		if (is_object($name)) {
			$this->db_table($name->db_table()."_group");
			if (is_bool($auto_create)) {
				$this->auto_create = $auto_create;
			}
			else if (is_array($auto_create)) {
				$this->auto_create = true;
				$this->db_column_options($auto_create);
			}
			$this->akin = $name;
		}

		else if (is_string($name)) {
			$db = $this->config("db");
			$db["table"] = $name;
			$db = $this->config("db",$db);

			if (is_bool($auto_create)) {
				$this->auto_create = $auto_create;
			}
			else if (is_array($auto_create)) {
				$this->auto_create = true;
				$this->db_column_options($auto_create);
			}

			$this->akin = $akin;
		}

		return either(
			$this->db_name, // <- Deprecated, but supported
			key_of("table",$this->config("db"))
		);

	}


	public function parents($ID = false, $include_self = false) {

		$return = array();

		if (is_empty($ID))
			return $return;

		if (!$this->group())
			return $return;

		$entry = $this->entry($ID);

		if (is_empty($entry))
			return $return;

		if ($include_self)
			$return[] = $entry;

		if ($parent = key_of($this->group, $entry))
			$return = $this->core->array->merge(
				$return,
				$this->parents($parent, true)
			);

		return $return;

	}


	public function children($ID = false, $include_self = false, $cache = NULL) {

		if ($ID === false)
	 		return false;

		if (is_null($array = $cache)) {
			$array = $this->read();
			$array = $this->core->array->group_by($array,"parent_ID");
		}

		$return = array();

		if ($include_self)
			$return[] = $this->entry($ID);

		if (isset($array[$ID])) {

			foreach ($array[$ID] as $value) {

				$return[] = $value;
				$return = $this->core->array->union(
					$return,
					$this->children(key_of($this->key(), $value), false, $array)
				);
			}
		}

		return $return;

	}


	public function duplicate($search = array(), $replace = array()) {

		if (is_empty($search))
			return false;

		// Backwards Compatibility

		if (is_string($replace) && not_empty($replace)) {

			deprecated("Replace should be an array when using duplicate()");

			$replace = array(
				"name" => $replace
			);
		}

		$entry = $this->entry($search);

		if (is_empty($entry))
			return false;

		$ID = key_of($this->key(), $entry);

		$replace = keys_of(array_keys($entry), $replace);

		$entry = $this->core->array->merge($entry, $replace);

		if ($this->key())
			unset($entry[$this->key()]);

		if ($this->position)
			unset($entry["position"]);

		if ($this->statistic)
			unset(
				$entry["updated_by"],
				$entry["updated"],
				$entry["created"],
				$entry["created_by"]
			);

		$new_ID = $this->create($entry);

		// Create Data Entries

		if($this->akin && $ID) {

			if(is_object($this->akin)) {

				$entries = $this->akin->read_per("group_ID",$ID);

				foreach($entries as $entry) {
					unset($entry["ID"]);
					$entry["group_ID"] = $new_ID;
					$this->akin->create($entry);
				}

			}

			else {

				$data = $this->core->data_o();
				$data->db_table($this->akin);
				$entries = $data->read_per("group_ID",$ID);

				foreach($entries as $entry) {
					unset($entry["ID"]);
					$entry["group_ID"] = $new_ID;
					$data->create($entry);
				}
			}
		}

		// Duplicate Subgroups

		if($ID && ($array = $this->read_per("parent_ID",$ID))) {

			foreach($array as $entry)
			$this->duplicate(array("ID"=>$entry["ID"]),array("parent_ID"=>$new_ID));
		}

	}


	public function select_parent($ID=false,$options=array()){

		$options = array_merge(array(
			"default" => false,
			"exclude" => $ID ? array_merge(array($ID),$this->read_children($ID)) : false,
			"spacing" => "&nbsp;&nbsp;&nbsp;&nbsp;",
			"ID" => "",
			"depth" => -1,
			"group" => false,
			"column"=>"name"
		),$options);

		$data = $this->read();

		if($this->group()){
			$data = $this->core->array->group_by($data,$this->group());
			return $this->array_list($data, $options["column"], $options,$this->group());
		}

		else if(!empty($data)){
			return $this->array_list($data, $column, $options);
		}

		return array();

	}


	public function group($create_object = false){
		/*
		if ($create_object) {

			$this->group = $this->core->data_group_o();
			$this->group->db_table($this);
			$this->group->db_name($this->db_name());
		}
		*/
		if(is_object($this->group))
			return "parent_ID";

		else
			return $this->group;

	}


	public function read_structured($key = false, $maxdepth = NULL) {

		// $array = $this->read();

		if (any_empty($this->key(), $this->group()))
			return [];

		if (is_numeric($key))
			$array = $this->children($key);

		else
			$array = $this->read();

		$array = $this->core->array->hierarchie($array, $this->key(), $this->group(), "depth");

		if (!is_numeric($maxdepth))
			return $array;

		if (are_equal($maxdepth, 0))
			$array = $this->core->array->filter($array, ["depth" => 0]);

		else
			$array = $this->core->array->range_by($array, "depth", [0,$maxdepth]);

		return $array;

	}

	// Private

	private function read_ungroup($array, $key = "", $maxdepth = NULL, $depth = 0) {

		$return = array();

		if (!is_null($maxdepth) && are_equal($depth, $maxdepth))
			return $return;

		foreach ((array) key_of($key, $array) as $key => $value) {

			$return[$key] = $this->core->array->union(
				$value,
				array("depth"=>$depth)
			);

			if (not_empty($sub_array = $this->read_ungroup($array, $key, $maxdepth, $depth+1)))
				$return = $this->core->array->union(
					$return,
					$sub_array
				);

		}

		return $return;
	}


}


class KB_data_property extends KB_data{

	private $core;

	// public $key = false;
	public $group = false;

	// Deprecated
	public $db_name = "";

	function __construct($core) {

		parent::__construct($core);
		$this->core = $core;

		$this->config("db",array(
			"table" => "kb_data_properties",
			"name" => ""
		));

		// $this->key(false);
		$this->key("ID");

	}


	public function db_table($name = false, $auto_create = false) {

		if(is_string($name) && not_empty($name)) {
			deprecated("db_table discourages strings at the moment");
			$db = $this->config("db","table",$name."_properties");

			if (is_bool($auto_create)) {
				$this->auto_create = $auto_create;
			}

			else if (is_array($auto_create)) {
				$this->auto_create = true;
				$this->db_column_options($auto_create);
			}
		}

		else if (is_object($name) && ($name instanceof KB_data)) {
			$db = $this->config("db","table",$name->db_table()."_properties");
			if (is_bool($auto_create)) {
				$this->auto_create = $name->auto_create;
			}
			else if (is_array($auto_create)) {
				$this->auto_create = true;
				$this->db_column_options($auto_create);
			}
		}

		return either(
			$this->db_name, // <- Deprecated, but supported
			key_of("table",$this->config("db"))
		);

	}


	public function create($array = array()) {

		$ID = key_of("ID",$array);

		if (is_empty($ID))
			return false;

		unset($array["ID"]);

		$properties = array();

		foreach($array as $key => $value)
			if (not_empty($value))
				$properties[] = array($ID, $key, $value);

		if (not_empty($properties))
			return $this->core->db->insert(
				$this->db_table(),
				array("ID","property","value"),
				$properties
			);

		return true;
	}


	public function update($array, $where = array()) {

		$ID = key_of("ID",$array);

		if (is_empty($ID))
			return false;

		unset($array["ID"]);

		$this->core->db->delete(
			$this->db_table(),
			array(
				"ID" => $ID,
				"property" => array_keys($array)
			)
		);

		$properties = array();

		foreach ($array as $key => $value) {
			if (not_empty($value)) {
				$properties[] = array($ID,$key,$value);
			}
		}

		if (is_empty($properties))
			return true;

		return $this->core->db->insert(
			$this->db_table(),
			array("ID","property","value"),
			$properties
		);

	}


	function set($array){

		if(is_array($array)){

			$result = $this->read_per(array("ID"=>$array["ID"],"property"=>$array["property"]));
			//echo "ID=>".$array["ID"].",property=>".$array["property"];
			// New entry
			if(empty($result)){
				if(!empty($array["value"]))
										$this->create($array);
			}

			//Update or delete
			else{

				if(!empty($array["value"])) $this->update($array);

				else $this->delete(array("ID"=>$array["ID"],"property"=>$array["property"]));
			}

		}
		else
		{
			return false;
		}
	}

	function get($array){
		if(is_array($array) && isset($array["ID"]) && isset($array["property"])){
			$result = $this->read_per(array("ID"=>$array["ID"],"property"=>$array["property"]),"",1);
			if(!empty($result))
			{
				return	$result["value"];
			}
			else
			{
				return false;
			}
		}
		else if($array == (int) $array){
			$result = $this->read_per(array("ID"=>$array),"");
			if(!empty($result))
			{
				foreach($result as $entry)
				{
					$result_new[$entry["property"]] = $entry["value"];
				}
				return	$result_new;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
}


class KB_correlation extends KB_data {

	private $core;

	function __construct($core) {
		parent::__construct($core);
		$this->core = $core;
		$this->key(false);
		$this->db_table("kb_correlation");

	}


	public function set($table_1, $ID, $table_2, $entries = NULL) {

		if (any_empty($table_1, $ID, is_numeric($ID), $table_2, is_array($entries)))
			return false;

		if (is_object($table_1) && $table_1 instanceof KB_data)
			$table_1 = $table_1->table_ID();

		if (is_object($table_2) && $table_2 instanceof KB_data)
			$table_2 = $table_2->table_ID();

		if (any_empty(is_numeric($table_1), is_numeric($table_2)))
			return false;

		$this->delete(array(
			"table_ID_1" => $table_1,
			"entry_ID_1" => $ID,
			"table_ID_2" => $table_2,
		));

		if (is_empty($entries))
			return false;

		foreach($entries as $entry) {

			if (!is_numeric($entry))
				continue;

			$this->create(array(
				"table_ID_1" => $table_1,
				"entry_ID_1" => $ID,
				"table_ID_2" => $table_2,
				"entry_ID_2" => $entry,
			));
		}
	}


	public function get($table_1, $ID, $table_2) {

		if (is_indexed($ID)) {

			$correlates = [];

			foreach ($IDs = $ID as $ID) {

				if (not_empty($result = $this->get($table_1, $ID, $table_2)))
					$correlates[$ID] = $result;
			}

			return $correlates;

		}

		if (any_empty($table_1, $ID, is_numeric($ID), $table_2))
			return array();

		if (is_object($table_1) && $table_1 instanceof KB_data)
			$table_1 = $table_1->table_ID();

		if (is_object($table_2) && $table_2 instanceof KB_data)
			$table_2 = $table_2->table_ID();

		if (is_scalar($table_2) && !is_numeric($table_2))
			$table_2 = $this->core->data($table_2)->table_ID(true);

		if (any_empty(is_numeric($table_1), is_numeric($table_2)))
			return false;

		$read = array(
			"table_ID_1" => $table_1,
			"entry_ID_1" => $ID,
			"table_ID_2" => $table_2,
		);

		$read = $this->read($read);

		$read = key_fetch("entry_ID_2",$read);

		return $read;
	}


	function return_table_ID($name = false){

		deprecated();

		if ($name) {

			$data_o = $this->core->kb_tables();
			$entry = $data_o->read_per("name", $name, 1);

			if($entry)
			return $entry["ID"];

			else
			return false;

		}else{
			return false;
		}
	}


	function get_entry($entry = array()){
		deprecated();
		$return = $this->get($entry);
		$return = current($return);
		return $return["entry_ID_1"];
	}


	public function table_has_table($source_table = false, $source_filter = false, $has_table = false, $has_table_filter = false, $set_entry = false){

		deprecated();

		// REMOVE
		$this->remove_filter();

		if($source_table && $has_table) {

			// Initialize Objects
			$source_table_o = $this->core->data_o();
			$source_table_o->db_table($source_table);
			$has_table_o = $this->core->data_o();
			$has_table_o->db_table($has_table);

			if (any_empty($source_table_o->columns(), $has_table_o->columns()))
				return array();

			//Set Table IDs
			$source_table_ID = $source_table_o->table_ID(true);
			$has_table_ID = $has_table_o->table_ID(true);

			//Set Filter
			// $this->add_filter("table_ID_1 = ".$source_table_ID);
			// $this->add_filter("table_ID_2 = ".$has_table_ID);
			$filter["table_ID_1"] = $source_table_ID;
			$filter["table_ID_2"] = $has_table_ID;

			if($source_filter){
				if(!is_array($source_filter)) $source_filter = array($source_filter);
				// $this->add_filter("entry_ID_1 IN (".implode(",",$source_filter).")");
				$filter["entry_ID_1"] = $source_filter;
			}

			if (not_empty($has_table_filter))
				$filter["entry_ID_2"] = $has_table_filter;

			//Read Correlation
			$array = $this->read($filter);

			//Set Correlation
			if($set_entry)
			{
				//Purge
				if (not_empty($has_table_filter))

					$this->delete(array(
							"table_ID_1" 	=> $source_table_ID,
							"entry_ID_1"	=> $has_table_filter,
							"table_ID_2"	=> $has_table_ID,
							"entry_ID_2"	=> $set_entry,
					));

				else

					$this->delete(array(
							"table_ID_1" 	=> $source_table_ID,
							"table_ID_2"	=> $has_table_ID,
							"entry_ID_2"	=> $set_entry,
					));

				//Insert
				foreach($source_filter as $key => $value)
				//while(list($key,$value) = each($source_filter))
				{
					$this->create(array(
						"table_ID_1" 	=> $source_table_ID,
						"table_ID_2"	=> $has_table_ID,
						"entry_ID_1"	=> $value,
						"entry_ID_2"	=> $set_entry,
					));
				}
								// $this->remove_filter();
				return true;
			}
						// $this->remove_filter();
			return $array;
		}
		else
		{
			return array();
		}
	}


	public function user_has($has_table = false, $set_entries = false){
		deprecated();
		$array = $this->table_has_table("kb_user",$this->core->user->data("ID"),$has_table,false,$set_entries);
		$array = $this->core->array->implode($array,"entry_ID_2");
		return $array;
	}


	public function user_group_has($has_table = false, $set_entries = false, $filter = false){
		deprecated();
		$array = $this->table_has_table("kb_user_group",$this->core->user->data("group_ID"),$has_table,$filter,$set_entries);
		$array = $this->core->array->implode($array,"entry_ID_2");
		return $array;
	}


	public function user_groups_have($has_table = false, $array = false, $set_entries = false, $filter = false){
		deprecated();
		return $this->table_has_table("kb_user_group",$array,$has_table,$filter,$set_entries);
	}


	function entry_has_user_groups($has_table = false, $ID = false){
		deprecated();
		$array = $this->table_has_table("kb_user_group",false,$has_table,$ID);
		$array = $this->core->array->implode($array,"entry_ID_1");
		return $array;
	}

}
