<?php


class KB_form extends KB_class {


	private
		$core,
		$description,
		$files,
		$values = NULL;

	public
		$identifier = "",
		$message = [],
		$element = [];


	function __construct ($core) {

		$this->core = $core;

		$this->config(array(
			"exclude_from_mail" => [],
			"subject" => "Kitbooth Form Mail",
			"error_class" => "message",
			"table_class" => "form wrap",
			"array_fields" => array("checkbox", "hsl"),
			"values_fields" => array("checkbox", "radio","select"),
			"file_fields" => array("file", "image", "video"),
			"upload" => true
		));

	}


	public function element ($name, $array = []) {

		if (is_empty($element = key_of($name, $this->element)))
			return;

		if (any_empty($array, is_array($array), !is_indexed($array)))
			return $element;

		foreach ($array as $key => $value)
			$this->element[$name][$key] = $value;

		return $this->element[$name];
	}


	public function elements ($return_submit = false) {


		$elements = $this->element;

		if (is_empty($elements))
			return [];

		if ($return_submit)
			return $elements;

		foreach ($elements as $key => $element)
			if (are_equal(key_of("type", $element), "submit"))
				unset($elements[$key]);

		return $elements;


	}


	public function field ($name = "", $array = false, $value = NULL) {

		if (is_null(key_of($name, $this->element)))
			return "";

		$this->element[$name]["markup"] = $this->markup($name);

		if ($this->element[$name]["array"] && $array === false) {

			$return = "";

			$this->element[$name]["markup"]["data-array"] = $this->identifier.$this->element[$name]["name"].'[0]';

			if (is_empty($this->element[$name]["value"]))
				return $this->field($name, 0, "");

			foreach ((array) $this->element[$name]["value"] as $key => $value) {

				if (is_string($value))
					$value = stripslashes($value);

				$this->element[$name]["markup"]["data-array"] = $this->identifier.$this->element[$name]["name"]."[".$key."]";

				$return .= $this->field($name, $key, $value);
			}

			if (not_empty($this->element[$name]["value"]) && fetch(key_of("type", $this->element[$name]), $this->config("file_fields")))
				$return .= $this->field($name, $key, "");

			return $return;
		}

		// Backwards compatibility
		// Set values for elements with predefined values
		if (fetch($this->element[$name]["type"],$this->config("values_fields"))) {

			if (
				is_empty(key_of("values",$this->element[$name])) &&
				is_array($values = key_of("value", $this->element[$name])) &&
				count($values) === 2
			) {

				$this->element[$name]["values"] = $values[1];
				$this->element[$name]["value"] = $values[0];

			}
		}


		if ($value === NULL)
			$value = $this->element[$name]["value"];

		if (not_empty($this->values()))
			$value = $this->value(key_of("name", $this->element[$name]));

		switch ($this->element[$name]["type"]) {

			case "select":
				return $this->gen_select($name, $value);

			case "radio":
				return $this->gen_radio($name, $value);

			case "checkbox":
				return $this->gen_checkbox($name, $value);

			case "textarea":
				return $this->gen_textarea($name, $value);

			case "html":
				return $this->gen_html($name, $value);

			case "markdown":
				return $this->gen_markdown($name, $value);

			case "file":
				$return = $this->gen_file($name, $value);
				if ($array === false && not_empty($value)){
					$return .= $this->gen_file($name, "", ["required"=>false]);
				}
				return $return;

			case "image":
				$return = $this->gen_file($name, $value, array("accept"=>"image/*"));
				if ($array === false && not_empty($value)){
					$return .= $this->gen_file($name, "", ["accept"=>"image/*","required"=>false]);
				}
				return $return;

			case "video":
				$return = $this->gen_file($name, $value, array("accept"=>"video/*"));
				if ($array === false && not_empty($value)){
					$return .= $this->gen_file($name, "", ["accept"=>"video/*","required"=>false]);
				}
				return $return;

			case "submit":
				return $this->gen_button($name, "submit", $value);

			case "button":
				return $this->gen_button($name, false, $value);

			case "captcha":
				return $this->gen_captcha($name, $value);

			case "hsl":
				return $this->gen_hsl($name, $value);

			default:
				return $this->gen_input($name,$this->element[$name]["type"], $value);
		}
	}


	public function file ($element = "") {

		if (is_empty($element))
		return [];

		$files = $this->files();

		return key_of($element,$files);

	}


	public function files ($element = []) {

		if (!isset($this->files))
		$this->files = $this->get_files();

		if (not_empty($element))
		return keys_of($element, $this->files);

		return $this->files;

	}


	public function label ($form_array) {

		if (isset($this->element[$form_array]) && !empty($this->element[$form_array])) {

			if (value_fetch(key_of("type",$this->element[$form_array]), array("hidden", "submit", "button")))
				return "";

			$this->description	= $this->element[$form_array]["description"] == '' ? '' : 'title="'.$this->element[$form_array]["description"].'"';

			return $this->gen_label($form_array);
		}
	}


	public function message ($name) {
		return $this->gen_message($name);
	}


	public function new_element ($element, $array = false) {

		$this->element[$element] = $this->element_template($element);

		if ($array)
			$this->set_element($element,$array);
	}


	public function new_elements ($elements = false) {

		if (!is_array($elements) || is_empty($elements))
			return false;

		foreach ($elements as $element => $array) {

			if (is_int($element))
				continue;

			$this->new_element($element, $array);
		}

	}


	public function render ($primary_option = true, $options = []) {

		if (is_empty($this->element))
			return "";

		if (is_bool($primary_option) && !are_empty($groups = $this->config("groups"), is_array($groups))) {

			$return = [];
			$elements = $this->element;

			foreach ($groups as $label => $fields) {
				$elements = remove_keys_of($fields, $elements);
				$return[] = $this->core->html->fieldset([
					$this->core->html->legend($this->core->language->translate($label)),
					$this->render($fields)
				]);
			}

			$return[] = $this->render(array_keys($elements));


			if ($primary_option === true)
				$return[] = $this->render(array_keys((array)fetch(["type"=>"submit"],$this->element)),["include_submit"=>true]);

			return $this->core->html->render($return);
		}

		$options = $this->core->array->union(
			["columns" => 2],
			$options
		);

		$include_submit = false;
		$elements = true;

		if (is_bool($primary_option))
			$include_submit = $primary_option;

		else if (is_array($primary_option))
			$elements = $primary_option;

		else if (is_string($primary_option))
			$elements = array($primary_option);

		else
			return "";

		$table = $this->core->view->table();

		$table->config(
			"class", $this->config("table_class")
		);

		$table->pagination = false;

		$array = $this->render_as_array($elements, $options);

		if (are_equal(key_of("columns", $options), 1)) {

			$elements = fetch(["type"=>"submit"],$this->element);

			$array = $this->core->array->union(
				$array,
				$this->render_as_array(array_keys($elements), ["include_submit"=>true])
			);

			foreach($array as $row => $elements) {
				foreach($elements as $element)
					if (not_empty($element))
						$array[] = $element;
				unset($array[$row]);
			}

			return $this->core->html->table($array, "", ["class"=>"form"]);
		}

		$label = false;

		foreach ($array as $entry) {
			if (not_empty($entry["label"])) {
				$label = true;
				break;
			}
		}

		if ($label) {
			$table->new_column(
				"label", array(
					"title"	=> "",
					"cell_markup" => ifttt($label_class = $this->config("label_class"), 'class="'.$label_class.'"'),
				)
			);
		}

		$table->new_column(
			"field",array(
				"title"	=> "",
				"cell_markup" => ifttt($field_class = $this->config("field_class"), 'class="'.$field_class.'"'),
			)
		);

		$table->new_row("message",array(
			"cell_markup" => 'class="red"'
		));

		$return = [];

		if (not_empty($array))
			$return[] = $table->render($array);

		if ($include_submit) {

			$elements = fetch(array("type"=>"submit"),$this->element);

			if (is_empty($elements))
				return $this->core->html->render($return);

			$submit = [];

			foreach($elements as $key => $element)
				$submit[] = $this->field($key);

			if (not_empty($submit))
				$return[] = $this->core->html->tag("div", $submit, array("class"=>"submit"));

		}

		return $this->core->html->render($return);
	}


	public function render_as_array ($elements = true, $options = []) {

		$return = [];

		$options = $this->core->array->union(
			["include_submit" => false],
			$options
		);

		if ($elements === true)
			$elements = array_keys($this->element);

		if (is_empty($elements))
			return $return;

		foreach ((array)$elements as $name) {

			if (is_null($element = key_of($name, $this->element)))
				continue;

			if (
				are_equal(key_of("type", $element), "submit") &&
				are_equal(key_of("include_submit", $options), false)
				)
				continue;

			$return[] = array(
				"label"	=> $this->label($name),
				"field"	=> 	preg_replace('-\$-',"&dollar;",$this->field($name)),
				"message"	=> 	$this->message($name),
			);

		}
		return $return;
	}


	public function render_form ($include_submit = true, $options = []) {

		if (empty($this->element))
			return "";

		if (count(array_find(array("type"=>"submit"),$this->element)) == count($this->element))
			return "";

		return $this->core->html->render(
			$this->core->html->tag("form", array($this->render($include_submit, $options)), array(
				"method" => "post",
				"accept-charset" => "utf-8",
				"enctype" => ifttt(fetch(array("type" => "file"), $this->element), "multipart/form-data")
			)+(array)key_of("markup", $options))
		);

	}


	public function set_element ($element, $array = []) {

		if (any_empty($array, is_array($array), !is_indexed($array)))
			return;

		if (is_empty(key_of($element, $this->element)))
			return;

		foreach ($array as $key => $value)
			$this->element[$element][$key] = $value;

		return;
	}


	public function set_values ($values = []) {

		if (any_empty($values, is_array($values), !is_indexed($values)))
			return false;

		foreach ($values as $key => $value) {

			$value = $this->core->json->decode($value);

			if (is_empty($element = key_of($key, $this->element)))
				continue;

			if (is_empty($type = key_of("type", $element)))
				continue;

			$this->element[$key]["value"] = $value;

		}

	}


	public function submit () {

		// Remove possible cache
		$this->values = NULL;

		if (is_empty($this->element))
			return false;

		if (is_empty($this->values(true)))
			return false;

		$elements = fetch(array("type"=>"submit"),$this->element);

		if (is_empty($elements))
			return false;

		$values = key_fetch("name",$elements);

		if (is_array($values))
			$values = array_values($values);

		if (not_empty(key_of($values,$this->values(true))))
			return $this->check_form();

		return false;
	}


	public function upload () {

		if (is_empty($this->element))
			return true;

		foreach ($this->element as &$element) {

			if (fetch($element["type"], $this->config("file_fields"))) {

				$files = $this->file($element["name"]);

				if ($files && key_of("save", $element)) {

					if (is_indexed($files)) {

						foreach ($files as &$file) {

							if ($file["error"])
								continue;

							$this->core->file->move($file['tmp_name'],$this->core->path("root").$this->core->filesystem->dir().$file["name"]);

							$file = $file["name"];

							if (matches("/frontend/files", $this->core->filesystem->dir()))
								$file = preg_replace("-(.*)/frontend/files-","",$this->core->filesystem->dir()).$file;

							else
								$file = $this->core->filesystem->dir().$file;


						}
					}

					else {

						if ($files["error"])
							continue;

						$this->core->file->move($files['tmp_name'],$this->core->path("root").$this->core->filesystem->dir().$files["name"]);

						$files = $files["name"];

						if (matches("/frontend/files", $this->core->filesystem->dir()))
							$files = preg_replace("-(.*)/frontend/files-","",$this->core->filesystem->dir()).$files;

						else
							$files = $this->core->filesystem->dir().$files;

					}
				}

				$this->value($element["name"], $files);

			}
		}
	}


	public function values ($element = false) {

		if (is_array($element) && not_empty($element) && !is_indexed($element))
			return $this->set_values($element);

		if (is_null($this->values)) // Deactivating means severe performance degration
			$this->values = $this->get_values();

		if (not_empty($element) && !is_bool($element))
			return keys_of($element, $this->values);

		if ($element === true)
			return $this->values;

		$submits = fetch(array("type"=>"submit"),$this->element);

		return remove_keys_of(array_keys(either($submits, [])),$this->values);

	}


	public function value ($element = false, $set = NULL) {

		if (is_empty($element))
			return [];

		if (!is_null($set)) {
			$this->values[$element] = $set;
			$this->core->set($this->identifier.$element, $set);
		}

		if (key_of($element, $this->element))
			return key_of($element, $this->values(true));

		return 	$this->core->get($this->identifier.$element, array("set","post","get"));

	}



	// Alias


	public function check () {
		return $this->submit();
	}

	// public function element ($name = "", $array = false, $value = NULL) {
	// 	return $this->field($name, $array, $value);
	// }

	public function new_field ($element, $array = false) {
		return $this->new_element($element, $array);
	}

	public function new_fields ($elements = false) {
		return $this->new_elements($elements);
	}



	// Private

	private function check_captcha ($a) {

		$hash = htmlentities($this->values[$a['name']."_captcha"]);
		$value = htmlentities($this->values[$a['name']]);

		$db = $this->core->data_o("kb_tmp");
		$entry = $db->read(array("identity"=>"captcha","value01"=>$hash,"value02"=>$value));

		if (is_empty($entry)) {
			$this->element[$a["name"]]["message"] = $a["message"] ? $a["message"] : "Please verify your input";
			$this->element[$a["name"]]["error"] = true;
			$this->return = false;
		}

		if (not_empty($hash))
			$db->delete(array("identity"=>"captcha", "value01"=>$hash));

		$this->values[$a['name']."_captcha"] = '';
		$this->values[$a['name']] = '';
	}

	private function check_element ($a) {

		switch ($a["type"]) {

			case 'input':
			case 'datetime':
			case 'date':
			case 'hidden':
			case 'password':
			case 'password_double':
			case 'textarea':
			case 'html':

				return $this->check_text($a);

			case 'radio':
			case 'checkbox':
			case 'select':
			case 'select2':

				return $this->check_select($a);


			case 'file':
				return $this->check_file($a);

			case 'captcha':

				return $this->check_captcha($a);

			default:
				return $this->check_text($a);

		}
	}

	private function check_file ($element) {

		if (!key_of("essential", $element))
			return true;

		if (not_empty($this->file(key_of("name", $element))))
			return true;

		$this->element[key_of("name", $element)]["error"] = true;

		return false;

	}

	private function check_form () {

		if (is_empty($this->element) || !is_array($this->element))
			return false;

		$return = true;

		foreach ($this->element as $name => $element) {
			if (!$this->check_element($element)) {
				$return = false;
				$this->element[$name]["error"] = true;
			}
		}

		// Remove Cache
		$this->values = NULL;

		return $return;

	}

	private function check_select ($a) {

		if (key_of("essential",$a) && is_empty($this->value($a["name"])))
			return false;

		return true;
	}

	private function check_text ($a) {

		$value = $this->value($a["name"]);
		$charset = key_of("charset",$a);

		if (key_of("essential",$a) && is_empty($value))
			return false;

		if (is_empty($value))
			return true;

		if (not_empty($charset)) {

			if (preg_match(preg_mod("mail"), $charset) && is_email($value))
				return true;

			if (is_empty(preg_replace(preg_mod("(".$charset.")+"), "", $value)))
				return true;

			return false;
		}

		return true;

	}

	private function element_template ($element) {

		$element = [
			"array" =>	false,
			"charset" =>	"",
			"class" =>	"",
			"description" => "",
			"disabled" =>	false,
			"error" =>	false,
			"essential" =>	false,
			"label" =>	$element,
			"markup" =>	"",
			"message" =>	$this->core->language->translate("A valid value is required"),
			"name" =>	$element,
			"output" => false,
			"placeholder"	=> "",
			"translate" =>	false,
			"type" =>	"text",
			"value" =>	"",
			"values" =>	"",
		];

		$element["required"] =& $element["essential"];
		$element["options"] =& $element["values"];
		return $element;
	}

	private function element_value ($value = "", $type = "") {

		if (is_numeric($value))
			return $value;

		if (is_bool($value))
			return $value;

		if (is_string($value))
			return htmlspecialchars($value);

		if (is_array($value) || is_object($value))
			return $this->element_value($this->core->json->encode($value));

	}

	private function gen_button ($name = "", $type = "submit", $value = "") {

		$element = key_of($name, $this->element);

		if (any_empty($element, $type))
			return "";

		$label = either(key_of("label", $element), key_of("value", $element), $name);

		// Backwards compatibility
		if (are_equal($label, $name))
			$label = either(key_of("value", $element), $label);

		if (is_string($label))
			$label = $this->core->language->translate($label);

		$field = $this->core->html->button(array($label), array(
			"type" => $type,
			"id" => $this->id($name),
			"name"	=> $this->name($name),
			"value"	=> htmlspecialchars(either($value, key_of("value", $element), $name)),
			"disabled" => ifttt($element["disabled"], true, false),
		)+$element["markup"]);

		return $this->core->html->render($field);
	}

	private function gen_captcha ($form_arr, $m=0, $l=0, $i=0) {

			// Grossbuchstaben erzeugen
			for ($x = 65; $x <= 90; $x++) {
				if ($x != 73 && $x != 76 && $x != 79)
					$alphabet[] = chr($x);
			}

			// Kleinbuchstaben erzeugen
			for ($x = 97; $x <= 122; $x++) {
				 if ($x != 105 && $x != 108 && $x != 111)
				 	$alphabet[] = chr($x);
			}

			// Zahlen erzeugen
			for ($x = 48; $x <= 57; $x++) {
				if ($x != 48 && $x != 49)
					$alphabet[] = chr($x);
			}
			shuffle($alphabet);
			$alph = implode("", array_slice($alphabet,0,4));
			$hash = md5(microtime());

			$color = $this->core->config('captcha_background') ? $this->core->config('captcha_background') : "ffffff";
			$border = isset($form_arr[9])&&!empty($form_arr[9]) ? $form_arr[9] : "000000";

			$sql = "INSERT INTO ".DB_PREFIX."kb_tmp (identity, value01, value02, value03, value04) VALUES ('captcha', '".$hash."', '".$alph."', '".$color."', '".$border."')";
			$this->core->db->sql_query($sql);

			echo "<img src=\"/captcha/".$hash.".png\" alt=\"SecurityCode\" style=\"margin:0\"> ";
			echo "<input type=\"hidden\" name=\"".$form_arr."_captcha\" value=\"".$hash."\"> ";
			echo "<br><input type=\"input\" name=\"".$form_arr."\" value=\"\"> ";

			$sql = "DELETE FROM ".DB_PREFIX."kb_tmp WHERE identity = 'captcha' AND time < DATE_SUB(NOW(), INTERVAL 1 DAY)";
			$this->core->db->sql_query($sql);

			$this->gen_input($form_arr);
	}

	private function gen_checkbox ($name = "", $value = "") 	{

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$values = either($element["values"], []);

		if (is_indexed($values, true))
			$values = array_combine($values,$values);

		if (!is_array($value))
			$value = array($value);

		$elements = [];

		foreach ($values	as $key => $label) {

			$label = $this->core->language->translate($label);

			$elements[] =
				$this->core->html->tag("span", [
					$this->core->html->tag("input",array(
						"type" => "checkbox",
						"name" => $this->name($name),
						"id" => $this->id($name).$key,
						"value" => $key,
						"checked" => value_fetch($key,$value)
					)),
					$this->core->html->tag("label",$label, ["for"	=> $this->id($name).$key])
				],
				[
					"data-field-type" => "option"
				]);

			// if ($checkbox_separator = $this->config('checkbox_separator'))
				// $elements[] = $checkbox_separator;

		}

		$field = $this->core->html->tag("span",$elements,$element["markup"]);
		$field = $this->core->html->property($field,"data-field-name",$name);
		$field = $this->core->html->property($field,"data-field-type","checkbox");

		return $this->core->html->render($field);

	}

	private function gen_datalist($name = "") {

		if (any_empty(
			$name,
			$element = key_of($name, $this->element),
			$values = key_of("values",$element),
			is_array($values)
		))
			return;

			$options = [];

			if (is_indexed($values, true))
				$values = array_combine($values, $values);

			foreach ($values as $key => $value) {
				$options[] = $this->core->html->option($value, ["value" => $key]);
			}

			return $this->core->html->datalist($options, ["id"=>$this->id($name)."_datalist"]);

	}

	private function gen_file ($name = "", $value = "", $markup = []) {

		$element = key_of($name, $this->element);

		$element["markup"] = $this->core->array->merge(
			$element["markup"],
			$markup
		);

		if (is_empty($element))
			return "";

		if (is_empty($value)) {

			$field = $this->core->html->tag("input", array(
				"type" => "file",
				"id" => $this->id($name),
				"name"	=> $this->name($name),
				"required" => ifttt(($element["required"] && key_of("required", $markup) !== false), true, false),
				"multiple" => (bool) key_of("multiple", $element),
				"data-label" => $this->core->language->translate("Drag files here"),
			)+$element["markup"]);
		}

		else {

			if (is_array($value)) {
				$path = key_of("path", $value);
				$filename = key_of("filename", $value);
			}
			else {
				$filename = $value;
				$path = $this->core->folder->path(key_of("path", $element)).$value;
			}

			$element["markup"]["class"] = "file_preview";

			$field = $this->core->html->tag(
				"div",
				array(
					$this->core->format->file(array(
						"filename" => $filename,
						"path" => $path,
					), false,
						array(
							"return_as_html"=>false,
							"render_filetypes"=>true,
							"thumbnail" => either($this->core->config("form_thumbnail_size"),"150"),
							"size" => "1500",
						)
					),
					$field = $this->core->html->tag("textarea", $this->core->json->encode($value), array(
						"id" => $this->id($name),
						"name"	=> $this->name($name),
					))
				),
				array(
					"type" => "file",
					"id" => $this->id($name),
					"name"	=> $this->name($name),
					"required" => false, //ifttt($element["required"], true, false),
					"data-label" => $this->core->language->translate("Drag files here"),

				)+remove_keys_of("data-array",$element["markup"])
			);
		}

		return $this->core->html->render($field);

	}

	private function gen_hsl ($name, $value = "") {

		$elements[] = array(
			"input"	=> array(
				"type" => "range",
				"min" => "0",
				"max" => "255",
				"name"	=> $this->name($name),
				"value"	=> key_of("0",$value),
			)
		);

		$elements[] = array(
			"input"	=> array(
				"type" => "range",
				"min" => "0",
				"max" => "255",
				"name"	=> $this->name($name),
				"value"	=> key_of("1",$value),
			)
		);

		$elements[] = array(
			"input"	=> array(
				"type" => "range",
				"min" => "0",
				"max" => "255",
				"name"	=> $this->name($name),
				"value"	=> key_of("2",$value),
			)
		);

		$markup = $this->core->html->render($elements);

		return $markup;
	}

	private function gen_html ($name = "", $value = "") {

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$this->core->template->backend->head(
			$this->core->html->tag("style",array(
				'html .ck.ck-toolbar{font-size: 0.85em;}'
			))
		);

		$this->core->template->backend->head(
			$this->core->template->backend->js(DROOT.BACKEND_TEMPLATE."editor/ckeditor.js")
		);

		$this->core->template->backend->head(
			$this->core->template->backend->js(DROOT.BACKEND_TEMPLATE."editor/translations/".key_of(0,$this->core->language->accept()).".js")
		);

		$field = $this->core->html->tag("textarea",$value,array(
			"id" => $this->id($name),
			"name" => $this->name($name),
			"class" => "html_editor",
		)+$element["markup"]);

		return $this->core->html->render($field);
	}

	private function gen_input ($name = "", $type = "text", $value = "") {

		$element = key_of($name, $this->element);

		if (any_empty($element, $type))
			return "";

		$field = $this->core->html->tag("input", [
			"type" => $type,
			"id" => $this->id($name),
			"name" => $this->name($name),
			"value" => $this->element_value($value),
			"list" => ifttt($list = $this->gen_datalist($name), $this->id($name)."_datalist", NULL),
			"placeholder" => either($this->core->language->translate($element["placeholder"]), false),
			"disabled" => ifttt($element["disabled"], true, false),
			"required" => ifttt($element["required"], true, false),
			"data-field-name" => $name,
		]+$element["markup"]);

		return $this->core->html->render([$field, $list, $this->gen_output($name)]);
	}

	private function gen_label ($name) {

		if (is_empty($element = key_of($name,$this->element)))
		 	return "";

		$label = key_of("label",$element);
		$title = key_of("title",$element);

		if (is_empty($label))
			return "";

		$label = $this->core->language->translate($label);
		$title = $this->core->language->translate($title);

		if ($this->element[$name]["essential"] == true)
			$label = $this->core->html->tag("strong",$label);

		$return = $this->core->html->tag(
			"label",
			$label,
			[
				"for" => $this->id($name),
				"title" => $title,
			]
		);

		$return = $this->core->html->render($return);

	 	return $return;

	}

	private function gen_markdown ($name = "", $value = "") {

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$this->core->template->backend->head(
			$this->core->html->render($this->core->html->tag("style",array(
				'html .ck.ck-toolbar{font-size: 0.85em;}'
			)))
		);

		$this->core->template->backend->head(
			$this->core->template->backend->js(DROOT.BACKEND_TEMPLATE."editor/ckeditor_markdown.js")
		);

		$this->core->template->backend->head(
			$this->core->template->backend->js(DROOT.BACKEND_TEMPLATE."editor/translations/".key_of(0,$this->core->language->accept()).".js")
		);

		$field = $this->core->html->tag("textarea",$value,array(
			"id" => $this->id($name),
			"name" => $this->name($name),
			"class" => "html_editor markdown",
		)+$element["markup"]);

		return $this->core->html->render($field);
	}

	private function gen_message ($form_array) {
		if (isset($this->element[$form_array]["message"]) && $this->element[$form_array]["error"]) {
			return "<span class=\"".$this->config("error_class")."\">".$this->element[$form_array]["message"]."</span>".$this->config('error_separator');
		}
	}

	private function gen_output($name = "") {

		if (any_empty(
			$name,
			$element = key_of($name, $this->element),
			key_of("output", $element)
		))
			return;

			return $this->core->html->output(key_of("value", $element), ["name"=>$this->id($name)."_output", "for"=>$this->id($name)]);

	}

	private function gen_radio ($name = "", $value = ""){

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$values = either(key_of("values",$element), []);

		if (is_indexed($values, true))
			$values = array_combine($values,$values);

		$elements = [];

		foreach ($values as $key => $label) {

			$label = $this->core->language->translate($label);

			$elements[] = $this->core->html->tag("input",array(
				"type" => "radio",
				"id" => $this->id($name).$key,
				"name" => $this->name($name),
				"value"	=> $key,
				// "checked" => fetch($key,$value),
				"checked" => either(are_equal($key,$value), are_empty($key,$value)),
				"disabled" => ifttt($element["disabled"], true, false),
				"required" => ifttt($element["required"], true, false),
			)+$element["markup"]);

			$elements[] = $this->core->html->tag("label",$label,array(
				"for"	=> $this->id($name).$key,
			));
		}

		return $this->core->html->render($elements);

	}

	private function gen_select ($name = "", $value = "") {

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$values = either(key_of("values",$element), []);

		if (!is_array($values))
			$values = [];

		if (is_indexed($values, true))
			$values = array_combine($values,$values);

		$elements = [];

		foreach ($values as $key => $label) {

			if (is_array($label)) {

				if (is_indexed($label, true))
					$label = array_combine($label,$label);

				$child_elements = [];

				foreach ($label as $child_key => $child_label) {

					if (key_of("translate", $element))
						$child_label = $this->core->language->translate($child_label);

					$child_elements[] = $this->core->html->tag("option",$child_label,array(
						"value"	=> $child_key,
						"selected" => are_equal($child_key,$value),
					));
				}

				$elements[] = $this->core->html->tag("optgroup",array($child_elements),array(
					"label"	=> $key,
				));

			}

			else {

				if (key_of("translate", $element))
					$label = $this->core->language->translate($label);

				$elements[] = $this->core->html->tag("option",$label,array(
					"value"	=> $key,
					"selected" => either(are_equal($key,$value), are_empty($key,$value)),
				));
			}
		}


		$field = $this->core->html->tag("select", $elements, array(
			"id" => $this->id($name),
			"name"	=> $this->name($name),
			"disabled" => ifttt($element["disabled"], true, false),
			"required" => ifttt($element["required"], true, false),
		)+$element["markup"]);

		return $this->core->html->render($field);

	}

	private function gen_textarea ($name = "", $value = "") {

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return "";

		$field = $this->core->html->tag("textarea", $this->element_value($value), array(
			"id" => $this->id($name),
			"name"	=> $this->name($name),
			"readonly" => ifttt($element["disabled"], true, false),
			"required" => ifttt($element["required"], true, false),
			"placeholder" => either($element["placeholder"], false),
		)+$element["markup"]);

		return $this->core->html->render($field);
	}

	private function get_files (){

		$values = [];

		if (empty($this->core->session->session_ID))
			return $values;

		if (!empty($this->element))

			foreach($this->element as $element){

				if (fetch($element["type"],$this->config("file_fields"))) {

					$value = $this->core->get($this->identifier.$element["name"],array("files"));

					if (is_empty(key_of("name",$value)))
						continue;

					if (is_indexed(key_of("name",$value))) {

						$i = 0;

						while(not_empty(key_of($i,key_of("name",$value)))) {

							foreach(array_keys($value) as $key)
								$values[$element["name"]][$i][$key] = key_of($i,key_of($key,$value));

							$i++;
						}

					}

					else {

						$values[$element["name"]] = $value;
					}
				}
			}

		return $values;

	}

	private function get_values () {

		$values = [];

		if (empty($this->core->session->session_ID))
			return $values;

		if (!empty($this->element))
		foreach ($this->element as $element) {

			if ($this->core->array->match($element["type"], array("select","radio"))) {

				$value = $this->core->get($this->identifier.$element["name"], array("set", "post","get"));

				if ($value === NULL)
					continue;

				$values[$element["name"]] = $value;
			}

			else if (fetch($element["type"], $this->config("file_fields"))) {

				$value = $this->core->get($this->identifier.$element["name"], array("set","post","get"));
				$files = $this->file($element["name"]);

				if ($value === NULL && $files === NULL)
					continue;

				if (not_empty($value)) {

					if (is_array($value)) {

						foreach ($value as $key => $entry) {
							if (is_empty($entry))
								unset($value[$key]);
						}


						array_walk_recursive(
							$value,
							function (&$value){
								$value = $this->core->json->decode(htmlspecialchars_decode($value,ENT_QUOTES));

							}
						);

					}

					else{
						$value = $this->core->json->decode(htmlspecialchars_decode($value,ENT_QUOTES));
					}

				}

				if ($files && key_of("save", $element)) {

					if (is_indexed($files)) {

						foreach ($files as $file) {

							if ($file["error"])
								continue;

							if ($this->config("upload"))
								$this->core->filesystem->move($file['tmp_name'],DROOT.$this->core->filesystem->base_dir().$file["name"]);

							$return = $file["name"];

							if (matches("/frontend/files", $this->core->filesystem->base_dir()))
								$return = preg_replace("-(.*)/frontend/files-","",$this->core->filesystem->base_dir()).$return;

							else
								$return = $this->core->filesystem->base_dir().$return;

							$values[$element["name"]][] = $return;
						}
					}

					else {

						if ($files["error"])
							continue;

						if ($this->config("upload"))
							$this->core->filesystem->move($files['tmp_name'],DROOT.$this->core->filesystem->base_dir().$files["name"]);

						$return = $files["name"];

						if (matches("/frontend/files", $this->core->filesystem->base_dir()))
							$return = preg_replace("-(.*)/frontend/files-","",$this->core->filesystem->base_dir()).$return;

						else
							$return = $this->core->filesystem->base_dir().$return;

						$values[$element["name"]] = $return;
					}
				}

				else {

					if ($element["array"])
						$values[$element["name"]] = $value;

					else
						$values[$element["name"]] = either(key_fetch("name", $files), $value);

					// $values[$element["name"]] = $this->core->json->decode($value);
				}
			}

			else if (are_equal($element["type"],"checkbox")) {

				$value = $this->core->get_var($this->identifier.$element["name"], array("set", "post","get"));

				if ($value === NULL) {

					if (fetch($this->identifier.".*", array_keys($this->core->get_var(true, array("post", "get", "set")))))
						$value = [];
					else
						continue;
				}

				if (not_empty($value))
					foreach ($value as $k => $v)
						if ($v == "NULL")
					 		$value[$k] = NULL;

				$values[$element["name"]] = $value;
			}

			else if (in_list($element["type"],array("datetime-local","number","date"))) {

				$value = $this->core->get_var($this->identifier.$element["name"], array("set", "post","get"));

				if ($value === NULL)
					continue;

				if ($value === "")
					// $value = "NULL";
					$value = NULL;

				$values[$element["name"]] = $value;
			}

			else {

				$value = $this->core->get($this->identifier.$element["name"], array("set", "post","get"));

				if ($value === NULL)
					continue;

				if (is_array($value)){

					$values[$element["name"]] = $value;

						array_walk_recursive(
							$values[$element["name"]],
							function (&$value){
								$value = htmlspecialchars_decode($value,ENT_QUOTES);
							}
						);

				}
				else{
					$values[$element["name"]] = htmlspecialchars_decode(trim($value),ENT_QUOTES);
				}
			}
		}
		return $values;
	}

	private function id ($name = "") {

		if (is_empty($name))
			return false;

		if (is_empty($name = key_of("name",key_of($name,$this->element))))
			return false;

		return $this->identifier.$name;
	}

	private function markup ($name = "") {

		if (is_empty($name))
			return [];

		$element = key_of($name, $this->element);

		if (is_empty($element))
			return [];

		$markup = key_of("markup", $element);
		$class = key_of("class", $element);

		if (are_empty($markup,$class))
			return [];

		if (is_array($markup)) {

			if (is_empty($class))
				return $markup;

			$markup["class"] = ifttt(
				key_of("class",$markup),
				key_of("class",$markup)." ".$class,
				$class
			);

			return $markup;
		}

		$markup = "<markup ".$markup."></markup>";
		$markup = $this->core->html->parse($markup);

		if (not_empty($class))
			$markup = $this->core->html->property($markup, "class", $class);

		$markup = key_fetch("markup", $markup);

		return $markup;

	}

	private function name ($name = "") {

		if (is_empty($name))
			return false;

		if (is_empty($element = key_of($name, $this->element)))
			return false;

		$id = $this->id($name);

		if ($id === false)
			return false;

		if (in_array(key_of("type", $element), $this->config("array_fields")))
			return $id."[]";

		if (key_of("array", $element))
			return $id."[]";

		if (key_of("multiple", $element))
			return $id."[]";

		return $id;

	}


	// Deprecated

	public function gen_mail_text () {

		deprecated();

		$text ="";

		do {

			$entry = current($this->element);
			$entry['name'] = key($this->element);
			$entry["label"] = $this->core->json($entry["label"]);

			if (
				!in_array($entry['type'],array("button","submit"))
				&& !in_array($entry['name'],$this->setting["exlude_from_mail"])
				&& !empty($entry['value'])
				){

				if (is_array($entry['value'])){

					if (is_array($entry['value'][0])){
						$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	";
						foreach($entry['value'][0] as $entry_value_key){
							$text .=	$entry['value'][1][$entry_value_key]."\n";
						}
					}
					else{
						if (isset($entry['value'][1][$entry['value'][0]])){
							$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	";
							$text .=	$entry['value'][1][$entry['value'][0]]."\n";
						}
					}
				}
				else{
					$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	";
					$text .=	$entry['value']."\n";
				}
				$text .=	"\n";
			}
		}
		while(next($this->element));

		reset($this->element);
		return $text;
	}

	public function mail_text ($elements, $options = []){

		deprecated();

		$options = array_merge(
			array(
				"html" => false
			),
			$options
		);

		$text = "";

		if (empty($elements))
			return $text;

		foreach($elements as $key => $entry){

			$entry['name'] = $key;
			$entry['value'] = key_fetch($key,$this->values);
			$entry["label"] = $this->core->json($entry["label"]);

			if (
					!in_array($entry['type'],array("button","submit")) &&
					!in_array($entry['name'],$this->config("exclude_from_mail")) &&
					!empty($entry['value'])
				){
					if (is_array($elements[$key]['value'])){
						if (is_array($entry['value'])){
							$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	"."\n";
							foreach($entry['value'] as $entry_value){

								if (key_fetch($entry_value,$elements[$key]['value'][1]))
									$text .=	key_fetch($entry_value,$elements[$key]['value'][1]).br;

								else if (value_fetch($entry_value,$elements[$key]['value'][1]))
									$text .=	$entry_value.br;
							}
						}else{
							if (isset($elements[$key]['value'][1][$entry['value'][0]])){
								$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	";
								$text .=	$elements[$key]['value'][1][$entry['value'][0]]."\n";
							}
						}
					}else{
						$text .=	((!empty($entry['label']))? $entry['label'] : $entry['name']).":	";
						$text .=	$entry['value']."\n";
					}
					$text .=	"\n";
				}
			}

			if (!fetch("html",$options))
				$text = strip_tags($text);

			return $text;
	}

	public function send_mail () {

		deprecated();

		$mail = $this->core->mail_o();
		$this->config('mailto') ? $mail->add_recipient($this->config('mailto')) : $mail->add_recipient($this->core->config('form_email'));
		$mail->set_subject($this->config('subject'));
		$mail->add_body_txt($this->mail_text($this->element));

		return $mail->send_mail();
	}


}
