<?php


function any_empty () {

	foreach (func_get_args() as $arg)
		if (is_empty($arg))
			return true;

	return false;
}


function are_empty () {

	foreach (func_get_args() as $arg)
		if (not_empty($arg))
			return false;

	return true;
}


function are_equal ($arg1 = NULL, $arg2 = NULL) {

	if ($arg1 === NULL && $arg2 === NULL)
		return true;

	if (is_string($arg1) && is_string($arg2)) {

		if ($arg1 === $arg2)
			return true;

		return false;
	}

	if (is_bool($arg1) && is_bool($arg2)) {

		if ($arg1 === $arg2)
			return true;

		return false;
	}

	if (is_array($arg1) && is_array($arg2)) {

		if (serialize($arg1) === serialize($arg2))
			return true;

		return false;
	}

	if (is_int($arg1) && is_int($arg2)) {

		if ($arg1 === $arg2)
			return true;

		return false;
	}

	if (is_scalar($arg1) && is_scalar($arg2)) {

		if ($arg1."" === $arg2."")
			return true;

		return false;

	}

	return false;
}


function debug ($return = false){

	deprecated('debug() is deprecated. Use $core->debug(true) with debugme()');
	return;

}


function debugme () {

	global $core;

	if (empty($core) || !is_object($core))
		return;

	if ($core->debug() === false)
		return;

	call_user_func_array("printme", func_get_args());

	if (!is_int($core->debug()))
		return;

	$debug = debug_backtrace();
	printme(key_of($core->debug(),$debug));
	return;

}


function deprecated ($message = false, $backtrace = 1) {

	global $core;

	$backtrace = either(intval($backtrace), 1);
	$debug_backtrace = debug_backtrace();

	if (is_empty($message))
		$message = key_of("function",key_of($backtrace,$debug_backtrace))."() is deprecated";

	if (!is_object($core))
		return NULL;

	if (!isset($core->message) || !is_object($core->message))
		return NULL;

	$core->message->dev(
		"<b>".$message."</b><br>".
		"File: ".key_fetch("file",key_of($backtrace,$debug_backtrace))." ".
		"Line: ".key_fetch("line",key_of($backtrace,$debug_backtrace)),
		"Deprecated",
		1
	);
}


function dumpme () {

	$args = func_get_args();

	global $core;

	if (is_object($core) && is_object($core->user) && !$core->user->is_dev())
		return;

	foreach($args as $arg){
		echo "<pre>";
		ob_start();
		var_dump($arg);
		$print = ob_get_clean();
		echo htmlspecialchars($print,defined("ENT_SUBSTITUTE") ? ENT_SUBSTITUTE : NULL);
		echo "</pre>";
	}
}


function either () {
	$args = func_get_args();
	foreach($args as $arg)
		if(not_empty($arg))
			return $arg;

	if (count($args) > 1){
		return array_pop($args);
	}
	else
		return NULL;
}


function error_handler ($severity, $message, $file, $line) {
	throw new ErrorException($message, 0, $severity, $file, $line);
	return true;
}


function fetch ($needle = "", $haystack = []) {

	if (is_bool($needle)) {

		if (is_bool($haystack))
			return $needle === $haystack;

		if (is_array($haystack))
			return array_find($needle, $haystack, false, false);

	}

	else if (is_string($needle) || is_int($needle)) {

		if (is_empty($needle))
			return false;

		if (is_empty($haystack))
			return false;

		if (is_string($needle) && is_string($haystack) && strpos($haystack, $needle) !== false)
			return true;

		if (is_string($haystack) && are_equal($needle, $haystack))
			return true;

		if (is_string($haystack) && preg_match(preg_mod($needle), $haystack))
			return true;

		if (is_array($haystack) && not_empty($result = key_fetch($needle,$haystack)))
			return $result;

		if (is_array($haystack) && is_string($needle) && in_array($needle,$haystack,true))
			return true;

		else if (is_array($haystack) && is_numeric($needle) && in_array(intval($needle),$haystack,true))
			return true;

		else if (is_array($haystack) && is_numeric($needle) && in_array($needle."",$haystack,true))
			return true;

		if (is_array($haystack) && ($result = array_find($needle,$haystack)))
			return $result;

	}

	else if (is_array($needle)) {

		if (is_array($haystack) && ($result = array_find($needle,$haystack,"^","$")))
			return $result;

		else if (is_string($haystack)) {

			foreach ($needle as $entry) {

				if (fetch($entry, $haystack))
					return true;
			}

			return false;

		}
	}

	return false;

}


function key_fetch ($search = "", $config = array()){

	// Fail

	if(is_empty($search))
		return false;

	if(!is_array($config))
		return false;

	if (is_array($search))
		return fetch($search,$config);

	// Shortcut

	if(is_array($config) && array_key_exists($search,$config))
		return $config[$search];

	// Full search

	$return = array();

	foreach($config as $config_key => $config_value){
		if(is_array($config_value)) {
			if(!is_null($array = key_fetch($search,$config_value))){
				$return[$config_key] = $array;
			}
		}
	}

	if(is_empty(array_keys($return)))
		return NULL;

	else if(count($return) > 1)
		return $return;

	else
		return current($return);

	// return either($return,false);
}


function int ($int = false) {

	global $core;

	if (is_object($core) && isset($core->calc) && is_object($core->calc))
		return $core->calc->intify($int);

	if (!is_numeric($int))
		return NULL;

	if (is_int($int))
		return $int;

	if (is_float($int))
		return $int;

	if (strval(intval($int)) === $int)
		return intval($int);

	if (preg_match("/^".intval($int)."(\.|,)([0]+)$/",$int))
		return intval($int);

	if (strval(floatval($int)) === $int)
		return floatval($int);

	if (preg_match("/^".floatval($int)."([0]+)$/",$int))
		return floatval($int);

	return NULL;

}


function iftctt ($compare = NULL, $compare_with = NULL, $do = "", $else = NULL){

	if(are_equal($compare, $compare_with))
		return $do;

	return $else;
}


function ifttt ($case = false, $do = "", $else = NULL) {

	if(is_empty($case))
		return $else;

	return $do;
}


function in_list ($needle = "", $haystack = []) {

	if (!is_indexed((array) $haystack, true))
		return false;

	if (is_indexed($needle, true)) {

		foreach ($needle as $value) {
			if (in_list($value, $haystack))
				return true;
		}

		return false;

	}

	foreach ((array) $haystack as $compare)
		if (are_equal($needle, $compare))
			return true;

	return false;
}


function is_empty ($var = "") {

	 //Alternative to PHPs empty function without including Zeros

	 if (is_null($var))
	 	return true;

	 if (is_string($var) && $var === "")
	 	return true;

	if (is_bool($var))
		return !$var;

	if (is_numeric($var))
		return false;

 	if (is_array($var)) {

 		if (empty($var))
 			return true;

 		if (!is_indexed($var))
 			return false;

 		foreach ($var as $key => $entry) {
 			if (not_empty($entry))
 				return false;
 		}

 		return true;

 	}

	// return false;
	return empty($var);

 	// return !empty($var) || $var === '0' || $var === 0 /*|| $var === false*/;
}


function has_regex ($string = "", $must_have_delimiters = false) {

	if (is_empty($string))
		return false;

	if (is_string($string)) {

		// Test for simple strings

		if (is_numeric($string))
			return false;

		if (preg_match("/^[\w\s:@,._\-]+$/",$string))
			return false;

		if (preg_match("/^[\w\s()]+$/",$string))
			return false;

		// Test for SQL strings

		if (preg_match("/^%[\w\s:,._\-]+%$/",$string))
			return false;

		// Test for dates

		if (preg_match("/\d\-\d/",$string) && !preg_match("/[\[\(]/",$string))
			return false;

		// Test for meaningfulness

		if (!preg_match("/[\+\-\|\*\[\]\{\}\(\)]+/",$string))
			return false;


		// Test for delimiters

		if (!preg_match("/^([\W\D\S])(.+)(\1)([a-zA_Z]*)$/", $string)) {

			if ($must_have_delimiters)
				return false;

			$string = preg_mod($string);
		}

		// Test execution

		$reporting_level = error_reporting(0);
		$case = preg_match($string,"");
		error_reporting($reporting_level);

		if ($case === false)
			return false;

		return true;
	}

	if (is_array($string)) {

		foreach ($array = $string as $string) {

			if (has_regex($string))
				return true;

		}

		return false;

	}

	return false;

}


function key_of ($key = "", $array = []) {

	if (empty($array) || !is_array($array))
		return NULL;

	if (is_string($key) || is_int($key)) {

		if (array_key_exists($key, $array))
			return $array[$key];

	}

	else if (is_indexed($key)) {

		foreach ($key as $value) {

			if (neither(is_string($value), is_int($value)))
				continue;

			if (array_key_exists($value, $array))
				return $array[$value];

		}

	}

	return NULL;

}


function keys_of ($key = array(), $array = array()) {

	$return = array();

	if (is_string($key))
		$key = array($key);

	if (!is_array($key))
		return NULL;

	foreach ($key as $value) {
		if (is_scalar($value) && ($entry = key_of($value, $array)) !== NULL)
			$return[$value] = $entry;
		else if (is_array($value))
			$return = array_merge(
				$return,
				(array) keys_of($value, $array)
			);
	}

	return $return;

}


function matches($needle = "", $haystack = "") {

	if (any_empty(is_string($needle), is_string($haystack)))
		return false;

	if (are_empty($needle, $haystack))
		return true;

	if (is_empty($needle))
		return false;

	// if (strstr($haystack, $needle) !== false)
	if (stristr($haystack, $needle) !== false)
		return true;

	return false;

}


function match_any ($needles = [], $haystack = []) {

	if (are_empty($needles))
		return true;

	foreach ((array) $haystack as $hay) {
		foreach ((array) $needles as $needle) {
			if (matches($needle, $hay))
				return true;
		}
	}

	return false;

}


function microtimes () {

	static $microtime_t2 = false;

	$t1 = microtime(true);

	if ($microtime_t2 === FALSE)
		$microtime_t2 = $t1;

	$return = $t1 - $microtime_t2;
	$microtime_t2 = $t1;

	return $return;
}


function neither () {

	$args = func_get_args();
	foreach($args as $arg)
		if(not_empty($arg))
			return false;

	return true;
}


function not_empty ($var = "") {

	 return !is_empty($var);

}


function preg_mod ($string = "", $contains_regex = true) {

	if (is_array($string)) {

		$return = array();

		foreach ($string as $value) {
			$return[] = preg_mod($value);
		}

		return $return;

	}

	$delimiter = array("/","-","#","~","@","!","<",">");

	$return = false;

	foreach ($delimiter as $k => $d) {

		$k += 1;
		$k %= count($delimiter);

		if (!preg_match($delimiter[$k].$d.$delimiter[$k],$string)) {

			if ($contains_regex)
				$return = $d.$string.$d;

			else
				$return = $d.preg_quote($string).$d;
			break;
		}
	}

	if ($return) {

		$reporting_level = error_reporting(0);
		$case = preg_match($return,"");
		error_reporting($reporting_level);


		if($case !== false){
			return $return;
		}
		else{
			$return = preg_quote($string);
			$return = preg_mod($return);
			return $return;
		}
	}

	// Error
	error_reporting(E_ALL);
	echo "<br>";
	echo $_SERVER['PHP_SELF'];
	echo "<br>";
	echo "Missing delimiter in :".$string;
	exit();
}


function preg_replace_array ($pattern,$replacement,$array) {

	if (is_string($array))
		return preg_replace($pattern, $replacement, $array);

	if (!is_array($array))
		return $array;

	foreach ($array as $array_key => $array_value)
		$array[$array_key] = preg_replace_array($pattern, $replacement, $array_value);

	return $array;
}


function printme () {

	$args = func_get_args();

	global $core;

	if (is_object($core) && isset($core->user) && is_object($core->user) && !$core->user->is_dev())
		return;

	if (
		is_object($core)
		&& isset($core->user)
		&& is_object($core->user)
		&& not_empty($debug_groups = (array) $core->config("debug_groups"))
		&& !in_list($core->user->ID(), $debug_groups)
		)
		return;

	foreach ($args as $arg) {

		echo "<pre>\n";

		if (!isset($core->array))
			$print = print_r($arg, true);

		else
			$print = $core->array->render($arg);

		echo htmlspecialchars($print,defined("ENT_SUBSTITUTE") ? ENT_SUBSTITUTE : NULL);
		echo "</pre>\n";
	}
}


function random ($length=8,$special_chars=false) {

	$dummy = $special_chars == true ? array_merge(range('0', '9'), range('a', 'z'), range('A', 'Z'), array('#','&','@','$','_','%','?','+')) : array_merge(range('0', '9'), range('a', 'z'), range('A', 'Z'));

	// shuffle array

	mt_srand((double)microtime()*1000000);

	for ($i = 1; $i <= (count($dummy)*2); $i++) {
		$swap = mt_rand(0,count($dummy)-1);
		$tmp = $dummy[$swap];
		$dummy[$swap] = $dummy[0];
		$dummy[0] = $tmp;
	}
	// Return
	return substr(implode('',$dummy),0,$length);
}


function remove_keys_of ($key = array(), $array = array()) {

	if (!is_array($array))
		return array();

	if (is_empty(array_keys($array)))
		return $array;

	if (is_string($key) || is_numeric($key))
		$key = array($key);

	if (!is_array($key))
		return $array;

	foreach ($key as $value)
		if (array_key_exists($value, $array))
			unset($array[$value]);

	return $array;

}


function runtime ($options = array()) {

	if ($options === true)
		$options = array(
			"info" => true
		);

	$options = array_merge(
		array(
			"raw" => false,
			"info" => false,
			"decimals" => 2,
		),$options
	);


	static $runtime = false;

	if (!$runtime)
		$runtime = microtime(true);

	$time = microtime(true)-$runtime;

	if (key_of("raw",$options))
		return $time;

	else if (!key_of("info",$options))
		return number_format($time, key_of("decimals",$options));

	$debug = debug_backtrace();

	return number_format($time, key_of("decimals",$options)).":".key_of("file",$debug[0]).":".key_of("line",$debug[0]);

}


function toggle ($case = "", $toggle = array()) {

	if (any_empty($case, $toggle, is_indexed($toggle, true), count($toggle) == 2))
		return $case;

	if (are_empty(is_string($case), is_bool($case), is_numeric($case)))
		return $case;

	if (!in_array($case, $toggle, true))
		return $case;

	$key = array_search($case, $toggle, true);

	$key = 1-int($key);

	return key_of($key, $toggle);

}


function UID ($input = false) {

	if (any_empty($input, is_string($input), is_numeric($input)))
		$input = random();


	$UID = time() . runtime(["decimals"=>10]) . $input;
	$UID = md5($UID);

	return $UID;
}


function value_fetch ($needle = "", $haystack = array()){

	if(is_array($haystack) && in_array($needle, $haystack, true))
		return true;

	else if(is_string($haystack) && are_equal($needle, $haystack))
		return true;

	if (not_empty($needle) && not_empty(array_find($needle, $haystack)))
		return true;

	return false;

}



// Should get deprecated

function add_slashes($string = "") {

	if (is_object($string))
		return $string;

	if (is_int($string))
		return $string;

	if (is_bool($string))
		return $string;

	if (is_null($string))
		return $string;

	if (is_string($string))
		return addslashes($string);

	if (is_array($string)) {

		foreach ($string as $key => $value){

			unset($string[$key]);

			$key = add_slashes($key);
			$value = add_slashes($value);

			$string[$key] = $value;

		}
		return $string;
	}
}

function are_indexed() {

	$arrays = func_get_args();

	if (is_empty($arrays))
	return false;

	foreach ($arrays as $array)
	if (!is_indexed($array, true))
	return false;

	return true;
}

function array_multimerge ($array1, $array2) {

	// deprecated();

	if (!is_array($array1) && is_array($array2))
		return $array2;

	if (!is_array($array2) && is_array($array1))
		return $array1;

	if (!is_array($array1) && !is_array($array2))
		return array();

	if (are_empty($array1, $array2))
		return array();

	if (is_empty($array2))
		return $array1;

	if (is_empty($array1))
		return $array2;

	foreach ($array2 as $k => $v) {
		if (is_array($v) && array_key_exists($k, $array1)) {
			$array1[$k] = array_multimerge($array1[$k], $v);
		}
		else {
			$array1[$k] = $v;
		}
	}
	return $array1;
}

function array_shuffle($array) {

	deprecated();

	if(is_empty($array) || !is_array($array))
	return $array;

	$keys = array_keys($array);
	shuffle($keys);

	foreach ($keys as $key)
	$return[$key] = $array[$key];

	return $return;
}

function array_find ($needle, $haystack, $d1="^", $d2="$", $d3="i", $key = false) {

	/*
	Find all occurrences in an array
	Supports regex
	Needle can be
	"value"
	["key"=>"value"]
	[["key"=>"value"],["key"=>"value"]]
	["key"=>"value","key"=>"value"]
	["key"=>["value","value"]]
	*/

	$regex = true;

	if (is_empty($haystack) || !is_array($haystack))
	return array();

	if (is_empty($needle))
	return $haystack;

	if (is_object($haystack))
	return $haystack;

	// Case when regex is off

	if ($d1 === false)
	$regex = false;

	$return = array();

	// Multiple needles
	// Create union
	if (
	is_array($needle) &&
	is_array(current($needle)) &&
	is_indexed($needle)
	){

		$return = array();

		// $key might be missing here
		foreach($needle as $current_find)
		$return = $return+array_find($current_find,$haystack,$d1,$d2,$d3,$key); //:

		return $return;
	}


	// Multiple conditions
	// Create intersection
	else if (
	is_array($needle) &&
	count(array_keys($needle)) > 1
	){

		foreach ($needle as $current_find_key => $current_find)
		$haystack =	array_find(array($current_find_key => $current_find), $haystack, $d1, $d2, $d3, $key);

		return $haystack;
	}

	else {

		// Needle is string
		if (is_string($needle) || is_int($needle)) {

			foreach ($haystack as $haystack_key => $haystack_value) {

				if (is_string($haystack_value) && ($regex ? preg_match(preg_mod($d1.$needle.$d2).$d3,$haystack_value) : are_equal($needle, $haystack_value)) )
				$return[$haystack_key] = $haystack_value;

				//	Walk multidimisional arrays
				if (is_array($haystack_value) && not_empty($return_value = array_find($needle,$haystack_value,$d1,$d2,$d3,$haystack_key))) {
					$return[$haystack_key] = $haystack_value;
					// $return[$haystack_key] = $return_value; // Didnt work with global search
				}
			}
			return $return;
		}

		// Needle is array
		else if (is_array($needle)) {

			$needle_key = key($needle);
			$needle_value = current($needle);

			foreach ($haystack as $haystack_key => $haystack_value) {

				// Associated find_keys

				if (either(is_numeric($haystack_value), is_string($haystack_value),is_null($haystack_value),is_bool($haystack_value)) && are_equal($needle_key,$haystack_key)) {

					// Indexed $needle_value

					if (is_indexed($needle_value) && is_string($haystack_value) &&  $needle_key === $haystack_key) {

						if (in_array($haystack_value, $needle_value)) {
							if ($key !== false)
							return $haystack;
							else
							$return[$haystack_key] = $haystack_value;
						}
					}

					else if (is_array($needle_value))
					continue;

					else if
					(
					are_equal($needle_value,$haystack_value) ||
					($regex && not_empty($haystack_value) && preg_match(preg_mod($d1.$needle_value.$d2).$d3,$haystack_value))
					) {

						if($key !== false)
						return $haystack;

						else
						$return[$haystack_key] = $haystack_value;

					}


					else if ($needle_key === $haystack_key && $needle_value === $haystack_value) {

						if ($key !== false)
						return $haystack;

						else
						$return[$haystack_key] = $haystack_value;
					}
					break;
				}

				// Indexed find_keys
				else if (is_int($needle_key) || are_equal($needle_key,intval($needle_key))) {

					if (is_string($haystack_value) && ($regex ? preg_match(preg_mod($d1.$needle_value.$d2).$d3,$haystack_value) : are_equal($needle_value, $haystack_value)) ) {

						if($key)
						$return[$key] = $haystack;

						else
						$return[$haystack_key] = $haystack_value;
					}
				}

				//	Walk multidimisional arrays
				else if (is_array($haystack_value)) {

					if
					(
					$needle_key === $haystack_key &&
					is_array($needle_value) &&
					not_empty($return_value = array_find($needle_value,$haystack_value,$d1,$d2,$d3,$haystack_key))
					) {

						if($key)
						return $haystack;

						else
						$return[$haystack_key] = $haystack_value;
					}

					else if (not_empty($return_value = array_find($needle,$haystack_value,$d1,$d2,$d3,$haystack_key))) {
						$return[$haystack_key] = $return_value;
					}
				}
			}
			return $return;
		}
	}
}

function array_match($needle = array(), $haystack = array()) {

	if (any_empty($needle, $haystack, is_array($haystack)))
	return false;

	// $needle = (array) $needle;

	if (is_array($needle) && count($needle) > 1) {
		foreach ($needle as $key => $value)
		if (!array_match(array($key => $value), $haystack))
		return false;

		return true;
	}

	else {

		foreach ($haystack as $hay_key => $hay_value) {

			if (is_array($needle) && are_equal(key($needle), $hay_key) && are_equal(current($needle), $hay_value))
			return true;

			if (!is_array($needle) && are_equal($needle, $hay_value))
			return true;

			if (is_array($hay_value))
			if (array_match($needle, $hay_value))
			return true;

		}

		return false;

	}
}

function db_encode($string) {
	$string = url_encode($string);
	$string = preg_replace("/-/","_",$string);
	return $string;
}

function downsample($img, $size = "1000", $folder = "") {

	global $core;


	if (!is_object($core))
		return $img;

	if (!$core->array->match(
		$core->file->extension($img),
		array("jpg", "jpeg", "png", "pdf", "mp4", "mov")
	))
		return $img;

	if ($core->file->exists($img))
		$source = $core->file->get_path($img);

	else if ($core->file->exists($core->folder->path($core->path("root", "files"),$core->folder->part($img)).$core->file->name($img)))
		$source = $core->folder->path($core->path("root", "files"),$core->folder->part($img)).$core->file->name($img);

	else if ($core->file->exists($core->folder->path($core->path("root", "kitbooth"),$core->folder->part($img)).$core->file->name($img)))
		$source = $core->folder->path($core->path("root", "kitbooth"),$core->folder->part($img)).$core->file->name($img);

	else
		return $img;


	$file_name = $core->file->name($img);

	$target_folder = $core->filesystem->path($core->folder->part($img),$folder);
	$source_folder = $core->folder->part($source);
	$target = $core->filesystem->path($core->folder->part($source),$folder).$file_name;
	$filemtime = $core->file->age($source, array("age"=>"modified"));

	// PDF

	if (are_equal(
		$core->file->extension($file_name),
		"pdf"
	)) {

		// Check if downsample allready exists

		$target = preg_replace("/pdf$/i","jpg",$target);

		if (
			$core->file->exists($target) &&
			$filemtime == $core->file->age($target, array("age"=>"modified"))
		)
			return "/".$target_folder.$core->file->name($target)."?version=".$filemtime;

		$file_name = $core->pdf->image($source, $folder, array(
			"pages" => 0,
			"size" => $size,
			"modified" => $filemtime,
		));

		if ($file_name)
			return "/".$target_folder.$file_name."?version=".$filemtime;

		else
			return $img;

	}

	// Video

	else if ($core->array->match(
		$core->file->extension($file_name),
		array("mp4", "mov")
	)) {

		// Check if downsample allready exists

		// $target = preg_replace("/".$core->file->extension($file_name)."$/i","jpg",$target);
		$target = $core->folder->part($target).$core->file->basename($target).".jpg";

		if (
			$core->file->exists($target) &&
			$filemtime <= $core->file->age($target, array("age"=>"modified"))
		)
			return "/".$target_folder.$core->file->name($target)."?version=".$filemtime;

		$file_name = $core->video->image($source, $folder, array(
			"size" => $size,
			"modified" => $filemtime,
		));

		if ($file_name)
			return "/".$target_folder.$file_name."?version=".$filemtime;

		else
			return $img;

	}

	else if ($core->array->match(
	$core->file->extension($file_name),
	array("jpg", "jpeg", "png", "gif", "bmp")
	)) {

		// Check if downsample allready exists

		if (
		$core->file->exists($target) &&
		$filemtime <= $core->file->age($target, array("age"=>"modified"))
		)
		return "/".$target_folder.$core->file->name($target)."?version=".$filemtime;

		$core->folder->create($core->folder->part($target));

		list($width_orig, $height_orig) = getimagesize($source);

		$width = $size;
		$height = $size;

		$ratio_orig = $width_orig/$height_orig;

		if ($width/$height > $ratio_orig)
		$width = $height*$ratio_orig;

		else
		$height = $width/$ratio_orig;

		if ($width_orig <= $size && $height_orig <= $size)
		return $img;

		// Resample

		if (preg_match("/.jpg/i",$file_name) || preg_match("/.jpeg/i",$file_name)) {

			if (!matches("image/jpeg", mime_content_type($source)))
			return $img;

			if (!function_exists("imagecreatefromjpeg"))
			return $img;

			$image = imagecreatefromjpeg($source);

			if (function_exists('exif_read_data')) {

				$exif = exif_read_data($source,'EXIF');

				if (not_empty($orientation = key_of("Orientation", $exif))) {
					$degree = array("1"=>"0","3"=>"180","6"=>"270","8"=>"90");
					$degree = key_of($orientation, $degree);

					if ($degree)
					$image = imagerotate($image, $degree, 0);

					if (value_fetch($orientation, array("6","8"))) {
						list($width,$height) = array($height,$width);
						list($width_orig,$height_orig) = array($height_orig,$width_orig);
					}

				}

			}

			$image_p = imagecreatetruecolor($width, $height);

			imagecopyresampled($image_p, $image, 0, 0, 0, 0, $width, $height, $width_orig, $height_orig);
			imagejpeg($image_p, $target, 90);
		}

		else if (preg_match("/.png/",$file_name)) {

			$image_p = imagecreatetruecolor($width, $height);

			if (!matches("image/png", mime_content_type($source)))
			return $img;

			$image = imagecreatefrompng($source);
			imageAlphaBlending($image_p, false);
			imagesavealpha( $image_p, true );
			imagecopyresampled($image_p, $image, 0, 0, 0, 0, $width, $height, $width_orig, $height_orig);
			imagepng($image_p, $target, 9);
		}

		else if (preg_match("/.gif/",$file_name)) {

			if (!matches("image/gif", mime_content_type($source)))
			return $img;

			$image = imagecreatefromgif($source);
			imagecolortransparent($image_p);
			imagecopyresampled($image_p, $image, 0, 0, 0, 0, $width, $height, $width_orig, $height_orig);
			imagegif($image_p, $target);
		}

		else if (preg_match("/.bmp/",$file_name)) {

			if (!matches("image/bmp", mime_content_type($source)))
			return $img;

			if (!is_callable("imagecreatefrombmp") || !is_callable("imagebmp"))
			return $img;

			$image = imagecreatefrombmp($source);
			imagecopyresampled($image_p, $image, 0, 0, 0, 0, $width, $height, $width_orig, $height_orig);
			imagebmp($image_p, $target);
		}

		touch($target,$filemtime);
		return "/".$target_folder.$file_name."?version=".$filemtime;
	}

	return $img;
}

function encode_html($value) {
	return htmlentities($value, ENT_COMPAT, 'UTF-8');
}

function group_array($array = false, $group = false) {

	if(empty($array))
	return $array;

	if(empty($group))
	return $array;

	foreach($array as $entry){

	}

	return $array;
}

function is_email($mail) {

	if (!is_string($mail))
		return false;

	$r = '/^((\"[^\"\f\n\r\t\v\b]+\")|([\w\!\#\$\%\&\'\*\+\-\~\/\^\`\|\{\}\=\?]+(\.[\w\!\#\$\%\&\'\*\+\-\~\/\^\`\|\{\}\=\?]+)*))@((\[(((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9])))\])|(((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9]))\.((25[0-5])|(2[0-4][0-9])|([0-1]?[0-9]?[0-9])))|((([A-Za-z0-9\-])+\.)+[A-Za-z\-]+))$/D';
	return (bool) preg_match($r, trim($mail));
}

function is_indexed($array = false, $must_be_in_order = false) {

	if (empty($array) || !is_array($array))
		return false;

	if (array_keys($array) === range(0, count($array) - 1))
		return true;

	if ($must_be_in_order === true)
		return false;

	foreach ($array as $key => $value) {
		if (is_int($key))
			continue;

		return false;
	}

	return true;
}

function kb_date($output="d.m.Y",$input="NOW",$locale=false) {

	// UNIX time
	if ($input && preg_match("/^[0-9]+$/",$input))
		$input = "@".$input;

	if ($locale) {
		setlocale(LC_TIME, $locale);
		return utf8_encode(strftime($output, strtotime($input)));
	}

	else {
		return date($output, strtotime(either($input,"NOW")));

	}

}

function kb_escapeshellarg($string = "") {

	// Replaces escapeshellarg()

	if (any_empty(
		$string,
		either(is_string($string), is_int($string))
		))
		return "";

	// Return string untouched after first escape
	if (preg_match("/^'([^']|\\')+'$/",$string))
		return $string;

	// Start escape
	// Remove single quote around string
	if (preg_match("/^(')(.+)(')$/", $string))
		$string = preg_replace("/^(')(.+)(')$/", "$2", $string);

	// Escape additional single quotes
	$string = preg_replace("/'/", "'\"'\"'", $string);

	// Add single quotes around string
	$string = "'".$string."'";

	return $string;

}

function loaded ($name = "", $message = true) {

	if (is_empty($name))
	return false;

	if (extension_loaded($name))
	return true;

	$shell = array(
	"ghostscript" => "gs --version",
	"ffmpeg" => "ffmpeg -version"
	);

	if ($command = key_of($name, $shell))
	if(shell_exec($command))
	return true;

	if ($message) {
		global $core;
		$core->message->dev("$name requested but not loaded","Modules",	2);
	}
	return false;

}

function preg_match_any($needles = [], $haystack = []) {

	if (is_empty($needles))
	return true;

	foreach ((array) $haystack as $hay) {
		foreach ((array) $needles as $needle) {
			if (preg_match(preg_mod($needle), $hay))
			return true;
		}
	}

	return false;

}

function preg_match_array($value,$array) {

	// deprecated();


	foreach($array as $array_value){
		// delimiter
		if(!preg_match('/[a-zA-Z0-9]/',substr($array_value,0,1)) && $array_value[0] === $array_value[strlen($array_value)-1]) {
			if(preg_match($array_value,$value)) {
	return true;
			}
		}
		else {
			if($array_value == $value) return true;
		}
	}
	return false;
}

function preg_replace_all_reverse($matches=array()) {

	// FAIL
	if(empty($matches))
	return $matches;

	if(!isset($matches[0][0]))
	return $matches;

	// PREPARE
	$entries = count($matches[0]);
	$found_matches = count($matches);

	// EXECUTE
	for ($i = 0; $i < $entries; $i++){
		$new_matches[$i] = array();
		for ($j = 0; $j < $found_matches; $j++){
			$new_matches[$i][$j] = $matches[$j][$i];
		}
	}

	// RETURN
	return $new_matches;
}

function rsort_by_length($a,$b) {

	return (strlen($a) - strlen($b))*(-1);

}

function sort_array($array=array(), $sort = false, $flag="") {

	if(!$sort)
	return $array;

	if(!is_array($sort))
	$sort = array($sort);

	$flag = $flag == "DESC" ? SORT_DESC : "";


	foreach ($array as $key => $row) {
		$new_array[] = isset($row[$sort]) ? $row[$sort] : "";
	}

	if(!empty($flag) && !empty($new_array)){
		array_multisort($new_array,$flag,$array, SORT_NATURAL);
	}else if(!empty($new_array)){
		array_multisort($new_array,$array, SORT_NATURAL);
	}

	return $array;
}

function sort_by_length($a,$b) {

	return (strlen($a) - strlen($b));

}

function string_to_array($string, $default_key = 0) {
	if(is_array($string))
	return $string;

	$string = explode(",",$string);
	foreach($string as $value){
		$parts = explode("=",$value);
		if(isset($parts[1]))
		$array[$parts[0]]	= $parts[1];
		else
		$array[$default_key++]= $parts[0];
	}
	return $array;
}

function thumbnail($file, $size = "1000", $folder = ".cache/thumbs/") {

	global $core;

	if (!is_object($core))
	return $file;

	if ($core->array->match(
	$core->file->extension($file),
	array("jpg", "jpeg", "png", "pdf", "mp4", "mov")
	)) {

		$folder = $folder.$size;
		return downsample($file, $size, $folder);

	}

	$filetype = $core->file->type($file);
	$filetype = explode("/",$filetype);
	$filetype = key_of(0, $filetype);

	if ($icon = $core->template->icon->src("document_$filetype"))
	return $icon;

	if ($icon = $core->template->icon->src("document"))
	return $icon;

	return $file;

}

function to_html($input) {

	deprecated();

	if(is_string($input)){

		return preg_replace(
		array("-\n-","-\r-"),
		array("<br>",""),
		$input
		);

	}else if(is_array($input)){

		global $core;

		return $core->html->render($input);
	}
}

function url_encode($url, $true_encode = false) {

	if ($true_encode)
		return preg_replace(
			array(
				"/\s/",
				"/\?/",
			),
			array(
				"%20",
				"%253F",
			),
			$url
		);

	//Lowercase
	//$url = strtolower($url);
	// replace spaces by "-"
	// replace specials chars manually
	$special_chars		= array("/ä/",	"/Ä/",	"/ö/",	"/Ö/",	"/ü/",	"/Ü/",	"/ß/",	"/ẞ/",	"/ /");
	$chars_replacement	= array("ae",	"Ae",	"oe",	"Oe",	"ue",	"Ue",	"ss",	"SS",	"-");
	$url = preg_replace($special_chars,$chars_replacement,$url);

	$url = strtolower($url);

	// Use iconv function if exists to convert textcoding automatically
	if(function_exists('iconv')) {
		global $core;
		// setlocale(LC_CTYPE, $core->get_config('localization').'.utf8');
		$url = iconv('UTF-8', 'ASCII//TRANSLIT', $url);
	}
	// Else, manual conversion via html entities
	else {
		// convert accents to html entities
		$url = htmlentities($url, ENT_COMPAT, 'UTF-8');

		// remove the accent from the letter
		$url = preg_replace(array('/&([a-zA-Z]){1,2}(acute|grave|circ|tilde|uml|ring|elig|zlig|slash|cedil|strok|lig){1};/', '/&(euro){1};/'), array('${1}', 'EUR'), $url);
	}

	// 1:Everything but alphanumeric and -_ can be removed 
	// 2: remove double dashes
	// 3-4: remove dashes on start and end of string
	$url = preg_replace(array('/[^a-zA-Z0-9\-_]/', '/[\-]{2,}/','/^(-)(.*)$/','/^(.*)(-)$/'), array('', '-','$2','$1'), html_entity_decode($url));


	return $url;
}



// Backwards Compatibility

if (!function_exists('preg_replace_callback_array')) {

	function preg_replace_callback_array (array $patterns_and_callbacks, $subject, $limit=-1, &$count=NULL) {
		$count = 0;

		foreach ($patterns_and_callbacks as $pattern => &$callback) {
			$subject = preg_replace_callback($pattern, $callback, $subject, $limit, $partial_count);
			$count += $partial_count;
		}
		return preg_last_error() == PREG_NO_ERROR ? $subject : NULL;
	}

}



// Deprecated

function add_class($class=false,$string=false) {
	//Add a CSS class to a string
	deprecated();

	if(!$class && !$string)
	return false;

	//add class property
	if(!preg_match('-class=".*"-U',$string))
	$string .= ' class=""';

	//add class if new
	if(!preg_match(preg_mod('class=".*'.$class.'.*"').'U',$string))
	return preg_replace(preg_mod('(class=".*)(")').'U',"$1 ".$class."$2",$string);
}

function file_content($path) {

	deprecated('file_content() is deprecated, use $core->file->read() instead');

	if(preg_match("-^//-",$path)) {
		if(fetch("80",fetch("SERVER_PORT",$_SERVER)))
		$path = "http:".$path;
		else
		$path = "https:".$path;
	}

	set_error_handler("error_handler");

	try {
		$file_content = file_get_contents($path);
	}

	catch (Exception $e) {
		$file_content = false;
	}

	restore_error_handler();

	return $file_content;
}

function generatePW($length=8,$special_chars=false) {
	deprecated("Use random() instead of generatePW()");
	return random($length, $special_chars);
}

function kb_json_encode($value) {

	deprecated();

	//Alternative to PHPs json_encode Function with proper Encoding

	if(is_int($value) || is_string($value))
	return $value;

	$value = add_slashes($value);
	$value = json_encode($value);
	$value = preg_replace_callback(
		'/\\\\u(\w{4})/',
		function ($matches) {
			return html_entity_decode('&#x' . $matches[1] . ';', ENT_COMPAT, 'UTF-8');
		},
		$value
	);
	return $value;
}

function kb_json_decode($value) {

	deprecated();

	// No JSON

	if(is_array($value))
		return $value;

	if(!preg_match("-([\{\[]+)-",$value)){
		return $value;
	}

	$value_backup = $value;

	// Convert newlines within strings
	$value = preg_replace_callback("-\"([^\"]*)\"-s", function($matches){
		$matches[0] = preg_replace(array("-\r-","-\n-","-\t-"),array("\\r","\\n","\\t"),$matches[0]);
		return $matches[0];
	},$value);

	// JSON
	if (!is_null($array = json_decode($value,true)))
		return preg_replace_array("-\\\\\"-","\"",$array);
		// return $array;

	// HTML special chars
	if (matches("&quot;", $value) && !matches('"', $value))
		$value = htmlspecialchars_decode($value);

	// HJSON
	// Remove line breaks
	$value = preg_replace('/^(.+)"(\s*)$/m','\1",',$value);
	// Remove comments
	$value = preg_replace('-^(\s*)(//|#)(.*)$-m','',$value);
	$value = preg_replace('-([\t])*/\*([^*]*)\*/([\n\r])*-','',$value);
	//Add missing commas
	$value = preg_replace('/([\w}])([\r\n]*)$/m','\1,',$value);
	// Add missing quotes
	$value = preg_replace('/^([{\r\n\s\t]+)(\w+)(\s*)(:)/m','\1"\2"\3\4',$value);
	$value = preg_replace('/(\"\s*:)(\s*)(?!null|true|false)(\w[\w\s$=]*)(\s*)([},])/','\1\2"\3"\4\5',$value);
	// Remove trailing commas
	$value = preg_replace('/,([\s]*)([}\]]){1}/','\1\2',$value);
	$value = preg_replace('/,$/','',$value);

	if (!is_null($array = json_decode($value,true)))
		return preg_replace_array("-\\\\\"-","\"",$array);

	// return $value;
	return $value_backup;


}

function lowercase($input = "") {

	deprecated();

	if (is_empty($input))
	return $input;

	if (is_string($input))
		return mb_strtolower($input);

	if (is_array($input)) {

		foreach ($input as $key => $value) {
			$input[$key] = lowercase($value);
		}

		return $input;

	}

	return $input;

}

function markdown_decode($text = "", $options = array()) {

	deprecated();

	if (is_array($texts = $text)) {

		$return = array();

		foreach ($texts as $key => $text) {
			$return[$key] = markdown_decode($text, $options);
		}

		return $return;

	}

	global $core;

	if (is_object($core) && is_object($core->array))
		$options = $core->array->merge(
			array(
				"shift" => 0
			),
			$options
		);
	else
		$options = array_merge(
			array(
				"shift" => 0
			),
			$options
		);

	$replacement = array(
		'/^######(.*)$/m'
		=> '<h6>\1</h6>',

		'/^#####(.*)$/m'
		=> '<h5>\1</h5>',

		'/^####(.*)$/m'
		=> '<h4>\1</h4>',

		'/^###(.*)$/m'
		=> '<h3>\1</h3>',

		'/^##(.*)$/m'
		=> '<h2>\1</h2>',

		'/^#(.*)$/m'
		=> '<h1>\1</h1>',

		'/^([ \t]+)(-|\*)(\s)(.+)$/m'
		=> '<ul><ul><li>\4</li></ul></ul>',

		'/^(-|\*)(\s)(.+)$/m'
		=> '<ul><li>\3</li></ul>',

		'/<\/ul><\/ul>([\r\n]+)<ul><ul>/m'
		=> '\1',

		'/<\/ul>([\r\n]+)<ul>/m'
		=> '\1',

		'/^([0-9]+\.)(\s+)(.+)$/m'
		=> '<ol><li>\3</li></ol>',

		'/<\/ol>([\r\n]+)<ol>/m'
		=> '\1',

		'/^(> )([^<\r\n]{1})(.+)$/m'
		=> '<blockquote><p>\2\3</p></blockquote>',

		'/^([^<\r\n]{1})(.+)$/m'
		=> '<p>\1\2</p>',

		'/([^\s]+)(\s?)<\/p>(\s?)<p>(\s?)([\w\[+]+)/'
		=> '\1<br>\5',

		'/<p>([\s]*)<\/p>/'
		=> '',

		'/(\[)([^\]]+)(\])(\()([^)]+)(\))/'
		=> '<a href="\5">\2</a>',

		'/(\[)([\w.-]+@[\w.-]+\.\w+)(\])/'
		=> '<a href="mailto:\2">\2</a>',

		'/(\[)(http)([^\]]+)(\])/'
		=> '<a href="\2\3">\2\3</a>',

		'/\*\*([^\n]+)\*\*/m'
		=> '<strong>\1</strong>',

		'/\*([^\n*>]+)\*/m'
		=> '<em>\1</em>',

		'/^_([^\n_>]+)_/m'
		=> '<em>\1</em>',

		'/\s_([^\n_>]+)_/m'
		=> '<em>\1</em>',

		'/(\d)\\\(\.)/m'
		=> '\1\2',
	);

	$pattern = array_keys($replacement);

	$replacement = array_values($replacement);

	// Shift headlines
	if (key_of("shift", $options)) {
		for ($i = 6; $i > 0; $i--) {
			$key = $i - intval(key_of("shift", $options));
			if ($key < 0)
			$key = 0;
			$replacement[$i] = $replacement[$key];
		}
	}

	$text = preg_replace($pattern,$replacement,$text);

	return $text;

}

function uppercase($input = "") {

	deprecated();

	if (is_empty($input))
	return $input;

	if (is_string($input))
	return mb_strtoupper($input);

	if (is_array($input)) {

		foreach ($input as $key => $value) {
			$input[$key] = uppercase($value);
		}

		return $input;

	}

	return $input;

}