<?php

class KB_kit extends KB_class {

	public
		$caller,
		$current,
		$toolkits = [],
		$data,
		$id,
		$kit_options = false,
		$title;

	private
		$core,
		$correlation_o,
		$dir_exclude,
		$filesystem;

	function __construct($core) {

		$this->core = $core;
		$this->correlation_o =  $this->core->correlation_o();
    $this->dir_exclude = array(".", "..", ".svn", "_notes","kb_kits");
		$this->data = $this->object();
		$this->install_shipped();
		$this->config(array(
			"default_kitgroup" => "Kits",
		));

		$this->filesystem = $this->core->filesystem($this->core->path("root","kits"));

	}

	// PUBLIC

	public function icon ($kit) {

		if (are_empty(is_string($kit), is_int($kit), is_array($kit)))
			return false;

		if (is_string($kit) || is_int($kit))
			$kit = $this->installed($kit);

		if (is_empty($kit))
			return false;

		if (is_empty(key_of("folder", $kit)))
			return false;

		if (not_empty($icon = key_of("icon", $kit)) &&
			$this->filesystem->exists($kit["folder"]."/".$icon)
			)
			$file = $this->core->folder->path($this->core->path("url"),"kits",$kit["folder"])."/".$icon;

		else if ($this->filesystem->exists($kit["folder"].'/icon.png'))
			$file = $this->core->folder->path($this->core->path("url"),"kits",$kit["folder"])."/icon.png";

		else if ($this->filesystem->exists($kit["folder"].'/icon.svg'))
			$file = $this->core->folder->path($this->core->path("url"),"kits",$kit["folder"])."/icon.svg";

		else
			return false;

		return "/".$file."?version=".$this->core->file->age($this->filesystem->get_path($kit["folder"]).$this->core->file->name($file));

	}


	public function install($kit = []) {

		if (is_empty($kit))
			return false;

		$config = [];

	  $kit['group'] = either(
			key_of("group", $kit),
			$this->config("default_kitgroup")
		);

		$run = key_of("run", $kit);

		// Backwards Compatibility
		if (not_empty(key_of("frontend", $kit)))
			$run["frontend"] = key_of("frontend", $kit);

		if (not_empty(key_of("backend", $kit)))
			$run["backend"] = key_of("backend", $kit);

		if (not_empty(key_of("core", $kit)))
			$run["core"] = key_of("backend", $kit);

		if (not_empty(key_of("file_handling", $kit)))
			$run["file_handling"] = key_of("file_handling", $kit);

		if (not_empty(key_of("toolbar", $kit)))
			$run["toolbar"] = key_of("toolbar", $kit);

		if (not_empty(key_of("toolbar", $kit)))
			$run["widget"] = key_of("widget", $kit);

	  if ($group_ID = $this->data->group->read_per("name",$kit["group"],1))
	    $group_ID = $group_ID["ID"];

	  else
	    $group_ID = $this->data->group->create(array("name"=>$kit["group"]));

	  $create = array(
	    "name" => $kit["title"],
	    "folder" => $kit["folder"],
	    "unique_name" => either(fetch("unique_name",$kit),fetch("unique-name",$kit)),
	    "group_ID" => $group_ID,
			"description" => key_of(["info_text","description"],$kit),
			"info_text" => key_of(["info_text","description"],$kit),
			"config" => "",
	    "run" => $run,
			"icon" => either(key_of("icon", $kit), ""),
	  );

		return $this->data->create($create);

  }


	public function installed ($kit = NULL, $interface = NULL) {

		if ($interface === NULL)
			$interface = $this->core->runs();

		if (is_empty($kit))
			return false;

		$installed = $this->private_config("installed");

		if ($installed === NULL) {

			$correlates = $this->core->user->group->correlates($this->core->user->group->ID(),$this->data);

			$entries = $this->data->read();

			foreach ($entries as $entry) {

				// Cleanup for framework 35
				if (not_empty(key_of("config",$entry))) {

					if (file_exists($file = DROOT.KITS_FOLDER.key_of("folder",$entry).'/config.php')) {

						$kit_config = [];
						$run = [];
						$core = $this->core;

						include $file;

						if (is_empty($kit_config))
							continue;

						if (fetch("frontend",key_of("config",$entry)))
							$run["frontend"] = key_of("frontend", $kit_config);

						if (fetch("backend",key_of("config",$entry)))
							$run["backend"] = key_of("backend", $kit_config);

						$update = array(
							"ID" => key_of("ID",$entry),
							"config" => "",
							"run" => $run,
						);

						$this->data->update($update);

					}
				}

				if (any_empty(key_of("run", $entry), is_array($this->core->json->decode(key_of("run", $entry)))))
					continue;

				foreach ($this->core->json->decode(key_of("run", $entry)) as $set_interface => $run) {

					if (
						are_equal($set_interface, "backend") &&
						!$this->core->user->is_dev() &&
						!in_list(key_of("ID", $entry), $correlates)
					)
						continue;

					$installed[$set_interface][] = $this->core->array->merge(
						$entry,
						array("run" => $run)
					);
				}

			}

			$this->private_config("installed", $installed);

		}

		if (is_empty($installed))
			return false;

		if ($interface)
			$installed = key_of($interface, $installed);

		if ($kit === true)
			return $installed;

		if (is_numeric($kit))
			$kit = $this->core->array->filter($installed, ["ID" => $kit]);

		else if (is_string($kit))
			$kit = $this->core->array->filter($installed, ["unique_name" => $kit]);

		else
			return false;

		if (is_empty($kit))
			return false;

		if (is_indexed($kit))
			$kit = current($kit);

		return $kit;

	}


	public function load ($ID, $return = false, $interface = NULL) {

		if (are_equal($ID,"kits")) {

			if ($this->core->user->logged_in() && $this->core->runs("backend")) {
				$kit = array(
					"name" => "Kits",
					"folder" => "kb_standard/kb_kits",
					"unique_name" => "kb_kits",
					"run" => "backend.php"
				);
			}
			else {
				return false;
			}
		}

		else if (!$kit = $this->installed($ID, $interface)) {
			return false;
		}

		if ($interface === false)
			$kit = current($kit);

		if (is_indexed($kit))
			$kit = current($kit);

		// Save current base dir
		$base_dir = $this->core->filesystem->base_dir();

		$core = $this->core;
		$core_config_creator = $this->core->config_creator;

		$unique_name = key_of("unique_name", $kit);
		$this->core->config_creator = $unique_name;
		$this->core->filesystem->base_dir($this->core->path("files").$unique_name."/");

		$kit["path"] = $this->core->folder->path(
			$this->core->path("root"),
			$this->core->path("kits"),
			$kit["folder"]
		);

		// Backwards compatibility
		$kit["url"] =& $kit["path"];

		if (file_exists($kit["path"].'config.php')){

			include $kit["path"].'config.php';
			$kit["config"] = isset($kit_config) ? $kit_config : [];

			// Backwards Compatibility
			$kit["kit_config"] =& $kit["config"];

			if (isset($kit["config"]["unique-name"]))
				$kit["config"]["unique_name"] =& $kit["config"]["unique-name"];

			else if (isset($kit["config"]["unique_name"]))
				$kit["config"]["unique-name"] =& $kit["config"]["unique_name"];


			$kit["kit_option"] = isset($kit_option) ? $kit_option : [];

		}

		$kit_title = key_of("name",$kit);
		$this->id = key_of("ID",$kit);

		$this->core->config_creator = $core_config_creator;

		// Set current base dir
		$this->core->filesystem->base_dir($base_dir);

		if ($return){
			if (isset($$return))
				return $$return;
			else
				return false;
		}
		else{
			return $kit;
		}

	}


	public function load_toolkit ($ID = false, $options = false) {

		if (!$ID)
			return false;

		$kit = $this->load($ID);

		if (is_empty($kit))
			return false;

		if (is_empty(key_of("toolkit",key_of("config",$kit))))
			return false;

		$kit["caller"] = $this->caller;

		$this->toolkits[] = array($kit, $options);
		// $this->toolkits[] = $kit;
		return $kit;
	}


	public function run ($kit, $kit_options = [], $local_vars = []) {

		$kit_options = $this->core->json->decode($kit_options);

		if (is_string($kit_options))
			$kit_options = explode(",", $kit_options);

		// Makes $core available in functions

		$core = $this->core;
		$core_config_creator = $this->core->config_creator;

		if (is_string($kit) || is_int($kit))
			$kit = $this->installed($kit);

		if (is_empty($kit))
			return false;

		if (not_empty($local_vars))
			foreach ($local_vars as $name => $value)
				${$name} = $value;

		$kit["option"] = either(key_of("option", $kit), $kit_options);
		$kit["path"] = $this->core->path("root", "kits").key_of("folder", $kit)."/";


		// Backwards compatibility

		$kit["url"] =& $kit["path"];
		$kit["kit_option"] =& $kit["option"];
		$kit["kit_config"] =& $kit["config"];

		if (isset($kit["unique-name"]))
			$kit["unique_name"] =& $kit["unique-name"];
		else if (isset($kit["unique_name"]))
			$kit["unique-name"] =& $kit["unique_name"];


		// $kit["fs"] = new StdClass;
		$kit["fs"] = $this->core->filesystem($kit["path"]);
		$kit["fs"]->frontend = $this->core->filesystem($this->core->path("files").$kit["unique_name"]);

		if (file_exists($kit["path"].'config.php')) {

			try {
				$return = include $kit["path"].'config.php';
				if (!$return)
					return;
			}
			catch (Throwable $e){
				$this->core->message->dev($e,"Error",2);
				return false;
			}
			catch (Exception $e){
				$this->core->message->dev($e,"Error",2);
				return false;
			}
		}

		// For global access in kits
		$this->current =& $this->caller;
		$this->title = key_of("name", $kit);

		// Save current base dir
		$base_dir = $this->core->filesystem->dir();

		$unique_name = key_of("unique_name",$kit);
		$this->core->config_creator = $unique_name;

		// For Toolkits
		$this->caller = $kit;

		if ($this->core->runs("backend")) {

			$favicon = $this->icon($unique_name);

			if (fetch($this->core->file->extension($favicon), array("png")))
				$this->core->template->backend->favicon($favicon);
		}

		// if (!$this->core->file->exists($this->core->path("root").$this->core->path("kits").key_of("folder",$kit)."/".key_of("run",$kit)))
		if (!$this->core->file->exists($kit["fs"]->dir().key_of("run",$kit)))
			return false;

		$this->core->filesystem->dir($this->core->path("files").$unique_name."/");

		ob_start();

		try {
			// $return = include $this->core->path("root","kits").key_of("folder",$kit)."/".key_of("run",$kit);
			$return = include $kit["fs"]->dir().key_of("run",$kit);
		}
		catch (Throwable $e){
			$this->core->message->dev($e,"Error",2);
			$this->core->filesystem->dir($base_dir);
			return false;
		}
		catch (Exception $e){
			$this->core->message->dev($e,"Error",2);
			$this->core->filesystem->dir($base_dir);
			return false;
		}

		$return = ifttt($return === 1, "", $return);
		$this->core->config_creator = $core_config_creator;
		$this->core->filesystem->dir($base_dir);

		return either(ob_get_clean(), $return);

	}


	public function run_all ($area = "", $options = []) {

		$return = [];

		if (not_empty($kits = $this->installed(true, $area)))
			foreach($kits as $kit)
				if (not_empty($return_kit = $this->run($kit, $options)))
					$return[] = $return_kit;

		return either($return, false);

	}


	public function widget ($ID = false) {

		if (is_empty($ID))
			return false;

		if (is_empty($kit = $this->load($ID)))
			return false;

		if (is_empty($widget = $this->installed($ID, "widget")))
			return false;

		if (is_empty($widget = key_of("run", $widget)))
			return false;

		if (is_string($widget))
			$widget = array(
				"file" => $widget
			);

		$content = $this->core->file->read_php(key_of("path",$kit).key_of("file",$widget), ["kit"=>$kit]);

		if (is_empty($content))
			return false;

		$widget = $this->core->array->union(
			array(
				"size" => array("1","1"),
				"link" => true,
				"content" => $content,
			),
			$widget
		);

		return $widget;

	}


	// PRIVATE

	private function install_shipped() {

		if (!$this->core->user->is_dev())
			return;

		if (not_empty($this->data->columns()))
			return;

		if (is_empty($this->data->group->columns())) {
			$this->data->group->create(array("name"=>"Website"));
			$this->data->group->create(array("name"=>"System"));
			$this->data->group->create(array("name"=>"Tools"));
		}

		$kits = $this->check_dir(DROOT.KITS_FOLDER);

		if (not_empty($kits))
			foreach ($kits as $kit)
				$this->install($kit);

	}
	private function object(){

    if (isset($this->data) && is_object($this->data))
      return $this->data;

		$data = $this->core->data("kb_kit",["name","folder","unique_name","run","info_text"]);
		$data->indexed(true);
		$data->expandable(true);
		$data->position(true);
		$data->group(true);
		$data->group->expandable(true);
		$data->group->indexed(true);
		$data->group->position(true);
		$data->group->group = false;

		return $data;
	}




	// DEPRECATED

	public function render($string=""){
		deprecated();
		$render = '<div class="kitcontent">';
		$render .= $string;
		$render .= '</div>';
		return $render;
	}

	public function get_config($key="") {
		deprecated();
		if (empty($key)){ return $this->kit_config; }
		elseif (!is_array($key) && array_key_exists($key, $this->kit_config)){ return $this->kit_config[$key]; }
		elseif (is_array($key))
		{
			$return = $this->kit_config;
			foreach($key AS $k)
			{
				$return = (is_array($return)) ? $return[$k] : $return;
			}
			return $return;
		}
		return false;
	}

	public function run_frontend($options){
		deprecated();
		if ($this->kit_config['frontend']){
			return $this->folder.$this->kit_config['frontend'];
		}
		return false;
	}

	public function get_merged_options($cat, $key) {

		deprecated();

		if (isset($this->kit_option[$cat]) && !empty($this->kit_option[$cat]))
		{
			$return = [];

			$option_keys = array_merge(
				(isset($this->kit_option[$cat]['default']) && is_array($this->kit_option[$cat]['default'])) ? array_keys($this->kit_option[$cat]['default']) : [],
				(isset($this->kit_option[$cat][$key]) && is_array($this->kit_option[$cat][$key])) ? array_keys($this->kit_option[$cat][$key]) : [],
				(isset($this->kit_option['group_'.$this->user->user_data('group_ID')][$cat]) && is_array($this->kit_option['group_'.$this->user->user_data('group_ID')][$cat])) ? array_keys($this->kit_option['group_'.$this->user->user_data('group_ID')][$cat]) : [],
				(isset($this->kit_option['user_'.$this->user->user_ID()][$cat]) && is_array($this->kit_option['user_'.$this->user->user_ID()][$cat])) ? array_keys($this->kit_option['user_'.$this->user->user_ID()][$cat]) : []

				);

			if (is_array($option_keys) && !empty($option_keys))
			{
				foreach($option_keys AS $k)
				{
					$return[$k] = array_unique(array_merge(
										(isset($this->kit_option[$cat]['default'][$k])) ? (array)$this->kit_option[$cat]['default'][$k] : [],
										(isset($this->kit_option[$cat][$key][$k])) ? (array)$this->kit_option[$cat][$key][$k] : [],
										(isset($this->kit_option['group_'.$this->user->user_data('group_ID')][$cat][$k])) ? (array)$this->kit_option['group_'.$this->user->user_data('group_ID')][$cat][$k] : [],
										(isset($this->kit_option['user_'.$this->user->user_ID()][$cat][$k])) ? (array)$this->kit_option['user_'.$this->user->user_ID()][$cat][$k] : []
									));
					if (isset($return[$k][0]) && count($return[$k])==1){ $return[$k] = $return[$k][0]; }
				}
			}
			/*
			return array_merge(
				(isset($this->kit_option[$cat]['default']) && is_array($this->kit_option[$cat]['default'])) ? $this->kit_option[$cat]['default'] : [],
				(isset($this->kit_option[$cat][$key]) && is_array($this->kit_option[$cat][$key])) ? $this->kit_option[$cat][$key] : [],
				(isset($this->kit_option[$this->user->user_data('group_ID')][$cat]) && is_array($this->kit_option[$this->user->user_data('group_ID')][$cat])) ? $this->kit_option[$this->user->user_data('group_ID')][$cat] : [],
				(isset($this->kit_option[$this->user->user_ID()][$cat]) && is_array($this->kit_option[$this->user->user_ID()][$cat])) ? $this->kit_option[$this->user->user_ID()][$cat] : []

				);
			*/

			return $return;
		}
		return false;
	}

	public function set_data($key, $value)	{
		deprecated();
		$this->$key = htmlspecialchars($value);
	}

	public function get_data($value) {
		deprecated();
		if (isset($this->data[$value])){ return $this->data[$value];}
		return false;
	}

	public function info ($ID = false) {
		deprecated();

		if ($ID){
			if (empty($this->id) ||$this->id != $ID){
				if (!$this->load($ID)){
					return false;
				}
			}
			if (isset($this->kit_config["info_file"]) && is_file($this->folder.$this->kit_config["info_file"])){
				ob_start();
				include $this->folder.$this->kit_config["info_file"];
				return ob_get_clean();
			}
			else if (isset($this->kit_config["info_text"]) && $this->kit_config["info_text"] != ""){
				return $this->kit_config["info_text"];
			}else{
				return false;
			}
		}else{
			return false;
		}
	}


	// UNSORTED

  function read_local_kits() {

		$kits = $this->private_config("local_kits");

    if (!is_null($kits))
      return $kits;

    $dir_exclude = array(".", "..", ".svn", "_notes","kb_kits");

    $kits = $this->check_dir(DROOT.KITS_FOLDER,$dir_exclude);

		$this->private_config("local_kits", $kits);

    return $kits;
  }

  function return_group($ID){
    $count = $this->data->count_per("group_ID",$ID);
    return $count;
  }

	function check_dir($kit_dir,$dir_exclude=false, $kits = []){

    if (!$dir_exclude)
      $dir_exclude = $this->dir_exclude;

    $core = $this->core;
    $dir = DROOT.KITS_FOLDER;

    $files = scandir($kit_dir);
    // return false;

    if (!empty($files)) {

      if (in_array('config.php', $files)) {

        $kit_config = []; //Reset temporary config-array

				// Provide kit variable as kit->load does

				$kit = array(
					"path" => $kit_dir."/"
				);

				try {
					include $kit_dir."/config.php";
				}
				catch (Throwable $e){
					$this->core->message->dev($e,"Error",2);
					return $kits;
				}
				catch (Exception $e){
					$this->core->message->dev($e,"Error",2);
				 	return $kits;
				}

		    $kit_config["folder"] = trim(str_replace($dir, "",$kit_dir),"/");
        $kits[$kit_config["folder"]] = $kit_config;
      }

	    else if (in_array('index.php', $files)) {

        $kit_config = []; //Reset temporary config-array
        $kit_config["folder"] = trim(str_replace($dir, "",$kit_dir),"/");
        $kit_config["title"] = substr($kit_dir,strrpos($kit_dir,"/")+1);
        $kits[$kit_config["folder"]] = $kit_config;
      }

      else {

        foreach($files as $f)
          if (is_dir($kit_dir."/".$f) && !in_array($f, $dir_exclude))
            $kits = $this->check_dir($kit_dir."/".$f,$dir_exclude,$kits);

      }
    }
    return $kits;
  }


	function toolkits(){

		$toolkits = [];

		if (is_empty($this->toolkits))
			return false;

		// Make core object available to kit

		$core = $this->core;

		foreach ($this->toolkits as $array) {

			$kit = key_of(0, $array);
			$options = (array) key_of(1, $array);
			$options["kit"] = $kit;

			$toolkit = $this->core->file->read_php(fetch("url",$kit).fetch("toolkit",$kit["config"]), $options);

			$toolkit = array(
				$this->core->html->tag("h2", key_of("title",$kit["config"]), array("class"	=> "")),
				$toolkit
			);

			$toolkits[] = $this->core->html->tag("div",$toolkit,array(
				"class" => "kit"
			));

		}

		$markup = $this->core->html->render($toolkits);

		return $markup;
	}

	function run_toolkit($kit, $options){
		if (isset($kit["kit_config"]['toolkit'])){
			return array($kit["kit_config"]['title'],$kit["folder"].$kit["kit_config"]['toolkit'],$kit["kit_config"]);
		}
	}


}
