<?php

class KB_string extends KB_class_extended {

	public
		$format,
		$json,
		$markdown;

	private $core;

	function __construct($core) {
		parent::__construct($core);
		$this->core = $core;
		$this->format = new KB_format($core);
		$this->json = new KB_json($core);
		$this->markdown = new KB_markown($core);
	}


	// Public

	public function encode ($string = "", $encode_from = "") {

		if (is_empty($string))
			return $string;

		if (is_empty($encode_from))
			$encode_from = $this->encoding($string);

		if (is_empty($encode_from))
			return $string;

		$encode_to = mb_convert_encoding($string, "UTF-8", $encode_from);

		if (is_empty($encode_to))
			return $string;

		return $encode_to;

	}


	public function lowercase ($string = "") {

		if (is_empty($string))
			return $string;

		if (is_array($array = $string)) {

			foreach ($array as $key => $value)
				$array[$key] = $this->lowercase($value);

			return $array;

		}

		if (!is_string($string))
			return $string;

		if (function_exists("mb_strtolower"))
			return mb_strtolower($string);

		return strtolower($string);
	}


	public function matches ($needle = "", $haystack = "") {

		if (any_empty(is_string($needle), is_string($haystack)))
			return false;

		if (strstr($haystack, $needle) !== false)
			return true;

		return false;
	}


	public function recode ($string = "", $encode_to = "UTF-8") {

		if (is_empty($string))
			return $string;

		if (is_empty($encode_to) || !is_string($encode_to))
			return $string;

		$encode_from = $this->encoding($string);

		$encode_to = mb_convert_encoding($string, $encode_to, $encode_from);

		if (is_empty($encode_to))
			return $string;

		return $encode_to;

	}


	public function replace ($replace_this = "", $with_this = "", $string = "") {

		if (any_empty($replace_this, $string))
			return $string;


		$replace_this = preg_mod($replace_this);

		return preg_replace($replace_this, $with_this, $string);

	}


	public function trim ($start = "", $end = "", $string = "", $inner_cut = false) {

		$args = func_get_args();
		$count = func_num_args();

		if ($count === 0)
			return $this->trim("");

		if ($count === 1)
			return $this->trim("","",key_of(0,$args));

		if ($count === 2)
			return $this->trim(key_of(0,$args),"",key_of(1,$args));

		if (is_empty($string))
			return $string;

		if (are_empty($start,$end))
			return trim($string);

		if (not_empty($start)) {
			$start_pos = strpos($string, $start);

			if ($inner_cut)
				$start_pos += strlen($start);

			$string = substr($string, $start_pos);
		}

		if (not_empty($end)) {
			$end_pos = strpos($string, $end);

			if (!$inner_cut)
				$end_pos += strlen($end);

			$string = substr($string, 0, $end_pos);
		}

		return $this->trim($string);

	}


	public function uppercase ($string = "") {

		if (is_empty($string))
			return $string;

		if (is_array($array = $string)) {

			foreach ($array as $key => $value)
				$array[$key] = $this->uppercase($value);

			return $array;

		}

		if (!is_string($string))
			return $string;

		if (function_exists("mb_strtoupper"))
			return mb_strtoupper($string);

		return strtoupper($string);
	}

	// Private

	private function encoding($string = "") {


		if (either(
				is_empty($string),
				preg_match('//u', $string)
			))
			return "UTF-8";

		$bom_short = substr($string, 0, 2);
		$bom_long = substr($string, 0, 4);

		if (are_equal($bom_long,"\x00\x00\xFE\xFF"))
			return "UTF-32BE";

		if (are_equal($bom_long,"\xFF\xFE\x00\x00"))
			return "UTF-32LE";

		if (are_equal($bom_short,"\xFE\xFF"))
			return "UTF-16BE";

		if (are_equal($bom_short,"\xFF\xFE"))
			return "UTF-16LE";

		if (preg_match('/[\x00-\x08\x0E-\x1F\x81\x8D\x8F\x90\x9D]/', $string))
        return false;

		if (preg_match('/[\x80-\x9F]/', $string))
      return "Windows-1252";

		return "ISO-8859-1";

	}

}


class KB_format extends KB_class_extended {

	private $core;

	function __construct($core) {
		parent::__construct($core);
		$this->core = $core;
	}


	public function currency($value, $currency = "€") {

		if (is_array($values = $value) || is_array($values = $this->core->json->decode($value))) {

			$return = array();

			foreach ($values as $value) {
				$return[] = $this->currency($value, $currency);
			}
			// return $this->core->html->render($return);
			return $return;
		}
		if (is_empty($value))
			return $value;

		// Convert to integer
		$value = $this->core->calc->intify($value);

		return number_format($value,2,",",".").ifttt($currency," $currency");
	}


	public function date($date = "NOW", $format = "d.m.Y", $locale = false) {

		if (is_empty($date))
			return "";

		// UNIX time
		if (preg_match("/^[0-9]+$/",$date))
			$date = "@".$date;

		if ($locale) {
			$locale = key_of(0,$this->core->language->accept());
			setlocale(LC_TIME, $locale);
			return utf8_encode(strftime($format, strtotime($date)));
		}

		return date($format, strtotime($date));

	}


	public function datetime($date){
		if(is_empty($date))
			return "";

		// UNIX time
		if(preg_match("/^[0-9]+$/",$date))
			$date = "@".$date;


		return date("d.m.Y H:i:s",strtotime($date));
	}


	public function email($value){

		if (is_array($values = $value) || is_array($values = $this->core->json->decode($values))) {

			$return = array();

			foreach ($values as $value) {
				$return[] = $this->email($value);
			}

			return $this->core->html->render($return);
		}

		if (!preg_match(preg_mod("^[^@]+@[^.@]+[.][^@]+[\w]$"),$value))
			return $value;

		return $this->core->html->render(
			$this->core->html->tag("a", $value ,array("href"=>"mailto:".$value))
		);

	}


	public function file_icon($values = false, $options = []){

		if (is_empty($values))
			return "";

		$values = $this->core->json->decode($values);

		$options = array_merge(
			array(
				"return_as_html" => true,
				"render_filetypes" => false,
				"class" => false,
				"thumbnail" => false,
				"raster" => true,
				"size" => false,
				"version" => true,
				"target" => false,
				"container" => false,
				"link" => true,
			),
			$options
		);

		$icons = [];

		foreach ((array) $values as $file) {

			$icon = $this->core->template->file_icon($this->core->file->extension($file));

			if (key_of("link", $options))
				$icon = $this->core->html->a($icon, ["href"=>"/".$this->core->path("files").$file]);

			$icons[] = $icon;

		}

		// $return = "";

		if ($container = key_of("container", $options))
			$icons = $this->core->html->tag(key($container), $icons, current($container));

		if (fetch("return_as_html",$options))
			$icons = $this->core->html->render($icons);

		return $icons;

	}


	public function file($values = false, $path = false, $options = array()){


		if (is_empty($values))
			return "";

		$values = $this->core->json->decode($values);

		if (is_string($values) && $path === false)
			$values = array($values);

		if (is_string($values) && is_array($array = $this->core->json->decode($values)))
			$values = $array;

		if (is_string($values) && is_string($path))
			$values = array(
				"filename"	=> $values,
				"path"			=> $path
			);

		if (/*is_array($values) &&*/ is_array($path) && empty($options)) {
			$options = $path;
			$path = false;
		}

		$options = array_merge(
			array(
				"return_as_html" => true,
				"render_filetypes" => false,
				"class" => false,
				"thumbnail" => false,
				"raster" => true,
				"size" => false,
				"version" => true,
				"target" => false,
				"container" => false,
			),
			$options
		);

		if (!is_indexed($values))
			$values = array($values);

		/*
		if (
				(key_of("thumbnail",$options) || !key_of("render_filetypes",$options)) &&
				either(BACKEND, false/*, key_of("class",$options)*//*)

			) {
			$return["span"] = array(
					"class" => BACKEND ? "zoom light" : key_of("class",$options)
			);
			// $options["class"] = "";
		}
		*/

		foreach ($values as $file) {

			if (is_string($file))
				$file = array(
					"path" => $file,
					"filename" => $this->core->file->name($file)
				);

			if (neither(
				$this->core->file->exists($file["path"]),
				$this->core->file->exists(DROOT.FILES_FOLDER.$file["path"]),
				$this->core->file->exists(DROOT.TEMPLATE_FOLDER.$file["path"])
			)) {

				$markup = array(
					array(
						"img" => array(
							"class" => "icon",
							"src" => $this->core->template->icon->get("warning", "red")
						)
					),
					array(
						"span" => array(
							"class" => "red",
							"title" => $file["path"],
							$file["filename"]
						)
					)
				);
				$return[] = $markup;
				continue;
			}


			$file["path"] = preg_replace("/^([^\/])/","/$1",$file["path"]);

			if ($this->core->file->exists(DROOT.TEMPLATE_FOLDER.$file["path"]))
				$file["path"] = "/templates/".$file["path"];

			if ($this->core->file->exists(DROOT.FILES_FOLDER.$file["path"]))
				$file_path = DROOT.FILES_FOLDER.$file["path"];

			else if ($this->core->file->exists(DROOT.FRONTEND_FOLDER.$file["path"]))
				$file_path = DROOT.FRONTEND_FOLDER.$file["path"];

			else
				$file_path = $file["path"];

			if (fetch("version",$options) === true)
				$file["path"] = $file["path"]."?version=".filemtime($this->core->file->get_path($file_path));

			if (preg_match("-\.html-",$file["filename"]) && !preg_match("-//-",$file["path"]))
				$file["path"] = "/files".$file["path"];

			$markup = $file["filename"];

			if ($filetypes = key_of("render_filetypes",$options)) {

				$pathinfo["extension"] = $this->core->file->extension($file["filename"]);
				$filetypes = array();

				$filetypes["raster"] = array("jpg","jpeg","png","gif");
				$filetypes["rasterable"] = array("pdf", "mp4", "webm", "mov");
				$filetypes["vector"] = array("svg");
				$filetypes["video"] = array("mp4", "webm", "mov");

				if (key_of("thumbnail",$options)) {
					$options["thumbnail_class"] = either(key_of("thumbnail_class", $options), key_of("class", $options));
					$options["file_class"] = either(key_of("file_class", $options), "");
					$options["class"] = "";
				}

				if (fetch(strtolower($pathinfo["extension"]),$filetypes["video"]) && !key_of("raster",$options)) {
					$markup = array(
						"video" => $this->core->array->merge(array(
							"src" => $file["path"],
							"class" => either(key_of("file_class",$options), key_of("class",$options), ifttt($this->core->runs("backend"),"gen_file")),
							"preload" => "auto"
						),key_of("attr", key_of("video", $options)))
					);
				}

				else if (fetch(strtolower($pathinfo["extension"]),keys_of(array("raster", "rasterable"),$filetypes))) {
					$markup = array(
						"img" => array(
							"src" => key_of("thumbnail",$options) ? thumbnail($file["path"], key_of("thumbnail",$options)) : ( key_of("size",$options) ? thumbnail($file["path"], key_of("size",$options)) : $file["path"]),
							"class" => either(key_of("file_class",$options), key_of("class",$options), ifttt($this->core->runs("backend"),"gen_file")),
						)
					);
				}

				else if(fetch(strtolower($pathinfo["extension"]),$filetypes["vector"])){
					$markup = array(
						"img" => array(
							"src" => $file["path"],
							"class" => either(key_of("file_class",$options), key_of("class",$options), ifttt($this->core->runs("backend"),"gen_file")),
						)
					);
				}
			}

			if (key_of("thumbnail",$options) || !key_of("render_filetypes",$options)) {
				$return[] = array(
					"a"	=> array(
						"href" => (fetch("size",$options) && fetch($pathinfo["extension"],$filetypes["raster"])) ? thumbnail($file["path"], fetch("size",$options)) : either(key_of("target", $options), $file["path"]),
						"title" => $this->core->runs("backend") ? $file["filename"] : false,
						"class" => either(key_of("thumbnail_class",$options), key_of("class",$options), ifttt($this->core->runs("backend"),"zoom")),
						$markup,
					)
				);
			}

			else {
				$return[] = $markup;
			}
		}

		if ($container = key_of("container", $options))
			$return = $this->core->html->tag(key($container), $return, current($container));

		if (fetch("return_as_html",$options))
			$return = $this->core->html->render($return);

		return $return;
	}


	public function filesize($value, $unit="auto"){

		if (is_empty($value))
			return $value;

		$value = (int) $value;

		$units = array(
			0 => "Bytes",
			1 => "KB",
			2 => "MB",
			3 => "GB",
		);

		if($unit == "auto"){
			$power = floor(log($value, 1024));
			$unit = $units[$power];
		}

		if(in_array($unit,$units))
		return number_format($value/pow(1024,array_search($unit, $units)),2,",",".")." ".$unit;

		return number_format($value/pow(1024,0),2,",",".")." Bytes";


	}


	public function phone($value) {

		if (is_array($values = $value) || is_array($values = $this->core->json->decode($values))) {

			$return = array();

			foreach ($values as $value) {
				$return[] = $this->phone($value);
			}

			return $this->core->html->render($return);
		}

		$number = preg_replace(array(
			preg_mod("[ \-()]"),
			"-&#40;-",
			"-&#41;-",
		),"",$value);
		$return = '<a onclick="javascript:return confirm(\'Call this number?\')" href="tel:'.$number.'">'.$value.'</a>'.br;

		return $return;
	}


	public function time($time){

		if(empty($time))
			return $time;

		// UNIX time
		if(preg_match("/^[0-9]+$/",$time))
			$time = "@".$time;

		$time = gmdate("U",strtotime($time));

		return floor($time/3600).":".str_pad($time%3600/60, 2, "0");
	}


	public function website($values){

		$return = "";

		if (!is_array($values))
			$values = array($values);

		foreach ($values as $value) {

			if (!preg_match(preg_mod("[\w][.][\w]"),$value))
				$return .= $value;

			$href = $value;

			if (!preg_match(preg_mod("[\w]+://[\w]+"),$value))
				$href = "//".$href;

			$return .= '<a href="'.$href.'" target="_blank">'.$value.'</a>'.br;
		}
		return $return;
	}


	// Alias

	public function euro ($value) {
		return $this->currency($value, "€");
	}

	public function mail ($value) {
		return $this->email($value);
	}

}


class KB_json extends KB_class_extended {

	private $core;

	function __construct($core) {
		parent::__construct($core);
		$this->core = $core;
	}


	public function encode($value = "") {

		//Alternative to PHPs json_encode Function with proper Encoding

		if (is_int($value) || is_string($value))
		return $value;

		$value = add_slashes($value);
		$value = json_encode($value);
		$value = preg_replace_callback(
			'/\\\\u(\w{4})/',
			function ($matches) {
				return html_entity_decode('&#x' . $matches[1] . ';', ENT_COMPAT, 'UTF-8');
			},
			$value
		);
		return $value;

	}

	public function decode($value = "") {

		// No JSON

		if (is_array($value))
			return $value;

		if (is_empty($value))
			return $value;

		if (!preg_match("-([\{\[]+)-",$value)){
			return $value;
		}

		$value_backup = $value;

		// Convert newlines within strings
		$value = preg_replace_callback("-\"([^\"]*)\"-s", function($matches){
			$matches[0] = preg_replace(array("-\r-","-\n-","-\t-"),array("\\r","\\n","\\t"),$matches[0]);
			return $matches[0];
		},$value);

		// JSON
		if (!is_null($array = json_decode($value,true)))
			return preg_replace_array("-\\\\\"-","\"",$array);
			// return $array;

		// HTML special chars
		if (matches("&quot;", $value) && !matches('"', $value))
			$value = htmlspecialchars_decode($value);

		// HJSON
		// Remove line breaks
		$value = preg_replace('/^(.+)"(\s*)$/m','\1",',$value);
		// Remove comments
		$value = preg_replace('-^(\s*)(//|#)(.*)$-m','',$value);
		$value = preg_replace('-([\t])*/\*([^*]*)\*/([\n\r])*-','',$value);
		//Add missing commas
		$value = preg_replace('/([\w}])([\r\n]*)$/m','\1,',$value);
		// Add missing quotes
		$value = preg_replace('/^([{\r\n\s\t]+)(\w+)(\s*)(:)/m','\1"\2"\3\4',$value);
		$value = preg_replace('/(\"\s*:)(\s*)(?!null|true|false)(\w[\w\s$=]*)(\s*)([},])/','\1\2"\3"\4\5',$value);
		// Remove trailing commas
		$value = preg_replace('/,([\s]*)([}\]]){1}/','\1\2',$value);
		$value = preg_replace('/,$/','',$value);

		if (!is_null($array = json_decode($value,true)))
			return preg_replace_array("-\\\\\"-","\"",$array);

		// return $value;
		return $value_backup;


	}
}


class KB_markown extends KB_class_extended {


	private $core;


	function __construct($core) {
		parent::__construct($core);
		$this->core = $core;
	}


	public function decode ($text = "", $options = array()) {

		if (is_empty($text))
			return "";

		if (is_array($texts = $text)) {

			$return = array();

			foreach ($texts as $key => $text) {
				$return[$key] = $this->decode($text, $options);
			}

			return $return;

		}

		$options = $this->core->array->merge(
			array(
				"shift" => 0
			),
			$options
		);

		$replacement = array(
			'/^######(.*)$/m'
			=> '<h6>\1</h6>',

			'/^#####(.*)$/m'
			=> '<h5>\1</h5>',

			'/^####(.*)$/m'
			=> '<h4>\1</h4>',

			'/^###(.*)$/m'
			=> '<h3>\1</h3>',

			'/^##(.*)$/m'
			=> '<h2>\1</h2>',

			'/^#(.*)$/m'
			=> '<h1>\1</h1>',

			// '/^([\s]+)(-|\*)(\s)(.+)$/m'
			'/^([ \t]+)(-|\*)(\s)(.+)$/m'
			=> '<ul><ul><li>\4</li></ul></ul>',

			'/^(-|\*)(\s)(.+)$/m'
			=> '<ul><li>\3</li></ul>',

			'/<\/ul><\/ul>([\r\n]+)<ul><ul>/m'
			=> '\1',

			'/<\/ul>([\r\n]+)<ul>/m'
			=> '\1',

			'/^([0-9]+\.)(\s+)(.+)$/m'
			=> '<ol><li>\3</li></ol>',

			'/<\/ol>([\r\n]+)<ol>/m'
			=> '\1',

			'/^(> )([^<\r\n]{1})(.+)$/m'
			=> '<blockquote><p>\2\3</p></blockquote>',

			'/^([^<\r\n]{1})(.+)$/m'
			=> '<p>\1\2</p>',

			'/([^\s]+)(\s?)<\/p>(\s?)<p>(\s?)([\w\[+]+)/'
			=> '\1<br>\5',

			'/<p>([\s]*)<\/p>/'
			=> '',

			'/(\[)([^\]]+)(\])(\()([^)]+)(\))/'
			=> '<a href="\5">\2</a>',

			'/(\[)([\w.-]+@[\w.-]+\.\w+)(\])/'
			=> '<a href="mailto:\2">\2</a>',

			'/(\[)(http)([^\]]+)(\])/'
			=> '<a href="\2\3">\2\3</a>',

			'/\*\*([^\n]+)\*\*/m'
			=> '<strong>\1</strong>',

			'/\*([^\n*>]+)\*/m'
			=> '<em>\1</em>',

			'/^_([^\n_>]+)_/m'
			=> '<em>\1</em>',

			'/\s_([^\n_>]+)_/m'
			=> '<em>\1</em>',

			'/(\d)\\\(\.)/m'
			=> '\1\2',
		);

		$pattern = array_keys($replacement);

		$replacement = array_values($replacement);

		// Shift headlines
		if (key_of("shift", $options)) {
			for ($i = 6; $i > 0; $i--) {
				$key = $i - intval(key_of("shift", $options));
				if ($key < 0)
				$key = 0;
				$replacement[$i] = $replacement[$key];
			}
		}

		$text = preg_replace($pattern,$replacement,$text);

		return $text;

	}
}
