<?php

class KB_template extends KB_data {

	private $core, $filesystem;
	public $backend, $frontend, $kit, $icon;

	function __construct($core){

		parent::__construct($core);

		$this->core = $core;
		$this->filesystem = $this->core->filesystem();
		$this->filesystem->dir($this->core->path("templates"));
		$this->backend = new KB_template_backend($core);
		$this->frontend = new KB_template_frontend($core);
		$this->kit = new KB_template_kit($core);
		// $this->kit->panel = new KB_template_panel($core);
		$this->icon = new KB_template_icon($core);

		$colors = array(
			"csv" => "hsl(144, 54%, 30%)",
			"doc" => "hsl(200, 67%, 39%)",
			"pdf" => "hsl(7, 81%, 50%)",
			"zip" => "hsl(43, 65%, 49%)",
		);

		$colors = array_merge(
			$colors,
			array(
				"docx" => &$colors["doc"],
				"gz" => &$colors["zip"],
				"tar" => &$colors["zip"],
				"xls" => &$colors["csv"],
				"xlsx" => &$colors["csv"],
			)
		);

		$this->icon->config("file_icon_colors",$colors);
	}


	// Public

	public function create($array = []) {

		$array = $this->create_filename($array);
		$array = $this->create_folder($array);

		if ($array === false)
			return false;

		$this->filesystem->file->write(
			$this->filesystem->path($array["folder"]).$array["filename"],
			html_entity_decode(key_of("content",$array),ENT_QUOTES)
		);

		unset($array["content"]);

		return parent::create($array);
	}


	public function delete($value = false){

		$entry = $this->entry($value);

		if (is_empty($entry))
			return false;

		$entries = $this->read(array("folder"=>$entry["folder"]));

		// Last/single entry for a folder
		// Delete folder
		if (count($entries) < 2)
			$this->filesystem->folder->delete($entry["folder"]);

		// Multiple entries with same folder
		// Delete file only
		else
			$this->filesystem->folder->delete($entry["folder"]."/".$entry["filename"]);

		// Delete database entry
		return parent::delete($value);
	}


	public function entry($filter = false, $options = false){

		$entry = parent::entry($filter, $options);

		if (is_empty($entry))
			return $entry;

		$entry["content"] = $this->filesystem->file->read(
			key_of("folder",$entry)."/".key_of("filename",$entry)
		);
		return $entry;
	}


	public function file_icon ($extension = "", $color = "", $markup = []) {

		if (is_empty($extension))
			return $this->icon("document", $color, $markup);

		$colors = $this->icon->config("file_icon_colors");

		$badge_color = either(
			$color,
			either(
				key_of($this->core->string->lowercase($extension), $colors),
				"black"
			)
		);

		$icon = $this->icon->get("document");

		if (is_empty($icon))
			return "";

		$icon = $this->core->svg->parse($icon);

		if (is_empty(key_of("svg", key_of(0,$icon))))
			return "";

		if (strlen($extension) <= 3)

			$badge = [
				$this->core->xml->rect([
					"x"=>"2", "y"=>"10", "width"=>"20", "height"=>"10", "style"=>"fill:$badge_color"
				]),
				$this->core->xml->text($this->core->string->uppercase($extension), [
					"x"=>"3", "y"=>"18", "fill"=>"white", "font-size"=>"8", "font-family"=>"system-ui","font-weight"=>"500"
				])
			];

		else

			$badge = [
				$this->core->xml->rect([
					"x"=>"0", "y"=>"10", "width"=>"24", "height"=>"10", "style"=>"fill:$badge_color"
				]),
				$this->core->xml->text($this->core->string->uppercase($extension), [
					"x"=>"2", "y"=>"17", "fill"=>"white", "font-size"=>"7", "font-family"=>"system-ui","font-weight"=>"500"
				])
			];

		$icon[0]["svg"][] = $badge;

		if ($markup)
			$icon = $this->core->svg->property($icon, $markup);

		/*
		$markup = $this->core->xml->svg([$icon, $badge],$this->core->array->union([
			"width" => "1em",
			"height" => "1em",
			"preserveAspectRatio" => "xMidYMid meet",
			"viewBox" => "0 0 24 24",
			"style" => ifttt($color, "color:$color", NULL)
		],$markup));
		*/

		return $this->core->svg->render($icon);

	}


	public function icon($icon = false, $color = "", $markup = []) {

		if ($icon === false)
			return "";

		$icon = $this->icon->get($icon);

		if (is_empty($icon))
			return "";

		if (is_array($color)) {
			$markup = $color;
			$color = "";
		}

		if ($color)
			$icon = $this->core->svg->property($icon, "style","color:$color");

		if ($markup)
			$icon = $this->core->svg->property($icon, $markup);

		/*
		$markup = $this->core->xml->svg($icon,$this->core->array->union([
			"xmlns" => "http://www.w3.org/2000/svg",
			"width" => "1em",
			"height" => "1em",
			"preserveAspectRatio" => "xMidYMid meet",
			"viewBox" => "0 0 24 24",
			"style" => ifttt($color, "color:$color", NULL)
		],$markup));

		*/
		return $this->core->xml->render($icon);
	}


	public function import($values = [], $path = "", $file = "") {

		if (is_empty($path))
			return false;

		if (is_empty(key_of("name",$values)))
			return false;

		if (is_empty($file))
			$file = fetch("(.+)\.php",$this->filesystem->folder->read($path));

		if (is_empty($file))
			return false;

		if (is_array($file))
			$file = current($file);

		return parent::create(array(
			"name" => key_of("name",$values),
			"folder" => $path,
			"filename" => $file,
			"title" => key_of("title",$values),
			"doctype" => key_of("doctype",$values),
			"interfaces" => key_of("interfaces",$values),
			"favicon" => key_of("favicon",$values),
		));
	}


	public function read_per($key = false, $value = false, $limit = false){

		$entry = parent::read_per($key, $value, $limit);

		if (is_empty($entry))
			return $entry;

		$entry["content"] = $this->filesystem->file->read(
			key_of("folder",$entry)."/".key_of("filename",$entry)
		);
		return $entry;
	}


	public function update($array, $where = []) {

		if (!is_array($array))
			return false;

		$ID = key_of($this->key, $array);

		if (is_empty($where))
			$where = array($this->key=>$ID);

		$filter = [];

		foreach($where as $key => $entry){
			if (is_numeric($key))
				$filter = $filter+array($entry=>fetch($entry,$array));
			else
				$filter = $filter+fetch(array($key=>$entry),$array);

		}

		$entry = $this->entry($filter);

		$array = $this->create_filename($array);
		$array = $this->create_folder($array);

		if ($entry["folder"] != $array["folder"]){
			$this->filesystem->move($entry["folder"],$array["folder"]);
			parent::update(array("folder"=>$array["folder"]),array("folder"=>$entry["folder"]));
		}

		// $this->filesystem->delete($array["folder"]."/".$entry["filename"]);
		$this->filesystem->write($array["folder"]."/".$array["filename"],$array["content"]);

		unset($array["content"]);
		return parent::update($array,$where);
	}



	// Private

	private function create_filename($array) {

		if (preg_match("/(.+)\.php$/",key_of("filename",$array)))
			return $array;

		if (are_empty(key_of("filename",$array),key_of("name",$array)))
			return false;

		$array["filename"] = db_encode(either(
			key_of("filename",$array),
			key_of("name",$array)
		)).".php";

		return $array;
	}

	private function create_folder($array) {

		// if (preg_match("/(.+)\.php$/",key_of("filename",$array)))
			// return $array;

		if (are_empty(key_of("folder",$array),key_of("name",$array)))
			return false;

		$array["folder"] = db_encode(either(
			key_of("folder",$array),
			key_of("name",$array)
		));

		return $array;
	}


}


class KB_template_backend extends KB_class {


	private $core;

	function __construct($core) {
		$this->core = $core;
	}


	// Public

	public function body($set = false) {

		$body = (array) $this->private_config("body");

		if (any_empty(not_empty($set), is_array($set)))
			return $body;

		$body = $this->core->array->merge($set, $body);

		$this->private_config("body", $body);

		return $body;


	}


	public function content($add = false){
		return $this->add_html("content",$add);
	}


	public function css($filenames = "") {

		if (is_empty($filenames))
			return false;

		if (!is_array($filenames))
			$filenames = array($filenames);

		$return = [];

		foreach ($filenames as $filename) {
			$href = preg_replace(array(preg_mod($this->core->path("root")), preg_mod($this->core->path("kitbooth"))), array("/",""), $filename);

			if (preg_match("/^\//",$href))
				$href = "/".$this->core->path("url").preg_replace("/^\//","", $href);

			if (!matches("version",$href) && not_empty($age = $this->core->file->age($filename)))
				$href = $href."?version=".$age;

			$return[] = $this->core->html->tag("link",array("rel"=>"stylesheet","href"=>$href,"type"=>"text/css"));
		}

		return $this->core->html->render($return);
	}


	public function favicon($set = false) {
		return $this->set("favicon",$set);
	}


	public function footer($add = false){
		return $this->add_html("footer",$add);
	}


	public function head($add = false){
		return $this->add_html("head",$add);
	}


	public function header($add = false){
		return $this->add_html("header",$add);
	}


	public function js($filenames = "") {

		if (is_empty($filenames))
			return false;

		if (!is_array($filenames))
			$filenames = array($filenames);

		$return = [];

		foreach ($filenames as $filename) {
			$href = preg_replace(array(preg_mod($this->core->path("root")), preg_mod($this->core->path("kitbooth"))), array("/",""), $filename);

			if (preg_match("/^\//",$href))
				$href = "/".$this->core->path("url").preg_replace("/^\//","", $href);

			if (!matches("version",$href) && not_empty($age = $this->core->file->age($filename)))
				$href = $href."?version=".$age;

			$return[] = $this->core->html->tag("script",false, array("type"=>"text/javascript","src"=>$href));
		}

		return $this->core->html->render($return);
	}


	public function main_menu($add = false){
		return $this->add_html("main_menu",$add);
	}


	public function notice($add = false){
		return $this->add_html("notice", $add);
	}


	public function sub_menu($add = false){
		return $this->add_html("sub_menu",$add);
	}


	public function template($template = "",$vars = false) {

		// Make core available within included files
		$core = $this->core;

		if ($vars)
			foreach ($vars as $var_name => $var_value)
				$$var_name = $var_value;

		switch ($template) {

			case 'login':
				ob_start();
			 	include DROOT.BACKEND_TEMPLATE.'login.php';
				return ob_get_clean();

			case 'backend':

				ob_start();
				$base_dir = $this->core->filesystem->base_dir();
			 	$file = either($this->core->config("backend_template"),DROOT.BACKEND_TEMPLATE.'backend_template.php');
				$folder = preg_replace("/(.*)(\/)([^\/]+)$/","$1",$file);
				$this->core->filesystem->base_dir($folder);
				include $file;
				$backend = ob_get_clean();
				$backend = $this->parse($backend);
				$this->core->filesystem->base_dir($base_dir);
				return $backend;

			case 'home':
				ob_start();
			 	include $this->core->path("root","template").'home.php';
				return ob_get_clean();

			case 'kitgroup':
				ob_start();
			 	include DROOT.BACKEND_TEMPLATE.'kitgroup.php';
				return ob_get_clean();

			case 'db':
				ob_start();
			 	include DROOT.BACKEND_TEMPLATE.'db.php';
				return ob_get_clean();

			default:
				return "";
		}
	}


	public function title($set = false){
		return $this->set("title",$set,true);
	}


	public function touch_icon($set = false) {
		return $this->set("touch_icon",$set);
	}


	public function toolkits($add = false){
		return $this->add_html("toolkits",$add);
	}



	// Private

	private function add_html($part = "", $add = false) {

		if (is_empty($part))
			return NULL;

		$parts = (array) $this->private_config($part);

		if ($add !== false) {

			if (is_empty($parts) || !in_array($add,$parts))
				$parts[] = $add;

			$this->private_config($part, $parts);

		}

		// if (!isset($this->$part))
			// return NULL;

		return $this->private_config($part);
	}


	private function set($part = "", $set = false) {

		if (is_empty($part))
			return NULL;

		$parts = $this->private_config($part);

		if ($set !== false)
			$this->private_config($part, $set);

		// if (!isset($this->$part))
		// 	return "";

		return $this->private_config($part);
	}


	private function parse($content = "") {

		if (is_empty($content))
			return $content;

		$backend_template = BACKEND_TEMPLATE;

		if (!matches("kitbooth", key_of("REQUEST_URI", $_SERVER)))
			$backend_template = preg_replace("-kitbooth/-","",$backend_template);

		$pattern[] = "/(<link[ a-zA-Z0-9=\"\/]*href=\")([^\/]{1})/";
		$pattern[] = "/<input[ a-zA-Z0-9=\"]*src=\"/";
		$pattern[] = "/(<script[ a-zA-Z0-9=\"\/]*src=\")([^\/]{1})/";
		$pattern[] = "/(<img[^>]*src=['|\"])([^'|^\"|^\/]*)(['|\"][^>]*>)/";
		$replacement[] = '\1/'.$backend_template.'\2';
		$replacement[] = '\0/'.$backend_template;
		$replacement[] = '\1/'.$backend_template.'\2';
		$replacement[] = "\\1/".$backend_template."\\2\\3";

		$content = preg_replace($pattern, $replacement, $content);

		$pattern = [];

		$pattern[] = "/(<a[^>]*href=['|\"])([^'|^\"]*)(['|\"][^>]*>)/";
		$pattern[] = "/(<form [^>]*action=['|\"])([^'|^\"]*)(['|\"][^>]*>)/";
		$pattern[] = "/(<meta [^>]*URL=)([^'|^\"|^>]*)(['|\"][^>]*>)/";

		$that = $this;

		$content = preg_replace_callback(
					$pattern,
					function ($matches) use($that){
							return $that->replace($matches);
					},
					$content
			);

		$time_end =	explode(" ", microtime());

		if (empty($this->core->content))
			return $content;

		$content = $this->core->content->frontend->json($content);

		return $content;
	}

	// PHP 5.3 needs this to be public in order
	// to be called from $that->replace
	public function replace($matches){
		global $core;
		$b = $matches[1];
		$c = $matches[2];
		$d = $matches[3];

		if (isset($core) and !empty($c) && !preg_match("/[a-z]+\:\/\//", $c)){
				return stripslashes($b.$core->session->append_SID($c).$d);
		}
		else {
				return stripslashes($b.$c.$d);
		}
	}


}


class KB_template_frontend extends KB_data {


	private $title = "";
	private $keywords = "";
	private $updated = "";
	private $description = "";
	private $core;


	function __construct($core) {

		parent::__construct($core);
		$this->core = $core;

		$this->frontend_templates = $this->core->filesystem($this->core->path("templates"));
		$this->backend_template = $this->core->filesystem($this->core->path("backend_template"));

		$this->config("doctype", [
			'HTML5' =>	'<!DOCTYPE html>',
			'HTML 4.01 Transitional' => '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">',
			'HTML 4.01 Strict' => '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">',
			'HTML 4.01 Frameset' => '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">',
			'XHTML 1.0 Transitional' => '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
			'XHTML 1.0 Strict' => '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
			'XHTML 1.0 Frameset' =>	'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">',
			'XHTML 1.1' => '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
		]);

		$this->config("title"," • ");

		$this->config("charset", [
			'meta' => [
				"http-equiv" => "content-type",
				"content"=> "text/html; charset=utf-8"
			]
		]);

		$this->private_config(
			"template",
			// "/".$this->filesystem->folder->path($core->path("url"),"core","template")
			"/".$this->backend_template->dir()
		);
		$this->private_config(
			"template_path",
			// $this->filesystem->folder->path($core->path("root"),$core->path("core"),"template")
			$this->backend_template->dir()
		);

		// $this->core->filesystem->base_dir($this->private_config("template_path"));

		$this->config("interfaces", [
			"XML RPC" => [],

			'jQuery' => [
				"script" => [
					"type" => "text/javascript",
					"src" => $this->private_config("template")."jquery.js?version=".$this->backend_template->file->age("jquery.js")
				]
			],

			'Core' => [
				"script" => [
					"type" => "text/javascript",
					"src" => $this->private_config("template")."core.js?version=".$this->backend_template->file->age("core.js")
				]
			],

			'Zoom5' => [
				"script" => [
					"type" => "text/javascript",
					"src" => $this->private_config("template")."Zoom5/Zoom5.js?version=".$this->backend_template->file->age("Zoom5/Zoom5.js")
				],
				"link" => [
					"rel" => "stylesheet",
					"type" => "text/css",
					"href" => $this->private_config("template")."Zoom5/Zoom5.css?version=".$this->backend_template->file->age("Zoom5/Zoom5.css")
				],
			],

			'Slide5' => [
				"script" => [
					"type" => "text/javascript",
					"src" => $this->private_config("template")."Slide5/Slide5.js?version=".$this->backend_template->file->age("Slide5/Slide5.js")
				],
				"link" => [
					"rel" => "stylesheet",
					"type" => "text/css",
					"href" => $this->private_config("template")."Slide5/Slide5.css?version=".$this->backend_template->file->age("Slide5/Slide5.css")
				],
			],

			'Video5' => [
				"script" => [
					"type" => "text/javascript",
					"src" => $this->private_config("template")."Video5/Video5.js?version=".$this->backend_template->file->age("Video5/Video5.js")
				],
				"link" => [
					"rel" => "stylesheet",
					"type" => "text/css",
					"href" => $this->private_config("template")."Video5/Video5.css?version=".$this->backend_template->file->age("Video5/Video5.css")
				],
			],
		]);

	}


	public function content_tag ($template_content = "", $content = "") {

		$template_content = preg_replace("/<!-- CONTENT -->/i",$content,$template_content);
		return $template_content;

	}


	public function css ($filenames = "") {

		if (is_empty($filenames))
			return false;

		if (!is_array($filenames))
			$filenames = array($filenames);

		$return = [];

		foreach ($filenames as $filename) {
			$href = preg_replace(array(preg_mod(DROOT), preg_mod(KB_FOLDER)), array("/",""), $filename);

			if (preg_match("/^\//",$href))
				$href = "/".$this->core->path("url").preg_replace("/^\//","", $href);

			if (!matches("version",$href) && not_empty($age = either($this->core->file->age($filename),$this->core->file->age(DROOT.KB_FOLDER.$filename))))
				$href = $href."?version=".$age;

			$return[] = $this->core->html->tag("link",array("rel"=>"stylesheet","href"=>$href,"type"=>"text/css"));
		}

		return $this->core->html->render($return);

	}


	public function description ($set = NULL) {

		if (!is_null($set))
			$this->description = $set;

		return $this->description;

	}


	public function doctype_tag ($template_content = "", $template = []) {

		if (preg_match("/<!--\s?DOCTYPE\s?-->/i",$template_content,$matches)) {
			$doctype = $this->config("doctype");

			if (key_of("doctype", $template))
				$template_content = str_replace($matches[0],$doctype[$template['doctype']],$template_content);

		}

		return $template_content;

	}


	public function head ($set = "", $replace = false) {

		$header = either($this->config("head"), []);

		if ($replace === true)
			$header = [];

		if (is_empty($set))
		 	return $header;

		if (!fetch($set, $header)) {

			$header[] = $set;
			$this->config("head", $header);
		}

		return $header;

	}


	public function header_tag ($template_content = "", $template = []) {

		if (!preg_match("/<!--\s*HEADER\s*-->/i",$template_content,$matches))
			return $template_content;

		$html = $this->core->html;

		if (preg_match("/XHTML/", key_of("doctype", $template)))
			$html =  $this->core->xhtml;

		$doctype = $this->config("doctype");

		// Header Start
		$header = "";

		// Charset
		$header .= $html->render($this->config("charset"))."\n";

		// Title tag
		$header .= ifttt(
			$this->title(),
			$html->render($html->tag("title", $this->title().$this->config("title").$template['title'])).br,
			$html->render($html->tag("title", $template['title'])).br
		);

		// Favicon
		if ($template['favicon'] != "") {

			$header .= $html->render($html->tag("link", [
				"rel" => "shortcut icon",
				"href" => $template['favicon']
			])).br;
		}

		// Keywords
		if (not_empty($meta_keywords = key_of("meta_keywords", $template))) {

			$header .= $html->render($html->tag("meta", [
				"name" => "keywords",
				"content" => $meta_keywords
			])).br;
		}

		// Author
		if (not_empty($meta_author = key_of("meta_author", $template))) {

			$header .= $html->render($html->tag("meta", [
				"name" => "author",
				"content" => $meta_author
			])).br;
		}


		// Description
		if (not_empty($meta_description = $this->description())) {

			$header .= $html->render($html->tag("meta", [
				"name" => "description",
				"content" => $meta_description
			])).br;
		}

		// Date
		if (not_empty($meta_date = $this->updated())) {

			$header .= $html->render($html->tag("meta", [
				"name" => "date",
				"content" => kb_date("c", $meta_date)
			])).br;
		}


		//Viewport
		if (not_empty($viewport = key_of("viewport", $template))) {

			$header .= $html->render($html->tag("meta", [
				"name" => "viewport",
				"content" => $viewport
			])).br;
		}


		// Theme Color
		if (not_empty($color = key_of("meta_theme-color", $template))) {

			$header .= $html->render($html->tag("meta", [
				"name" => "theme-color",
				"content" => $color
			])).br;
		}

		$this->interfaces($this->core->json->decode(key_of("interfaces", $template)));

		// Interfaces
		if (not_empty($interfaces = $this->interfaces())) {

			foreach ($interfaces as $interface) {

				if (are_equal($interface, "XML RPC")) {

					foreach ($this->core->content->group->lids() as $ID) {
						if (are_equal(
							key_of("template_ID", $this->core->content->group->entry($ID)),
							key_of("ID", $template)
						)) {

							$s_id = $ID;
							$s_path = implode("/", (array) key_fetch("url_path_name", $this->core->content->group->parents($ID)));

							$header .= $html->render($html->tag("link", [
								"rel" => "EditURI",
								"type" => "application/rsd+xml",
								"title" => "RSD",
								"href" => "/xmlrpc.php?rsd&amp;s_id=$s_id&amp;s_path=$s_path",
							])).br;

							$header .= $html->render($html->tag("link", [
								"rel" => "wlwmanifest",
								"type" => "application/wlwmanifest+xml",
								"href" => "/xmlrpc.php?wlw",
							])).br;

							break;
						}
					}
				}

				else {
					$header .= $html->render(key_of($interface, $this->config("interfaces")), ["spacing"=>br]).br;
				}
			}
		}

		// Additional Headers set
		$header .= $this->core->html->render($this->head(),["spacing"=>br]);

		// Include
		if (not_empty($includes = key_of("load", $template)) && is_array($includes = $this->core->json->decode($includes))) {

			foreach ($includes as $include) {

				if (preg_match("/\.js$/",$include))
					$header .= $html->render($html->tag("script","", [
						"type" => "text/javascript",
						"src" => $include
					])).br;

				else if (preg_match("/\.css$/",$include))
					$header .= $html->render($html->tag("link", [
						"rel" => "stylesheet",
						"type" => "text/css",
						"href" => $include
					])).br;
			}
		}

		// Style

		if (not_empty($style = key_of("style", $template))) {
			$header .= $html->style($style).br;
		}

		// Head

		if (not_empty($head = key_of("head", $template))) {
			$header .= $head.br;
		}


		// Template replacements
		$template_content = str_replace($matches[0],$header,$template_content);

		return $template_content;

	}


	public function interfaces ($set = "") {

		$interfaces = (array) $this->config("interfaces_set");

		if (is_empty($set))
			return $interfaces;

		foreach ((array) $set as $interface) {

			if (!$this->core->array->match($interface, $interfaces))
				$interfaces[] = $interface;

		}

		$this->config("interfaces_set",$interfaces);
		return $interfaces;

	}


	public function js ($filenames = "") {

		if (is_empty($filenames))
			return false;

		if (!is_array($filenames))
			$filenames = array($filenames);

		$return = [];

		foreach ($filenames as $filename) {

			$href = preg_replace(array(preg_mod(DROOT), preg_mod(KB_FOLDER)), array("/",""), $filename);

			if (preg_match("/^\//",$href))
				$href = "/".$this->core->path("url").preg_replace("/^\//","", $href);

			$return[] = $this->core->html->tag("script",false,array("type"=>"text/javascript","src"=>$href."?version=".either($this->core->file->age($filename),$this->core->file->age(DROOT.KB_FOLDER.$filename))));
		}

		return $return;
	}


	public function keywords ($set = NULL) {

		if (!is_null($set))
			$this->keywords = $set;

		return $this->keywords;

	}


	public function logo_tag ($template_content = "", $template = []) {

		$doctype = $this->config("doctype");

		if (preg_match("/<!--\s?logo\s?-->/i",$template_content,$matches)) {

			$output = $this->core->html;

			if (isset($doctype) && preg_match('/XHTML/',$template['logo']))
				$output = $this->core->xhtml;

			$img = $output->img($template['logo'], array(
				"class" => "logo",
				"alt" => key_of("title", $template),
			));

			if ($template['logo'] != "")
				$template_content = str_replace($matches[0],$img,$template_content);
		}

		return $template_content;

	}


	public function message ($set = false) {

		$message = $this->private_config("message");

		if (is_empty($set))
			return $message;

		if (is_string($set))
			return key_of($set, $message);

		$this->private_config("message", $message);
	}


	public function render ($content = "", $IDs = []) {

		if (is_empty($IDs))
			return $content;

		if (is_string($IDs))
			$IDs = array($IDs);

		if (!is_array($IDs))
			return $content;

		$return = "";

		$content = $this->config("content").$content;

		// Make $content available

		$this->config("content",$content);

		foreach ($IDs as $ID) {

			if (is_empty($ID))
				continue;

			$base_dir = $this->core->filesystem->base_dir();
			$template = $this->core->template->entry($ID);

			if (is_empty($template))
				continue;

			$this->core->filesystem->base_dir($this->core->path("root", "templates").$template["folder"]);
			$template_content = $this->core->file->read_php(
				$this->core->folder->path($this->core->path("root", "templates"),$template["folder"]).$template["filename"]
			);

			$this->core->filesystem->base_dir($base_dir);

			if (is_empty($template_content))
				continue;

			// Replace Blog Tags in templates with requested menus.
			$template_content = $this->core->content->frontend->blog_tags($template_content);
			// Replace Kit Tags with requested Kits
			$template_content = $this->core->content->frontend->kit_tags($template_content);
			// Built Template with execution of included JSON
			$template_content = $this->core->content->frontend->json($template_content);
			// Replace Menu Tags with requested menus.
			$template_content = $this->core->content->frontend->menu_tags($template_content);
			// Replace Content Tags with requested menus.
			$template_content = $this->core->content->frontend->content_tags($template_content);
			// Replace Doctype Tags
			$template_content = $this->doctype_tag($template_content, $template);
			// Replace Header Tags
			$template_content = $this->header_tag($template_content, $template);
			// Replace Logo Tags
			$template_content = $this->logo_tag($template_content, $template);
			// Replace Paths
			$template_content = $this->replace_paths($template_content, $template);
			// Replace Content Tags
			$content = $this->content_tag($template_content, $content);
			// Replace Title Tags
			$content = $this->title_tag($content);
			// Make new $content available
			$this->config("content",$content);

		}

		return $content;

	}


	public function replace ($i, $start, $search, $template_name = "") {

		$core = $this->core;

		$search =  stripslashes($search);
		$replace_callback = array(

			4 => array(
				"/((href|action){1}=['|\"])(\/[^'|^\"]*)(['|\"]{1}[^>]*)/Ui"
				=> function ($match) use($core, $template_name) {return $core->content->frontend->replace_content($match[1], $match[3], $match[4]);}
			),
			5 => array(
				"/((background|href|poster|src|url|name=['|\"]?movie['|\"]? value|name=['|\"]?src['|\"]? value){1}=['|\"])([^'|^\"]*)(['|\"]{1}[^>]*)/Ui"
				=> function ($match) use($core, $template_name) {return $core->template->frontend->replace_template($match[1], $match[3], $match[4], $template_name);}
			),
			7 => array(
				"/((url\(){1}['|\"]?)([^'|^\"^\)]*)(['|\"]?\)[^>]*)/Ui"
				=> function ($match) use($core, $template_name) {return $core->template->frontend->replace_template($match[1], $match[3], $match[4], $template_name);}
			),
			9 => array(
				"/((src|url)\([|'|\"]?)([^'|^\"|\)]*)(['|\"]?\))/Ui"
				=> function ($match) use($core, $template_name) {return $core->template->frontend->replace_template($match[1], $match[3], $match[4], $template_name);}
			),

		);


		$search = preg_replace_callback_array(
			$replace_callback[$i],
			$search
		);

		return $start.$search;


	}


	public function replace_paths ($content = "", $template = []) {

		//  TEMPLATE REPLACE FUNCTIONS
		//	Add template path to to all html elements, except anchors and headers, containing specific keywords in contents.
		//	Add default structures and base directories to links.
		//	Ignore all text outside html tags.
		//	Include relative links to all but anchors and headers, ignore relative links in anchors and headers which include special links like mailto:.
		//	Case insensitive.

		$core = $this->core;

		$replace_callback = array(
			"/(<[^a])([^>]+)/i"
			=>	function($matches) use($core, $template) { return $core->template->frontend->replace(5, $matches[1],$matches[2], $template["folder"]);},

			"/(<(a|form) )([^>]+)/i"
			=>	function($matches) use($core, $template) { return $core->template->frontend->replace(4, $matches[1],$matches[3]);},
			// "/(<(a|form) )([^>]+)/i"
			// =>	function($matches){ return replace(1, $matches[1], $matches[3]);},
			"/(<\b\w+\b )([^>]+)/i"
			=>	function($matches) use($core, $template) { return $core->template->frontend->replace(7, $matches[1],$matches[2], $template["folder"]);},
			"/(header\(['|\"]?Location:\s?)([^\)|^'|\"]+)/i"
			=>	function($matches) use($core, $template) { return $core->template->frontend->replace(8, $matches[1], $matches[2]);},
			"/(<style)([^>]*>)([^<]+)/i"
			=>	function($matches) use($core, $template) { return $core->template->frontend->replace(9, $matches[1].$matches[2],$matches[3], $template["folder"]);},
		);

		$content = preg_replace_callback_array($replace_callback, $content);
		return $content;

	}


	public function replace_template ($b, $c, $d, $folder) {

		$b =  stripslashes($b);
		$d =  stripslashes($d);

		if (substr($c, 0, 7) != "http://" && substr($c, 0, 8) != "https://" && substr($c, 0, 5) != "/kits" && substr($c, 0, 1) != "/") {

			$version = "";

			if (!matches("version=", $c)) {

				if (preg_match("-\?-",$c)) {
					$version = preg_split("-\?-",$c);
					$version = array_slice($version,0,-1);
					$version = implode("",$version);
					$version = "&version=".$this->core->file->age($this->core->path("root","templates").$folder."/".$version);
				}

				else {
					$version = "?version=".$this->core->file->age($this->core->path("root","templates").$folder."/".$c);
				}
			}

			return $b."/".$this->core->path("url")."templates/".$folder."/".$c.$version.$d;

		}

		else {
			return $b.$c.$d;
		}

	}


	public function title ($set = NULL) {

		if (!is_null($set))
			$this->title = $set;

		return $this->title;

	}


	public function title_tag ($template_content = "") {

		if (preg_match("/<!--\s?title\s?-->/i",$template_content,$matches)) {

			$template_content = str_replace($matches[0],$this->title(),$template_content);

		}

		return $template_content;

	}


	public function updated ($set = NULL) {

		if (!is_null($set))
			$this->updated = $set;

		return $this->updated;

	}



	// ALIAS

	public function header($set = "") {
		return $this->head($set);
	}

}


class KB_template_kit extends KB_class {


	private $core;

	public $panel;


	function __construct($core) {
		$this->core = $core;
		$this->panel = new KB_template_panel($core);
	}


	public function button($icon = "", $label = "", $options = []) {

		$args = func_get_args();

		if (count($args) === 2 && is_array($args[1])) {
			$icon = "";
			$label = $args[0];
			$options = $args[1];
		}

		$options["class"] = trim("button ".either(key_of("class", $options), ""));

		$color = NULL;

		if (is_indexed($icon, true)) {
			$color = key_of(1, $icon);
			$icon = key_of(0, $icon);
		}

		if (not_empty($icon))
			$icon = $this->core->template->icon($icon, $color);

		if (is_empty($label))
			return $this->core->html->a($icon, $options);

		if (is_string($label))
			$label = $this->core->language->translate($label);

		if (not_empty($icon))
			$label = $this->core->html->tag("span", $label);

		return $this->core->html->a(array($icon, $label), $options);

	}


	public function content ($title = false, $body = [], $options = []) {

		if (neither($title, $body))
			return "";

		if (not_empty($title)) {

			if (is_string($title))
				$title = $this->core->html->tag("h3",$this->core->language->translate($title));

			$title = $this->core->html->tag("div",$title,array("class"=>"section".ifttt($class = key_of("class",key_of("title",$options))," $class")));
		}

		if (not_empty($body)) {
			$body = $this->core->html->tag("div", (array) $body,array("class"=>"section"));
		}

		$markup = $this->core->html->tag("div",array($title,$body),array("class"=>"kitcontent"));

		return $this->core->html->render($markup);

	}


	public function menu ($menu, $options = []) {

		$array = [];

		foreach ($menu as $get_var => $entry) {

			$get_var = ifttt (is_numeric ($get_var), $entry, $get_var);

			$url_options = either (key_of ("url_options",$entry), []);

			$item = array(
				"a" => array(
					"class" => ifttt(
						either(
							are_equal(
								$this->core->get("view"),
								either(key_of("view", $url_options), $get_var)
							),
							are_empty(
								$this->core->get("view"),
								either(key_of("view", $url_options), $get_var)
							)
						),
						key_of("class", $entry)." active button",
						key_of("class", $entry)." button"
					),
					"href" => $this->core->url(["kit" => $this->core->get("kit"), "view" => either(key_of("view", $url_options), $get_var)] + $url_options, true),
					$this->core->language->translate(either(key_of("title", $entry), ifttt(is_string($entry), $entry, $get_var)))
				)
			);
			$array[] = $item;
		}
		/*
		$array = array(
			"div" => array(
				"class" => "section",
				$array
			)
		);
		*/
		return $array;
	}


	public function panel($title = false, $body = [], $footer = [], $options = []) {

		$options = either($options, $this->core->template->kit->panel->options());

		$options = $this->core->array->merge(
			array(
				"wrap_inner" => false,
				"wrap_outer" => false,
			),$options
		);

		$title = either($title, $this->core->template->kit->panel->title());

		if (is_string($title))
			$title = array(
				"h3" => array(
					"class" => "section",
					$this->core->language->translate($title)
				)
			);

		$body = either($body ,$this->core->template->kit->panel->body());

		$body = array(
			"div" => array(
				"class" => "section scroll",
				$body
			)
		);

		$footer = either($footer ,$this->core->template->kit->panel->footer());

		if (not_empty($footer))
		$footer = array(
			"div" => array(
				"class" => "section".ifttt($class = key_of("class",key_of("footer",$options))," $class"),
				// $this->core->language->translate($footer)
				$footer
			)
		);

		$markup = array(
			"div" => array (
				"class" => "panel".ifttt($class = key_of("class",key_of("panel",$options))," $class"),
				$title,
				$body,
				$footer
			)
		);

		if ($form = key_of("wrap_form", $options)) {

			if (is_bool($form)) {
				$markup = $this->core->html->tag("form",$markup, array("method"=>"post"));
			}

			else if (is_object($form) && $form instanceof KB_form) {
				if (fetch(array("type"=>$form->config("file_fields")), $form->element))
					$markup = $this->core->html->tag("form",$markup, array("method"=>"post","enctype"=>"multipart/form-data"));
				else
					$markup = $this->core->html->tag("form",$markup, array("method"=>"post"));
			}
		}

		return $this->core->html->render($markup);

	}


	public function panes($pane = "", $content = "", $options = []) {

		$panes = [];

		if (not_empty($pane)) {

			if (!is_indexed($pane))
				$pane = array($pane);

			foreach ($pane as $index => $entry) {

				$panes[] = $this->core->html->tag("div", array($entry), array(
					"class" => "pane pane$index". ifttt($class = key_of("class", key_of("pane", $options)), " $class"),
				));
			}

		}

		$contents = [];

		if (not_empty($content)) {

			if (!is_indexed($content))
				$content = array($content);

			foreach ($content as $index => $entry) {

				$contents[] = $this->core->html->tag("div", array($entry), array(
					"class" =>"main main$index".ifttt($class = key_of("class", key_of("main", $options)), " $class"),
				));
			}

		}

		if (key_of("switch", $options))
			$array = $this->core->array->merge($contents, $panes);
		else
			$array = $this->core->array->merge($panes, $contents);


		$array = $this->core->html->tag("div", $array, array(
			"class" => "panes",
		));

		return $this->core->html->render($array);
	}


	public function small_button($icon = "", $label = "", $options = []) {

		$args = func_get_args();

		if (count($args) === 2 && is_array($args[1])) {
			$icon = "";
			$label = $args[0];
			$options = $args[1];
		}

		$options["class"] = trim("small ".either(key_of("class", $options), ""));

		return $this->button($icon, $label, $options);

	}


	public function table_menu($menu = false, $additional_markup = [], $options = []) {

		if (are_empty($menu, $additional_markup))
			return "";

		if (!is_array($additional_markup))
			$additional_markup = array($additional_markup);

		$markup = [];

		foreach ($menu as $get_var => $entry) {

			$title = key_of("title", $entry);

			if (is_string($title))
				$title = ["",$title];

			$title[1] = $this->core->language->translate(key_of("1", $title));

			$item = $this->button(
				key_of(0,$title),
				key_of(1,$title),
				[
					"class" => key_of("class",$entry),
					"href" => $this->core->url(
						either(
							key_of("href",$entry),
							$this->core->array->merge(
								array("do"=>$get_var),
								key_of("url_options",$entry)
							)
						)
					),
				]
			);

			/*
			$item = array(
				"a" => array(
					"class" => either(key_fetch("class",$entry), "button"),
					"href" => $this->core->url(
						either(
							key_of("href",$entry),
							$this->core->array->merge(
								array("do"=>$get_var),
								key_of("url_options",$entry)
							)
						)
					),
					$this->core->language->translate(key_of("title", $entry))
				)
			);
			*/
			$markup[] = $item;
		}

		$markup = array(
			"div" => array(
				"class" => "kit_menu",
				array_merge($markup,$additional_markup)
			)
		);

		return $this->core->html->render($markup);
	}


	public function tabs($array = []) {

		if (is_empty($array))
			return $array;

		$identifier = debug_backtrace();
		$identifier = "i".md5($identifier[0]["file"].$identifier[0]["line"]);

		$tabs = [];
		$content = [];

		// Backwards compatibility

		if (is_indexed($indexed = $array)) {

			$array = [];

			foreach ($indexed as $entry) {

				if (any_empty(key_of(0,$entry), key_of(1,$entry)))
					continue;

				$array[key_of(0,$entry)] = key_of(1,$entry);

			}
		}

		if (is_empty($array))
			return $array;

		foreach ($array as $key => $tab) {

			if (are_empty($key,$tab))
				continue;

			$tabs[] = array(
				"div" => array(
					"class" => "tab",
					"data-id" => $identifier.db_encode($key),
					$this->core->language->translate($key)
				)
			);

			$content[] = array(
				"div" => array(
					"class" => "content",
					"data-id" => $identifier.db_encode($key),
					$tab
				)
			);

		}

		return $this->core->html->render(
			array(
				"div" => array(
					"class" => "tabbed_content",
					array(
						"div" => array(
							"class" => "tabs",
							"data-id" => $identifier,
							$tabs
						)
					),
					$content
				)
			)
		);
	}


}


class KB_template_panel extends KB_class {


	private $core;
	private $title = "";
	private $body = "";
	private $footer = "";
	private $options = "";


	function __construct($core) {
		$this->core = $core;
	}


	public function body($body = false) {

		if ($body === false) {
			$body = $this->body;
			$this->body = "";
			return $body;
		}

		if (is_array($body))
			$body = $this->core->html->render($body);

		$this->body = $this->body.$body;

		return;

	}


	public function button($button = false) {

		if ($buttons === false) {
			$buttons = $this->buttons;
			$this->buttons = "";
			return $buttons;
		}

		// if (is_array($buttons))
			// $buttons = $this->core->html->render($body);

		// $this->body = $this->body.$body;


		return;


	}


	public function buttons($buttons = false) {

		if ($buttons === false) {
			$buttons = $this->buttons;
			$this->buttons = "";
			return $buttons;
		}

		// if (is_array($buttons))
			// $buttons = $this->core->html->render($body);

		// $this->body = $this->body.$body;


		return;


	}


	public function footer($footer = false) {

		if ($footer === false) {
			$footer = $this->footer;
			$this->footer = "";
			return $footer;
		}


		if (is_array($footer))
			$footer = $this->core->html->render($footer);

		$this->footer = $this->footer.$footer;

		return;

	}


	public function menu($menu = []) {

	}


	public function options($options = false) {

		if ($options === false) {
			$options = $this->options;
			$this->options = "";
			return $options;
		}

		if (!is_array($options) || is_indexed($options))
			return;

		$this->options = $this->core->array->merge(
			$this->options,
			$options
		);

	}


	public function title($title = false) {

		if ($title === false) {
			$title = $this->title;
			$this->title = "";
			return $title;
		}

		if (is_array($title))
			$title = $this->core->html->render($title);

		$this->title = $this->title.$title;

		return;

	}


}


class KB_template_icon extends KB_class {


	private $icons = [];
	private $core;


	function __construct($core) {
		$this->core = $core;
	}


	public function color ($color = NULL) {

		deprecated();

		if (not_empty($color))
			$this->private_config("color", $color);

		return $this->private_config("color");

	}


	public function get ($name = false, $color = "") {

		if (any_empty($name, is_string($name)))
			return "";

		$icon = key_of($name, $this->icons);

		if (is_empty(key_of("source", $icon))) {
			$fs = $this->core->filesystem($this->core->path("root"));
			$this->icons[$name]["source"] = $fs->file->read(key_of("path", $icon));
			$icon = key_of($name, $this->icons);
		}

		return key_of("source", $icon);

	}


	public function path ($icon=false) {

		if (any_empty($icon, is_string($icon)))
			return "";

		$icon = key_of($icon, $this->icons);

		return key_of("path", $icon);

	}


	public function set ($name = "", $content = "") {

		if (any_empty($name, $content))
			return false;

		$this->icons[$name] = [
			"source" => $content,
			"path" => "",
			];
	}


	public function set_folder ($path="") {

		if (is_empty($path))
			return false;

		$filesystem = $this->core->filesystem($this->core->path("root"));

		// $this->core->filesystem->dir($this->core->path("root"));

		foreach ($filesystem->read($path) as $file)
			$this->set_path($filesystem->file->basename($file), $filesystem->folder->path($path).$file);

		return;
	}


	public function set_path ($name = "", $path = "") {

		if (any_empty($name, $path))
			return false;

		$this->icons[$name] = [
			"path" => $path,
			"source" => "",
			];
	}


}
