<?php

class KB_video extends KB_class {

	private $core;

	function __construct ($core) {

		$this->core = $core;

	}


	public function image($source = "", $target_folder = "", $options = array()) {

		// if (any_empty($source, $target_folder))
		if (is_empty($source))
			return false;

		if (!$this->core->array->match(
			$this->core->file->extension($source),
			array("mp4", "mov")
		))
			return false;

		$options = $this->core->array->merge(
			array(
				"filetype" => "jpg",
				"index" => false,
				"size" => 1000,
				"frame" => 0,
				"frame_percent" => false,
				"modified" => false,
			),
			$options
		);

		if (any_empty(
			loaded("ffmpeg")
			))
			return false;

		if ($this->core->file->exists($source))
			$source = $this->core->file->get_path($source);

		else if ($this->core->file->exists($this->core->folder->path(DROOT,FILES_FOLDER,$this->core->folder->part($source)).$this->core->file->name($source)))
			$source = $this->core->folder->path(DROOT,FILES_FOLDER,$this->core->folder->part($source)).$this->core->file->name($source);

		else if ($this->core->file->exists($this->core->folder->path(DROOT,KB_FOLDER,$this->core->folder->part($source)).$this->core->file->name($source)))
			$source = $this->core->folder->path(DROOT,KB_FOLDER,$this->core->folder->part($source)).$this->core->file->name($source);

		else
			return $source;

		if (is_array($size = key_of("size", $options)))
			list($size_w, $size_h) = $size;

		else
			list($size_w, $size_h) = array($size, $size);


		// $this->core->filesystem->create_folder($target_folder);

		$target_file = $this->core->file->basename($source).".".key_of("filetype",$options);

		$target = $this->core->folder->path($this->core->folder->part($source), $target_folder).$target_file;

		if (
			$this->core->file->exists($target)
			&& ($this->core->file->age($source) <= $this->core->file->age($target, array("age"=>"modified")))
			)
			return $target_file;

		$file = new FFmpegMovie($source);

		if ($frame = key_of("frame", $options)) {
			$frame = int($frame);
			$frame = $file->getFrame($frame);
		}

		else if ($frame_percent = key_of("frame_percent", $options)) {
			$frame_percent = ((float) $frame_percent)/100;
			$frame = $file->getFrameCount();
			$frame = floor($frame*$frame_percent);
			$frame = $file->getFrame($frame);

		}

		else {

			$frame = $file->getDuration();
			$frame = floor($frame/2);

			try {

				$frame = $file->getFrameAtTime($frame);
			}

			catch (Exception $e) {

				try {

					$frame = $file->getFrame(1);
				}

				catch (Exception $e) {
					return false;
				}
			}

		}

		if (!$frame)
			return false;

		if (($crop = key_of("crop_percent", $options)) && is_indexed($crop, true)) {

			$crop_left = (float) key_of(0, $crop) / 100 * $frame->getWidth();
			$crop_top = (float) key_of(1, $crop) / 100 * $frame->getHeight();
			$crop_right = (float) key_of(2, $crop) / 100 * $frame->getWidth();
			$crop_bottom = (float) key_of(3, $crop) / 100 * $frame->getHeight();

			$crop_left = floor($crop_left);
			$crop_top = floor($crop_top);
			$crop_right = floor($crop_right);
			$crop_bottom = floor($crop_bottom);

			if ($crop_left % 2 == 1) $crop_left--;
			if ($crop_top % 2 == 1) $crop_top--;
			if ($crop_right % 2 == 1) $crop_right--;
			if ($crop_bottom % 2 == 1) $crop_bottom--;

			$frame->crop($size_w, $size_h, $crop_top, $crop_bottom, $crop_left, $crop_right);

		}
		else {

			$frame->resize($size_w, $size_h, true);
		}

		if (!$this->core->file->write($target, $frame->to_jpg()))
			return false;

		if ($filemtime = key_of("modified", $options))
			$this->core->file->touch($target, $filemtime);

		return $target_file;

	}


}


/**
 * FFmpegMovie represents a movie file
 *
 * @author char0n (Vladimír Gorej, vladimir.gorej@gmail.com)
 * @package FFmpegPHP
 * @license New BSD
 * @version 2.6
 */
// class FFmpegMovie implements Serializable {
class FFmpegMovie {

    protected static $REGEX_DURATION            = '/Duration: ([0-9]{2}):([0-9]{2}):([0-9]{2})(\.([0-9]+))?/';
    protected static $REGEX_FRAME_RATE          = '/([0-9\.]+\sfps,\s)?([0-9\.]+)\stb(?:r|n)/';
    protected static $REGEX_COMMENT             = '/comment\s*(:|=)\s*(.+)/i';
    protected static $REGEX_TITLE               = '/title\s*(:|=)\s*(.+)/i';
    protected static $REGEX_ARTIST              = '/(artist|author)\s*(:|=)\s*(.+)/i';
    protected static $REGEX_COPYRIGHT           = '/copyright\s*(:|=)\s*(.+)/i';
    protected static $REGEX_GENRE               = '/genre\s*(:|=)\s*(.+)/i';
    protected static $REGEX_TRACK_NUMBER        = '/track\s*(:|=)\s*(.+)/i';
    protected static $REGEX_YEAR                = '/year\s*(:|=)\s*(.+)/i';
    protected static $REGEX_FRAME_WH            = '/Video:.+?([1-9][0-9]*)x([1-9][0-9]*)/';
    protected static $REGEX_PIXEL_FORMAT        = '/Video: [^,]+, ([^,]+)/';
    protected static $REGEX_PIXEL_ASPECT_RATIO  = '/Video:.+DAR ([0-9]+):([0-9]+)\]/';
    protected static $REGEX_BITRATE             = '/bitrate: ([0-9]+) kb\/s/';
    protected static $REGEX_VIDEO_BITRATE       = '/Video:.+?([0-9]+) kb\/s/';
    protected static $REGEX_AUDIO_BITRATE       = '/Audio:.+?([0-9]+) kb\/s/';
    protected static $REGEX_AUDIO_SAMPLE_RATE   = '/Audio:.+?([0-9]+) Hz/';
    protected static $REGEX_VIDEO_CODEC         = '/Video:\s([^,]+),/';
    protected static $REGEX_AUDIO_CODEC         = '/Audio:\s([^,]+),/';
    protected static $REGEX_AUDIO_CHANNELS      = '/Audio:\s[^,]+,[^,]+,([^,]+)/';
    protected static $REGEX_HAS_AUDIO           = '/Stream.+Audio/';
    protected static $REGEX_HAS_VIDEO           = '/Stream.+Video/';
    protected static $REGEX_ERRORS              = '/.*(Error|Permission denied|could not seek to position|Invalid pixel format|Unknown encoder|could not find codec|does not contain any stream).*/i';

    /**
     * FFmpeg binary
     *
     * @var string
     */
    protected $ffmpegBinary;
    /**
     * Output provider
     *
     * @var OutputProvider
     */
    protected $provider;

    /**
    * Movie file path
    *
    * @var string
    */
    protected $movieFile;

    /**
    * provider output
    *
    * @var string
    */
    protected $output;

    /**
    * Movie duration in seconds
    *
    * @var float
    */
    protected $duration;
    /**
    * Current frame index
    *
    * @var int
    */
    protected $frameCount;
    /**
    * Movie frame rate
    *
    * @var float
    */
    protected $frameRate;
    /**
    * Comment ID3 field
    *
    * @var string
    */
    protected $comment;
    /**
    * Title ID3 field
    *
    * @var string
    */
    protected $title;
    /**
    * Author ID3 field
    *
    * @var string
    */
    protected $artist;
    /**
    * Copyright ID3 field
    *
    * @var string
    */
    protected $copyright;
    /**
    * Genre ID3 field
    *
    * @var string
    */
    protected $genre;
    /**
    * Track ID3 field
    *
    * @var int
    */
    protected $trackNumber;
    /**
    * Year ID3 field
    *
    * @var int
    */
    protected $year;
    /**
    * Movie frame height
    *
    * @var int
    */
    protected $frameHeight;
    /**
    * Movie frame width
    *
    * @var int
    */
    protected $frameWidth;
    /**
    * Movie pixel format
    *
    * @var string
    */
    protected $pixelFormat;
    /**
    * Movie pixel aspect ratio
    *
    * @var float
    */
    protected $pixelAspectRatio;
    /**
    * Movie bit rate combined with audio bit rate
    *
    * @var int
    */
    protected $bitRate;
    /**
    * Movie video stream bit rate
    *
    * @var int
    */
    protected $videoBitRate;
    /**
    * Movie audio stream bit rate
    *
    * @var int
    */
    protected $audioBitRate;
    /**
    * Audio sample rate
    *
    * @var int
    */
    protected $audioSampleRate;
    /**
    * Current frame number
    *
    * @var int
    */
    protected $frameNumber;
    /**
    * Movie video cocec
    *
    * @var string
    */
    protected $videoCodec;
    /**
    * Movie audio coded
    *
    * @var string
    */
    protected $audioCodec;
    /**
    * Movie audio channels
    *
    * @var int
    */
    protected $audioChannels;

    /**
    * Open a video or audio file and return it as an FFmpegMovie object.
    * If ffmpeg and ffprobe are both installed on host system, ffmpeg
    * gets priority in extracting info from the movie file. However
    * to override this default behaviour use any implementation of OutputProvider interface
    * as the second constructor argumentwhile instantiating
    *
    *
    * @param string $moviePath full path to the movie file
    * @param OutputProvider $outputProvider provides parsable output
    * @param string $ffmpegBinary ffmpeg executable, if $outputProvider not specified
    * @throws Exception
    */
    public function __construct($moviePath, OutputProvider $outputProvider = null, $ffmpegBinary = 'ffmpeg') {
        $this->movieFile       = $moviePath;
        $this->frameNumber     = 0;
        $this->ffmpegBinary    = $ffmpegBinary;
        if ($outputProvider === null) {
            $outputProvider = new FFmpegOutputProvider($ffmpegBinary);
        }
        $this->setProvider($outputProvider);
    }

    /**
     * Setting provider implementation
     *
     * @param OutputProvider $outputProvider
     */
    public function setProvider(OutputProvider $outputProvider) {
        $this->provider = $outputProvider;
        $this->provider->setMovieFile($this->movieFile);
        $this->output = $this->provider->getOutput();
    }

    /**
     * Getting current provider implementation
     *
     * @return OutputProvider
     */
    public function getProvider() {
        return $this->provider;
    }

    /**
    * Return the duration of a movie or audio file in seconds.
    *
    * @return float movie duration in seconds
    */
    public function getDuration() {
        if ($this->duration === null) {
            $match = array();
            preg_match(self::$REGEX_DURATION, $this->output, $match);
            if (array_key_exists(1, $match) && array_key_exists(2, $match) && array_key_exists(3, $match)) {
                $hours     = (int)    $match[1];
                $minutes   = (int)    $match[2];
                $seconds   = (int)    $match[3];
                $fractions = (float)  ((array_key_exists(5, $match)) ? "0.$match[5]" : 0.0);

                $this->duration = (($hours * (3600)) + ($minutes * 60) + $seconds + $fractions);
            } else {
                $this->duration = 0.0;
            }

            return $this->duration;
        }

        return $this->duration;
    }

    /**
    * Return the number of frames in a movie or audio file.
    *
    * @return int
    */
    public function getFrameCount() {
        if ($this->frameCount === null) {
            $this->frameCount = (int) ($this->getDuration() * $this->getFrameRate());
        }

        return $this->frameCount;
    }

    /**
    * Return the frame rate of a movie in fps.
    *
    * @return float
    */
    public function getFrameRate() {
        if ($this->frameRate === null) {
            $match = array();
            preg_match(self::$REGEX_FRAME_RATE, $this->output, $match);
            $this->frameRate = (float) ((array_key_exists(1, $match)) ? $match[1] : 0.0);
        }

        return $this->frameRate;
    }

    /**
    * Return the path and name of the movie file or audio file.
    *
    * @return string
    */
    public function getFilename() {
        return $this->movieFile;
    }

    /**
    * Return the comment field from the movie or audio file.
    *
    * @return string
    */
    public function getComment() {
        if ($this->comment === null) {
             $match = array();
             preg_match(self::$REGEX_COMMENT, $this->output, $match);
             $this->comment = (array_key_exists(2, $match)) ? trim($match[2]) : '';
        }

        return $this->comment;
    }

    /**
    * Return the title field from the movie or audio file.
    *
    * @return string
    */
    public function getTitle() {
        if ($this->title === null) {
            $match = array();
            preg_match(self::$REGEX_TITLE, $this->output, $match);
            $this->title = (array_key_exists(2, $match)) ? trim($match[2]) : '';
        }

        return $this->title;
    }

    /**
    * Return the author field from the movie or the artist ID3 field from an mp3 file; alias $movie->getArtist()
    *
    * @return string
    */
    public function getArtist() {
        if ($this->artist === null) {
            $match = array();
            preg_match(self::$REGEX_ARTIST, $this->output, $match);
            $this->artist = (array_key_exists(3, $match)) ? trim($match[3]) : '';
        }

        return $this->artist;
    }

    /**
    * Return the author field from the movie or the artist ID3 field from an mp3 file.
    *
    * @return string
    */
    public function getAuthor() {
        return $this->getArtist();
    }

    /**
    * Return the copyright field from the movie or audio file.
    *
    * @return string
    */
    public function getCopyright() {
        if ($this->copyright === null) {
            $match = array();
            preg_match(self::$REGEX_COPYRIGHT, $this->output, $match);
            $this->copyright = (array_key_exists(2, $match)) ? trim($match[2]) : '';
        }

        return $this->copyright;
    }

    /**
    * Return the genre ID3 field from an mp3 file.
    *
    * @return string
    */
    public function getGenre() {
        if ($this->genre === null) {
            $match = array();
            preg_match(self::$REGEX_GENRE, $this->output, $match);
            $this->genre = (array_key_exists(2, $match)) ? trim($match[2]) : '';
        }

        return $this->genre;
    }

    /**
    * Return the track ID3 field from an mp3 file.
    *
    * @return int
    */
    public function getTrackNumber() {
        if ($this->trackNumber === null) {
            $match = array();
            preg_match(self::$REGEX_TRACK_NUMBER, $this->output, $match);
            $this->trackNumber = (int) ((array_key_exists(2, $match)) ? $match[2] : 0);
        }

        return $this->trackNumber;
    }

    /**
    * Return the year ID3 field from an mp3 file.
    *
    * @return int
    */
    public function getYear() {
        if ($this->year === null) {
            $match = array();
            preg_match(self::$REGEX_YEAR, $this->output, $match);
	    $this->year = (int) ((array_key_exists(2, $match)) ? $match[2] : 0);
        }

        return $this->year;
    }

    /**
    * Return the height of the movie in pixels.
    *
    * @return int
    */
    public function getFrameHeight() {
        if ($this->frameHeight == null) {
            $match = array();
            preg_match(self::$REGEX_FRAME_WH, $this->output, $match);
            if (array_key_exists(1, $match) && array_key_exists(2, $match)) {
                $this->frameWidth  = (int) $match[1];
                $this->frameHeight = (int) $match[2];
            } else {
                $this->frameWidth  = 0;
                $this->frameHeight = 0;
            }
        }

        return $this->frameHeight;
    }

    /**
    * Return the width of the movie in pixels.
    *
    * @return int
    */
    public function getFrameWidth() {
        if ($this->frameWidth === null) {
            $this->getFrameHeight();
        }

        return $this->frameWidth;
    }

    /**
    * Return the pixel format of the movie.
    *
    * @return string
    */
    public function getPixelFormat() {
        if ($this->pixelFormat === null) {
            $match = array();
            preg_match(self::$REGEX_PIXEL_FORMAT, $this->output, $match);
            $this->pixelFormat = (array_key_exists(1, $match)) ? trim($match[1]) : '';
        }

        return $this->pixelFormat;
    }

    /**
    * Return the pixel aspect ratio of the movie.
    *
    * @return float
    */
    public function getPixelAspectRatio() {
        if ($this->pixelAspectRatio === null) {
            $match = array();
            preg_match(self::$REGEX_PIXEL_ASPECT_RATIO, $this->output, $match);
            $this->pixelAspectRatio = (array_key_exists(1, $match) && array_key_exists(2, $match)) ? ($match[1] / $match[2]) : 0;
        }

        return $this->pixelAspectRatio;
    }

    /**
    * Return the bit rate of the movie or audio file in bits per second.
    *
    * @return int
    */
    public function getBitRate() {
        if ($this->bitRate === null) {
            $match = array();
            preg_match(self::$REGEX_BITRATE, $this->output, $match);
            $this->bitRate = (int) ((array_key_exists(1, $match)) ? ($match[1] * 1000) : 0);
        }

        return $this->bitRate;
    }

    /**
    * Return the bit rate of the video in bits per second.
    *
    * NOTE: This only works for files with constant bit rate.
    *
    * @return int
    */
    public function getVideoBitRate() {
        if ($this->videoBitRate === null) {
            $match = array();
            preg_match(self::$REGEX_VIDEO_BITRATE, $this->output, $match);
            $this->videoBitRate = (int) ((array_key_exists(1, $match)) ? ($match[1] * 1000) : 0);
        }

        return $this->videoBitRate;
    }

    /**
    * Return the audio bit rate of the media file in bits per second.
    *
    * @return int
    */
    public function getAudioBitRate() {
        if ($this->audioBitRate === null) {
            $match = array();
            preg_match(self::$REGEX_AUDIO_BITRATE, $this->output, $match);
            $this->audioBitRate = (int) ((array_key_exists(1, $match)) ? ($match[1] * 1000) : 0);
        }

        return $this->audioBitRate;
    }

    /**
    * Return the audio sample rate of the media file in bits per second.
    *
    * @return int
    */
    public function getAudioSampleRate() {
        if ($this->audioSampleRate === null) {
            $match = array();
            preg_match(self::$REGEX_AUDIO_SAMPLE_RATE, $this->output, $match);
            $this->audioSampleRate = (int) ((array_key_exists(1, $match)) ? $match[1] : 0);
        }

        return $this->audioSampleRate;
    }

    /**
    * Return the current frame index.
    *
    * @return int
    */
    public function getFrameNumber() {
        return ($this->frameNumber == 0) ? 1 : $this->frameNumber;
    }

    /**
    * Return the name of the video codec used to encode this movie as a string.
    *
    * @return string
    */
    public function getVideoCodec() {
        if ($this->videoCodec === null) {
            $match = array();
            preg_match(self::$REGEX_VIDEO_CODEC, $this->output, $match);
            $this->videoCodec = (array_key_exists(1, $match)) ? trim($match[1]) : '';
        }

        return $this->videoCodec;
    }

    /**
    * Return the name of the audio codec used to encode this movie as a string.
    *
    * @return string
    */
    public function getAudioCodec() {
        if ($this->audioCodec === null) {
            $match = array();
            preg_match(self::$REGEX_AUDIO_CODEC, $this->output, $match);
            $this->audioCodec = (array_key_exists(1, $match)) ? trim($match[1]) : '';
        }

        return $this->audioCodec;
    }

    /**
    * Return the number of audio channels in this movie as an integer.
    *
    * @return int
    */
    public function getAudioChannels() {
        if ($this->audioChannels === null) {
            $match = array();
            preg_match(self::$REGEX_AUDIO_CHANNELS, $this->output, $match);
            if (array_key_exists(1, $match)) {
                switch (trim($match[1])) {
                    case 'mono':
                        $this->audioChannels = 1; break;
                    case 'stereo':
                        $this->audioChannels = 2; break;
                    case '5.1':
                        $this->audioChannels = 6; break;
                    case '5:1':
                        $this->audioChannels = 6; break;
                    default:
                        $this->audioChannels = (int) $match[1];
                }
            } else {
                $this->audioChannels = 0;
            }
        }

        return $this->audioChannels;
    }

    /**
    * Return boolean value indicating whether the movie has an audio stream.
    *
    * @return boolean
    */
    public function hasAudio() {
        return (boolean) preg_match(self::$REGEX_HAS_AUDIO, $this->output);
    }

    /**
    * Return boolean value indicating whether the movie has a video stream.
    *
    * @return boolean
    */
    public function hasVideo() {
        return (boolean) preg_match(self::$REGEX_HAS_VIDEO, $this->output);
    }

    /**
    * Returns a frame from the movie as an FFmpegFrame object. Returns false if the frame was not found.
    *
    *   * framenumber - Frame from the movie to return. If no framenumber is specified, returns the next frame of the movie.
    *
    * @param int $framenumber
 	* @param int $width
    * @param int $height
	* @param int $quality
    * @return FFmpegFrame|boolean
    */
	public function getFrame($framenumber = null, $width = null, $height = null, $quality = null) {

		$framePos = ($framenumber === null) ? $this->frameNumber : (((int) $framenumber) - 1);

		// Frame position out of range
    if (!is_numeric($framePos) || $framePos < 0 || $framePos > $this->getFrameCount())
      return false;

		if (any_empty($this->getFrameCount(), $this->getDuration()))
			return false;

	  $frameTime = round((($framePos / $this->getFrameCount()) * $this->getDuration()), 4);

		$frame = $this->getFrameAtTime($frameTime, $width, $height, $quality);

    // Increment internal frame number
    if ($framenumber === null) {
        ++$this->frameNumber;
    }

		return $frame;
	}

    /**
     * Returns a frame from the movie as an FFmpegFrame object. Returns false if the frame was not found.
     *
     * @param float $seconds
     * @param int $width
	 * @param int $height
     * @param int $quality
     * @param string $frameFilePath
     * @param array $output
     *
     * @throws Exception
     *
     * @return FFmpegFrame|boolean
     *
    */
	 public function getFrameAtTime($seconds = null, $width = null, $height = null, $quality = null, $frameFilePath = null, &$output = null) {
        // Set frame position for frame extraction
        $frameTime = ($seconds === null) ? 0 : $seconds;


        // time out of range
        if (!is_numeric($frameTime) || $frameTime < 0 || $frameTime > $this->getDuration()) {
			throw(new Exception('Frame time is not in range '.$frameTime.'/'.$this->getDuration().' '.$this->getFilename()));
        }

        if(is_numeric($height) && is_numeric($width)) {
            $image_size = ' -s '.$width.'x'.$height;
        } else {
            $image_size = '';
        }

        if(is_numeric($quality)) {
            $quality = ' -qscale '.$quality;
        } else {
            $quality = '';
        }

		$deleteTmp = false;
        if ($frameFilePath === null) {
            $frameFilePath = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('frame', true).'.jpg';
			$deleteTmp = true;
        }

		$output = array();

		// Fast and accurate way to seek. First quick-seek before input up to
		// a point just before the frame, and then accurately seek after input
		// to the exact point.
		// See: http://ffmpeg.org/trac/ffmpeg/wiki/Seeking%20with%20FFmpeg
		if ($frameTime > 30) {
			$seek1 = $frameTime - 30;
			$seek2 = 30;
		} else {
			$seek1 = 0;
			$seek2 = $frameTime;
		}

    exec(implode(' ', array(
        $this->ffmpegBinary,
        '-ss '.$seek1,
        '-i '.kb_escapeshellarg($this->movieFile),
        '-f image2',
        '-ss '.$seek2,
        '-vframes 1',
        $image_size,
        $quality,
        kb_escapeshellarg($frameFilePath),
        '2>&1',
    )), $output, $retVar);
    $output = join(PHP_EOL, $output);

        // Cannot write frame to the data storage
        if (!file_exists($frameFilePath)) {
			// Find error in output
			preg_match(self::$REGEX_ERRORS, $output, $errors);
			if ($errors) {
				throw new Exception($errors[0]);
			}
			// Default file not found error
            throw new Exception('TMP image not found/written '. $frameFilePath);
        }

        // Create gdimage and delete temporary image
        $imageSize = getimagesize($frameFilePath);
        switch ($imageSize[2]) {
            case IMAGETYPE_GIF:
                $gdImage = imagecreatefromgif($frameFilePath);
                break;
            case IMAGETYPE_PNG:
                $gdImage = imagecreatefrompng($frameFilePath);
                break;
            default:
                $gdImage = imagecreatefromjpeg($frameFilePath);
        }

        if ($deleteTmp && is_writable($frameFilePath)) {
            unlink($frameFilePath);
        }

        $frame = new FFmpegFrame($gdImage, $frameTime);
        imagedestroy($gdImage);

        return $frame;
    }

    /**
    * Returns the next key frame from the movie as an FFmpegFrame object. Returns false if the frame was not found.
    *
    * @return FFmpegFrame|boolean
    */
    public function getNextKeyFrame() {
        return $this->getFrame();
    }

    public function __clone() {
        $this->provider = clone $this->provider;
    }

    public function serialize() {
        $data = serialize(array(
            $this->ffmpegBinary,
            $this->movieFile,
            $this->output,
            $this->frameNumber,
            $this->provider
        ));

        return $data;
    }

    public function unserialize($serialized) {
        list($this->ffmpegBinary,
             $this->movieFile,
             $this->output,
             $this->frameNumber,
             $this->provider
        ) = unserialize($serialized);

    }
}


/**
 * OutputProvider interface of all output providers
 *
 * @author char0n (Vladimír Gorej, vladimir.gorej@gmail.com)
 * @package FFmpegPHP
 * @subpackage provider
 * @license New BSD
 * @version 2.6
 */
interface OutputProvider {

    /**
     * Setting movie file path
     *
     * @param string $movieFile
     */
    public function setMovieFile($movieFile);

    /**
     * Getting parsable output
     *
     * @return string
     */
    public function getOutput();
}

// abstract class AbstractOutputProvider implements OutputProvider, Serializable {
abstract class AbstractOutputProvider implements OutputProvider {

    protected static $EX_CODE_FILE_NOT_FOUND = 334561;
    protected static $persistentBuffer       = array();

    /**
     * Binary that returns info about movie file
     *
     * @var string
     */
    protected $binary;

    /**
     * Movie File path
     *
     * @var string
     */
    protected $movieFile;

    /**
     * Persistent functionality on/off
     *
     * @var boolean
     */
    protected $persistent;

    /**
     * Base constructor for every provider
     *
     * @param string $binary binary that returns info about movie file
     * @param boolean $persistent persistent functionality on/off
     */
    public function __construct($binary, $persistent) {
        $this->binary     = $binary;
        $this->persistent = $persistent;
    }

    /**
     * Setting movie file path
     *
     * @param string $movieFile
     */
    public function setMovieFile($movieFile) {
        $this->movieFile = $movieFile;
    }

    public function serialize() {
        return serialize(array(
            $this->binary,
            $this->movieFile,
            $this->persistent
        ));
    }

    public function unserialize($serialized) {
        list(
            $this->binary,
            $this->movieFile,
            $this->persistent
        ) = unserialize($serialized);
    }
}

/**
 * FFmpegOutputProvider ffmpeg provider implementation
 *
 * @author char0n (Vladimír Gorej, vladimir.gorej@gmail.com)
 * @package FFmpegPHP
 * @subpackage provider
 * @license New BSD
 * @version 2.6
 */

class FFmpegOutputProvider extends AbstractOutputProvider {

    protected static $EX_CODE_NO_FFMPEG = 334560;

    /**
     * Constructor
     *
     * @param string $ffmpegBinary path to ffmpeg executable
     * @param boolean $persistent persistent functionality on/off
     */
    public function __construct($ffmpegBinary = 'ffmpeg', $persistent = false) {
        parent::__construct($ffmpegBinary, $persistent);
    }

    /**
     * Getting parsable output from ffmpeg binary
     *
     * @return string
     */
    public function getOutput() {

        // Persistent opening
        if ($this->persistent == true && array_key_exists(get_class($this).$this->binary.$this->movieFile, self::$persistentBuffer)) {
            return self::$persistentBuffer[get_class($this).$this->binary.$this->movieFile];
        }

        // File doesn't exist
        if (!file_exists($this->movieFile)) {
            throw new Exception('Movie file not found', self::$EX_CODE_FILE_NOT_FOUND);
        }

        // Get information about file from ffmpeg
        $output = array();

        exec($this->binary.' -i '.kb_escapeshellarg($this->movieFile).' 2>&1', $output, $retVar);
        $output = join(PHP_EOL, $output);

        // ffmpeg installed
        if (!preg_match('/FFmpeg version/i', $output)) {
            throw new Exception('FFmpeg is not installed on host server', self::$EX_CODE_NO_FFMPEG);
        }

        // Storing persistent opening
        if ($this->persistent == true) {
            self::$persistentBuffer[get_class($this).$this->binary.$this->movieFile] = $output;
        }

        return $output;
    }
}

/**
 * AbstractOutputProvider parent of all output providers
 *
 * @author char0n (Vladimír Gorej, vladimir.gorej@gmail.com)
 * @package FFmpegPHP
 * @subpackage provider
 * @abstract
 * @license New BSD
 * @version 2.6
 */


 /**
  * FFmpegFrame represents one frame from the movie
  *
  * @author char0n (Vladimír Gorej, vladimir.gorej@gmail.com)
  * @package FFmpegPHP
  * @license New BSD
  * @version 2.6
  */
 // class FFmpegFrame implements Serializable {
 class FFmpegFrame {

     protected static $EX_CODE_NO_VALID_RESOURCE = 334563;

     /**
     * GdImage binary data
     *
     * @var string
     */
     protected $gdImageData;
     /**
     * Presentation time stamp
     *
     * @var float
     */
     protected $pts;

     /**
     * Frame width in pixels
     *
     * @var int
     */
     protected $width;

     /**
     * Frame height in pixels
     *
     * @var int
     */
     protected $height;

     /**
     * Create a FFmpegFrame object from a GD image.
     *
     * @param resource $gdImage image resource of type gd
     * @param float $pts frame presentation timestamp; OPTIONAL parameter; DEFAULT value 0.0
     * @throws Exception
     */
     public function __construct($gdImage, $pts = 0.0) {
         if (!(is_resource($gdImage) && get_resource_type($gdImage) == 'gd')) {
             throw new Exception('Param given by constructor is not valid gd resource', self::$EX_CODE_NO_VALID_RESOURCE);
         }

         $this->gdImageData = $this->gdImageToBinaryData($gdImage);
         $this->width       = imagesx($gdImage);
         $this->height      = imagesy($gdImage);
         $this->pts         = $pts;
     }

     /**
     * Return the width of the frame.
     *
     * @return int
     */
     public function getWidth() {
         return $this->width;
     }

     /**
     * Return the height of the frame.
     *
     * @return int
     */
     public function getHeight() {
         return $this->height;
     }

     /**
     * Return the presentation time stamp of the frame; alias $frame->getPresentationTimestamp()
     *
     * @return float
     */
     public function getPTS() {
         return $this->pts;
     }

     /**
     * Return the presentation time stamp of the frame.
     *
     * @return float
     */
     public function getPresentationTimestamp() {
         return $this->getPTS();
     }


     public function resize($width, $height, $keep_aspect_ratio = true, $cropTop = 0, $cropBottom = 0, $cropLeft = 0, $cropRight = 0 ) {

      $widthCrop = (int) $cropLeft + (int) $cropRight;
      $heightCrop = (int) $cropTop + (int) $cropBottom;
      $gdImage = $this->toGDImage();

			if ($keep_aspect_ratio) {

				if ($this->getWidth() > $this->getHeight()) {

					$ratio = $this->getHeight()/$this->getWidth();
					$zoom = $width/($this->getWidth()-$widthCrop);
					$height = ($this->getHeight()-$heightCrop)*$zoom;

				}

				else {
					$ratio = $this->getWidth()/$this->getHeight();
					$width = floor($height * $ratio) - ($widthCrop*($height/$this->getHeight()));
				}

			}

			$resizedImage  = imagecreatetruecolor($width, $height);

      imagecopyresampled($resizedImage, $gdImage, 0, 0, $cropLeft, $cropTop, $width, $height, $this->getWidth()-$widthCrop, $this->getHeight()-$heightCrop);
      imageconvolution($resizedImage, array(
         array( -1, -1, -1 ),
         array( -1, 24, -1 ),
         array( -1, -1, -1 ),
      ), 16, 0);

      $this->gdImageData = $this->gdImageToBinaryData($resizedImage);
      $this->width       = imagesx($resizedImage);
      $this->height      = imagesy($resizedImage);
      imagedestroy($gdImage);
      imagedestroy($resizedImage);
     }

     /**
     * Crop the frame.
     *
     *      * croptop - Remove [croptop] rows of pixels from the top of the frame.
     *      * cropbottom - Remove [cropbottom] rows of pixels from the bottom of the frame.
     *      * cropleft - Remove [cropleft] rows of pixels from the left of the frame.
     *      * cropright - Remove [cropright] rows of pixels from the right of the frame.
     *
     * NOTE: Crop values must be even numbers.
     *
     * @param int $cropTop
     * @param int $cropBottom OPTIONAL parameter; DEFAULT value - 0
     * @param int $cropLeft OPTIONAL parameter; DEFAULT value - 0
     * @param int $cropRight OPTIONAL parameter; DEFAULT value - 0
     * @return void
     */
     public function crop($size_w, $size_h, $cropTop = 0, $cropBottom = 0, $cropLeft = 0, $cropRight = 0) {
         $this->resize($size_w, $size_h, true, $cropTop, $cropBottom, $cropLeft, $cropRight);
     }

     /**
     * Returns a truecolor GD image of the frame.
     *
     * @return resource resource of type gd
     */
     public function toGDImage() {
         return imagecreatefromstring($this->gdImageData);
     }

     public function to_jpg() {
			 $image = imagecreatefromstring($this->gdImageData);
			 ob_start();
			 imagejpeg($image);
			 return ob_get_clean();
         // return $this->gdImageData;
     }

     protected function gdImageToBinaryData($gdImage) {
         ob_start();
         imagegd2($gdImage);
         return ob_get_clean();
     }

     public function serialize() {
         $data = array(
             $this->gdImageData,
             $this->pts,
             $this->width,
             $this->height
         );

         return serialize($data);
     }

     public function unserialize($serialized) {
         $data = unserialize($serialized);
         list($this->gdImageData,
              $this->pts,
              $this->width,
              $this->height
         ) = $data;
     }
 }