<?php

// VIEW Objects

class KB_view extends KB_class {

	private $core;

	public
		$cards,
		$csv,
		$table;

	function __construct ($core) {
		$this->core = $core;
	}

	public function cards() {
		$identifier = debug_backtrace();
		$identifier = "i".md5($identifier[0]["file"].$identifier[0]["line"]);
		$cards = new KB_cards($this->core);
		$cards->identifier = $identifier;
		return $cards;
	}


	public function table() {

		$identifier = debug_backtrace();
		$identifier = "i".md5($identifier[0]["file"].$identifier[0]["line"]);

		$table = new KB_tables($this->core);
		$table->identifier = $identifier;
		return $table;

	}


	public function ul() {
		return new KB_lists($this->core);
	}


	public function csv() {
		$identifier = debug_backtrace();
		$identifier = "i".md5($identifier[0]["file"].$identifier[0]["line"]);
		$csv = new KB_csv_table($this->core);
		$csv->identifier = $identifier;
		return $csv;
	}

}

// LIST

class KB_lists extends KB_class {

	private $core;


	function __construct ($core){

		$this->core = $core;

	}

	function render($array, $group_ending = br, $element_ending = ", ") {

		$return = "";

		if (!empty($array) && is_array($array)) {

			foreach($array as $a) {

				if (!empty($a) && is_array($a)) {

					foreach($a as $e) {
						$return .= $e.$element_ending;
					}
					$return .= $group_ending;
				}

				else{
					$return .= $a.$group_ending;
				}
			}
		}
		return $return;
	}

}

// TABLE

class KB_tables extends KB_class {

	private
		$core,
		$count_entries,
		$even_odd_counter,
		$pagination_o,
		$post_actions = [],
		$preg_function = '/([^(^ ]+)\((.*)\)(.*)/s',
		$search_table = [];

	public
		$calc,
		$element = [],
		$format,
		$identifier = "",
		$pagination = 20,
		$row = [],
		$rows = [];


	function __construct ($core){

		$this->core = $core;

		$this->pagination_o = $core->pagination();

		$this->config("class", "data");
		$this->config("style_classes",array(
			'even' => '',
			'odd' => '',
		));

		// $this->config("spacing","&nbsp;&nbsp;&nbsp;&nbsp;");
		$this->config("spacing",$this->core->html->span("", array("class"=>"child_element")));
		$this->config("keyvarname","ID");
		$this->config("keygetname","ID");
		$this->config("exclude",[]);

		$this->config("sort",ifttt($this->core->runs("backend"),true,false));

		$this->calc = new KB_table_calc($this->core);
		$this->format = new KB_format($this->core);


	}


	// Public

	public function column ($name = false, $array = false) {

		if (is_empty($element = key_of($name, $this->element)))
			return false;

		if (any_empty($array, is_array($array), !is_indexed($array)))
			return $element;

		$template = array_keys($this->column_template());

		foreach ($array as $key => $value) {

			if (!fetch($key, $template))
				continue;

			$this->element[$name][$key] = $value;
		}

		return $this->element[$name];
	}

	public function create_nested ($tables, $index = 0, $filter=false) {

		$table = $tables[$index];

		if ($filter)
			$table["data"] = $this->core->array->filter_by($table["data"], key($filter), current($filter));

		if (!isset($table["nested"]))
			$table["nested"] = false;

		if (!isset($table["object"],$table["data"]))
			return false;

		if (empty($table["data"]))
			return false;

		$return = [];

		$head = $table["object"]->head($table["data"],$table["nested"]);

		foreach($table["data"] as $data){

			$nested_content = false;

			if (isset($tables[$index+1]) && isset($table["nested"]) && isset($data["ID"]))
				if ($nested_content = $table["object"]->create_nested(
					$tables,
					$index+1,
					array($table["nested"]=>$data["ID"])))
				$nested = array(
					"tr" => array(
						array(
							"td" => array(
								"colspan" => count($this->element),
								$nested_content
							)
						)
					)
				);

			if ($nested_content)
				$data["nested"] = array(
					"span" => array(
						"class" => "has_nested"
					)
				);

			$body[] = $table["object"]->body($data);
			$body[] = "\n";
			if ($nested_content)
				$body[] = $nested;

		}

		$return[] = array(
			"table" => array(
				"class" => $table["object"]->config("class"),
				$head,
				$body
			)
		);

		return $return;

	}


	public function new_column ($element = false, $array = false) {

		if (is_empty($element) || !is_string($element))
			return false;

		$this->element[$element] = $this->column_template($element);

		if ($array !== false)
			$this->set_column($element,$array);

		return true;

	}


	public function new_columns ($elements = false) {

		if (!is_array($elements) || is_empty($elements))
			return false;

		foreach ($elements as $element => $array)
			$this->new_column($element, $array);

		return true;
	}


	public function new_row ($element = "", $array = false) {

		if (is_empty($element))
			return;

		$this->row[$element] = $this->row_template($element);

		if ($array)
		{
			$this->set_row($element,$array);
		}
	}


	public function pagination ($set = NULL) {

		if (is_numeric($set) || $set === false)
			$this->pagination = $set;

		if ($this->core->runs("frontend") && is_empty($this->core->session->ID()))
			return false;

		return $this->pagination;

	}


	public function remove_column ($column = "") {

		if (is_empty($column))
			return true;

		if (is_indexed($column)) {

			$return = [];

			foreach ($columns = $column as $column)
				$return[] = $this->remove_column($column);

			return $return;
		}

		if (is_null(key_of($column, $this->element))) {

			if (not_empty($element = $this->core->array->filter($this->element, ["name"=>$column])))
				return $this->remove_column(key($element));

			return false;
		}

		unset($this->element[$column]);

		return true;

	}


	public function render ($body = false, $group = false, $foot = "") {

		$this->calc = new KB_table_calc($this->core);

		if (is_empty($body))
			$body = array([]);

		if (is_object($body))
			return $this->render_object($body, $group, $foot);

		$this->count_entries = 0;

		if (is_empty($this->element) && is_array(current($body))) {

			foreach (current($body) as $k => $b) {
				$this->new_element($k);
			}
		}

		$head = isset($this->element) ? array_intersect_key($this->element,array_flip(array_diff(array_keys($this->element),array_keys(array_flip($this->config("exclude")))))) : array_keys($this->config("exclude"));
		$this->rows = isset($this->row) ? array_keys($this->row) : false;

		if (isset($this->element))
			$foot = array_intersect_key($this->element,array_flip(array_diff(array_keys($this->element),array_keys(array_flip($this->config("exclude"))))));
		//else $head = array_keys($this->exclude);
		$return  = "";
		$return	.= $this->begin($head);
		$return	.= $this->extended_head($head);
		$return .= $this->search_row($head);

		if (not_empty($this->get_sort())){

			// if (isset($head[$this->get_sort("key")]["sort"]) && !is_bool($head[$this->get_sort("key")]["sort"]))
				// $this->get_sort() = $head[$this->get_sort("key")]["sort"];

			$body = $this->core->array->sort_by(
				$body,
				either(
					key_fetch("key",$this->get_sort()),
					key_fetch("sort",$this->get_sort())
				),
				$this->get_sort("flag")
			);

		}

		if ($group) {

			// $body = $this->core->array->group_by($body, "group_ID");
			/*
			// Remove entries with group IDs not found in groups
			$remove_keys = $this->core->array->complement(
				array_keys($body),
				key_fetch("ID",$group)
			);

			$body = remove_keys_of($remove_keys, $body);
			*/
			// $body = $this->core->array->ungroup($body,$group);
			// $body = $this->core->array->ungroup_by($body, "ID");
		}
		/*
		else if (not_empty(current($body)) && array_key_exists("parent_ID",current($body))){
				$get_highest_group_key = $this->core->array->get_highest_group_key($body,"parent_ID");
				$body = $this->core->array->group_by($body,"parent_ID");
				$body = $this->core->array->ungroup($body, $get_highest_group_key);
		}
		*/

		if (isset($this->search_table) && !empty($this->search_table)) {

			foreach($this->search_table as $key => $value){

				if (array_key_exists($key,$this->element)) {

					if (is_bool(key_of("search", $this->element[$key])) && is_array(key_of("search_config", $this->element[$key]))) {
						deprecated("search_config in tables is deprecated and should be set with search",2);
						$this->element[$key]["search"] = key_of("search_config", $this->element[$key]);
					}

					$filter = either(key_of("var", $this->element[$key]["search"]), $key);
					$value = ifttt(are_equal(key_of("type",$this->element[$key]["search"]), "select"), $value, "(.*)".$value."(.*)");

					$body = $this->core->array->filter_by(
						$body,
						$filter,
						$value
					);
				}
			}
		}


		if ($group) {

			$body = $this->core->array->sort_by_hierarchie($body, $group, "ID", "group_ID", "depth");
			$body = $this->pagination_o->set($body,$this->pagination);
			$body = $this->core->array->group_by($body, "group_ID");

			/*
			$body = $this->core->array->group_by($body, "group_ID");
			$body = $this->core->array->ungroup($body, $group);
			$body = $this->pagination_o->set($body,$this->pagination);
			$body = $this->core->array->group_by($body,"group_ID");
			*/

			// if (isset($body[""]))
				// $return .= $this->extended_body($body[""], $head, NULL);

			foreach ($group as $group_entry) {

				$group_key = key_of("ID", $group_entry);

				if (!isset($group_entry["depth"]))
					$group_entry["depth"] = 0;

				$group_entries[$group_entry["depth"]] = $group_entry;
				$group_entries = array_slice($group_entries,0,$group_entry["depth"]+1);

				$this->even_odd_counter = 0;

				if (!isset($body[$group_key]))
					continue;

				// if (isset($group_entry["depth"]))
					// $extended_body = $this->extended_body($body[$group_key], $head, false);

				// else
					$extended_body = $this->extended_body($body[$group_key], $head, $group_key);

				if (!empty($extended_body)){
					$return .= $this->group($group_entries,$head);
					// $group_entries = [];
				}

				$return .= $extended_body;
			}
		}

		else {
			$body = $this->pagination_o->set($body,$this->pagination());
			$return .= $this->extended_body($body, $head);
		}

		$return .= $this->foot($foot, $head);
		$return .= $this->end();

		if ($this->pagination())
			$return .= $this->pagination_o->render();

		return $return;
	}


	public function render_nested () {
		$tables = func_get_args();
		$return = $this->create_nested($tables);
		$return = $this->core->html->render($return);
		return $return;

	}


	public function set_column ($element = false, $array = false) {

		if (!array_key_exists($element, $this->element))
			return false;

		if (!is_array($array) || is_indexed($array))
			return false;

		$template = array_keys($this->column_template());

		foreach ($array as $key => $value) {

			if (!fetch($key, $template))
				continue;

			$this->element[$element][$key] = $value;
		}

		return $this->element[$element];
	}


	public function set_columns ($elements = []) {

		if (any_empty($elements, is_array($elements)))
			return false;

		foreach ($elements as $element => $values) {

			$this->set_column($element, $values);
		}

		return $this->element;
	}


	public function set_row ($element, $array) {
		$this->row[$element] = $array+$this->row[$element];
	}




	// Private

	private function begin ($head=false) {

		$return = "";

		if ($this->elements_have("search")){
			$this->search_begin();

			$return .=  '<form class="'.fetch("table",$this->config("style_classes")).'" action="'.$this->core->url(array('page'=>'')).'" method="post">';
		}

		if ($this->elements_have("sort")){
			$this->sort_begin();
		}

		$return .= '<table class="'.either(fetch("table",$this->config("style_classes")), $this->config("class")).'">'."\n";
		return $return;
	}

	private function body ($data) {

		if (!is_array($data))
			return false;

		$return = [];
		$row = [];

		foreach($this->element as $column => $element)
			$row[] = array(
				"td" => array(
					"class" => isset($element["class"]) ? $element["class"] : false,
					isset($data[$column]) ? $data[$column] : ""
				)
			);


		$return[] = array("tr"=>$row);

		return $return;

	}

	protected function column_template ($column = "") {

		if (!is_string($column))
			return $this->column_template();

		$template = array(
			"name" => $column,
			"title" => $column,
			"sort" => $this->config("sort"),
			"search" => false,
			"value" => preg_match('-\$-',$column) ? $column : '$'.$column,
			"values" => [],
			"markup" => "",
			"format" => "",
			"class" => "",
			"footer" => "",
			"cell_markup"	=> "",
			"collapsible"	=> false,
			"search_config"	=> array(
				"var" => $column
			),
		);

		return $template;
	}

	protected function elements_have ($property) {

		// FAIL

		if (is_empty($this->element))
			return false;

		if (in_array(true, $this->core->array->implode($this->element,$property), true))
			return true;

		if (not_empty($this->core->array->implode($this->core->array->implode($this->element,$property))))
			return true;

		return false;

	}

	private function end () {
		$return = "";
		$return .=  "</table>";
		//if ($this->search) $return .=  '<table class="'.$this->get_config(array('style_classes','table')).'">'."\n";

		if ($this->elements_have("search"))
			$return .=  '</form>';

		return $return;
	}

	protected function even_odd ($ID = false) {

		static $even_odd_counter = 0;

		$even_odd_counter++;

		$tr_class = [];


		if ($ID && $ID == $this->core->get_var())
			$tr_class[] = "active";

		$tr_class[] = ifttt(
			$even_odd_counter % 2 == 0,
			fetch("even",$this->config("style_classes")),
			fetch("odd",$this->config("style_classes"))
		);

		$tr_class = implode(" ", $tr_class);

		return $tr_class;

	}

	protected function extended_head ($head=false) {

		$return = "";

		if (not_empty($head) && is_array($head) && not_empty(array_values((array)key_fetch("title",$head)))) {

			$return .= "<thead><tr>";

			foreach($head as $h) {

				$h["title"] = $this->core->language->translate($h["title"]);

				if (not_empty($this->get_sort()) && $this->get_sort("key") == $h["name"])
					$h['markup'] = $this->core->html->add_class($h['markup'], "active");

				$h['markup'] = $this->core->html->add_class($h['markup'], key_of("class", $h));

				if ($h['collapsible'])
					$h['markup'] = $this->core->html->property($h['markup'], "data-collapsible", $h['collapsible']);

				if ($h['sort'])
					$return .=  "<th ".$h['markup'].">".$this->get_sort_url($h)."</th>".br;

				else
					$return .=  "<th ".$h['markup'].">".$h['title']."</th>".br;

			}
			$return .= "</tr></thead>".br;
		}
		return $return;
	}

	protected function extended_body ($body = [], $head = false, $group_key = false, $padding = 0) {

		$return = "";

		if (!is_array($body))
			return $return;

		$even = 1;

		if (any_empty($head, is_array($head)))
			return "";

		foreach($body as $b) {

			$even = 1 - $even;

			//Special Chars Encode
			$b = preg_replace_array(array("/\(/","/\)/","-,-"),array("&#40;","&#41;","&#44;"),$b);

			$this->count_entries++;

			if (!empty($b) && is_array($b)) {

				$var = $b;

				if (empty($var['depth']))
					$var['depth'] = 0;

				$var['depth'] = intval($var['depth']) + $padding;

				if ($group_key && (key_of("group_ID", $var)))
					$var['depth'] = 1;

				$classes = [];

				// Highlight active row
				if (
					not_empty($this->config("keygetname")) &&
					not_empty($this->core->get($this->config("keygetname"))) &&
					not_empty(key_of($this->config("keygetname"), $b)) &&
					are_equal(key_of($this->config("keygetname"), $b), $this->core->get_var($this->config("keygetname")))
					)
					$classes[] = "active";

				$classes[] = ifttt(are_equal($even,1),"even","odd");

				$return .= '<tr id="tr'.$this->count_entries.'" class="'.implode(" ",$classes).'">';

				$keys = array_keys($var);
				// $keys = array_unique(array_merge($keys,array_keys($this->element)));
				usort($keys, "rsort_by_length");

				$string	= "/^NULL$/";

				foreach($keys as $k => $v)
					$keys[$k] = '/(\$)('.$v.')(?!_)/';

				$first_column = true;

				foreach($head as $key => $value) {

					$markup = either(key_of("cell_markup",$head[$key]), key_of("markup",$head[$key]));
					$markup = $this->core->html->add_class($markup, key_of("class", $head[$key]));

					if ($format = key_of("format", $head[$key]))
						$markup = $this->core->html->add_class($markup, $format);


					$return .= "<td";
					$return .= ifttt($markup, " ".$markup);
					$return .= ">";

					if (($first_column || $group_key === false) && ($key == "name" || key_of("name",$value) == "name")){
						$spacing = $this->config("spacing");
						// $margin=str_repeat($spacing,intval($var['depth']));
						$margin = str_repeat($spacing, $var['depth']);

						$return .= $margin;
					}

					if ($head[$key]['value'] && !is_array($head[$key]['value'])) {
						$value = $head[$key]['value'];

						$value = $this->replace_var($keys,$var,$value);

						if (preg_match($this->preg_function,$value))
          		$value = $this->replace_function($value);

						if (is_scalar($value)) {
							$value = preg_replace($string,'',$value);
							$value = preg_replace_array(["/&#40;/","/&#41;/","/&#44;/"],["(",")",","],$value);
						}

						if (not_empty($values = key_of("values", $head[$key])) && is_array($values)) {
							$value = key_of($value, $values);
						}

						else if (!empty($head[$key]["array"]) && $json = $this->core->json->decode($value)){

							$value = $json;

							if (is_indexed($value)) {
								if (count($value) < 2)
									$value = implode($value);
							}
						}

						if ($format = key_of("format", $head[$key])) {
							$options = NULL;
							/*
							if (are_equal($format, "file"))
								$options = [
									"class" => true,
									"container" => [
										"span" => [
											"class" => "zoom light"
										]
									]
								];
							*/
							// if (are_equal($format, "file"))
							if (in_list($format, ["file","file_icon"]))
								$options = $this->core->array->merge([
									"class" => true,
									"container" => [
										"span" => [
											"class" => "zoom"
										]
									]
								], key_of("format_options", $head[$key])
							);


							if (not_empty($options))
								$value = $this->core->format->$format($value, $options);
							else
								$value = $this->core->format->$format($value);
						}

						if (is_string($value))
							$return .= $value;
					}

					else if ($head[$key]['value'] && is_array($head[$key]['value'])) {

						$value = $head[$key]['value'][0];
						$array = $head[$key]['value'][1];
						$value = $this->replace_var($keys,$var,$value);

            if (preg_match($this->preg_function,$value))
          		$value = $this->replace_function($value/*,$this->preg_function*/);

            if (not_empty($array)) {

							if (isset($array[$value]))
								$value = $array[$value];

							// else if (in_array($value,$array))
								// $value = $value;
						}

            // else
						 	// $value = "";

						$value = preg_replace($string,'',$value);

						$return .= $value;
					}
					//else if (isset($var[$head[$key]['name']]))	$return .= preg_replace($string,'',$var[$head[$key]['name']]);
					$return .= "</td>";
					$first_column = false;
				}

				$return .= "</tr>\n";

				// Additional rows

				if (not_empty($this->row)) {

					if ($row_keys = array_intersect_key($this->row,$b)) {

						foreach($row_keys as $row_key => $row_config){

							if (not_empty($b[$row_key])) {

								$cell_value = $this->replace_var($keys,$b,key_of("value", $row_config));

								if (preg_match($this->preg_function, $cell_value))
									$cell_value = $this->replace_function($cell_value);


								$class = [];
								$class[] = ifttt(are_equal($even,1),"even","odd");
								$class[] = key_of("class",$row_config);

								$return .= $this->core->html->render(
									$this->core->html->tr(
										$this->core->html->td($cell_value,["colspan"=>count($this->element)]),
										["class"=>$class]
									)
								);
							}
						}
					}
				}
				next($body);

			}

		}


		return $return;
	}

	private function foot ($foot, $head = []) {

		$return = "";

		$empty = true;

		if (!empty($foot) && is_array($foot)){

			$keys = array_keys($foot);

			foreach($keys as $k => $v)
				$keys[$k] = '/(\$)('.$v.')/';

			$return .= "<tr>";

			foreach($head as $key => $value){

				$return .= "<th";
				$return .= !empty($head[$key]['cell_markup']) ? " ".$head[$key]['cell_markup'] : "";
				$return .= ">";
				$footer = key_of("footer",key_of($key, $foot));

				if ($empty)
					$empty = is_empty($footer);

				$footer = $this->replace_var($keys,$foot,$footer);

				if (preg_match($this->preg_function,$footer))
					$footer = $this->replace_function($footer/*,$this->preg_function*/);

				$return .= (!in_array($foot, $this->config("exclude"))) ? $footer: "";
				$return .= "</th>".br;
			}
			$return .= "</tr>".br;
		}
		if (!$empty)
			return $return;

		return "";
	}

	protected function get_sort ($key = false) {

		if (is_empty($sort = $this->private_config("sort")))
			$sort = $this->private_config(
				"sort",
				$this->core->json->decode(
					html_entity_decode(
						either($this->core->session->get($this->identifier.'_sort'), "")
					)
				)
			);

		if ($key)
			return key_fetch($key, $sort);

		return $sort;

	}

	protected function get_sort_url ($config) {

		$name = key_fetch("title",$config);
		$col = key_fetch("name",$config);
		$sort = key_fetch("sort",$config);

		if (is_empty($sort))
		 	return key_fetch("title",$config);

		if ($sort === $col)
			$sort = false;

		if (fetch("key",$this->get_sort()) == $col && fetch("flag",$this->get_sort()) == "") {
			$gets['sort_'.$this->identifier] = $this->core->json->encode(array("key"=>$col,"flag"=>"DESC","sort"=>$sort));
			$img = $this->core->html->render($this->core->template->icon("up", array("class"=>"sort")));
		}

		else if (fetch("key",$this->get_sort()) == $col && fetch("flag",$this->get_sort()) == "DESC") {
			$gets['sort_'.$this->identifier] = "false";
			$img = $this->core->html->render($this->core->template->icon("down", array("class"=>"sort")));
		}

		else {
			$gets['sort_'.$this->identifier] = $this->core->json->encode(array("key"=>$col,"flag"=>"","sort"=>$sort));
			// $img = '<img src="sort_desc.png" class="sort">';
			$img = "";
		}

		return '<a href="'.$this->core->generate_url($gets).'">'.$name.'</a>'.$img;
	}

	protected function group ($group_entries,$head) {
		$colspan = count($head);

		if (isset($group_entry['depth']))
		{
			// $spacing = $this->get_config("spacing");
			// $margin=str_repeat($spacing,intval($group_entry['depth']));
			// $group_entry["name"] = $margin.$group_entry["name"];
		}

		$name = [];

		foreach($group_entries as $group_entry)
			if (not_empty(key_of("name", $group_entry)))
				$name[] = $group_entry["name"];

		if (is_empty($name))
			return;

		$name = implode(" / ",$name);

		$return = $this->core->html->tr(
			$this->core->html->td(
				$name,
				["colspan" => $colspan]
			),
			["class" => "category"]
		);

		return $return.br;
	}

	private function head ($data,$nested){

		if (!isset($this->element) && !is_array(current($data)))
			return false;

		if (!isset($this->element) && is_array(current($data))){
			foreach(array_keys($data) as $key){
				$this->new_element($key);
			}
		}

		if ($nested)
			$this->new_element("nested",array(
				"title"=>"",
				"class"=>"nested",
			));

		$return = [];

		if (is_empty(array_values((array)key_fetch("title", $this->element))))
			return "";

		foreach($this->element as $column => $element)
			$row[] = array(
				"th" => array(
					"class" => isset($element["class"]) ? $element["class"] : false,
					isset($element["title"]) ? $element["title"] : $column
				)
			);

		$return[] = array(
			"tr"=>array(
				$row
			)
		);

		return $return;

	}

	private function render_object ($object = false, $group = false, $foot = "") {

		if (is_empty($object) || !is_object($object) || !($object instanceof  KB_data))
			return false;

		if ($group)
			return $this->render($object->read(), $group, $foot);

		$pagination = $this->core->pagination();
		$this->sort_begin();
		$return = [];

		if (is_empty($this->element)) {

			if ($this->pagination())
				$object = $pagination->set($object, $this->pagination());

			if (not_empty($read = $object->read()))
				$return[] = $this->core->html->table($read, array_keys(current($read)), array("class"=>"data"));

		}

		else {

			$string	= "/^NULL$/";

			foreach ($this->element as $key => $element) {

				$element["title"] = $this->core->language->translate($element["title"]);

				$cell = $this->core->html->tag(
					"th",
					array($this->get_sort_url($element))
				);

				$cell = $this->core->html->property($cell,	key_of("markup", $element));

				if (
					not_empty(key_of("sort", $element)) &&
					ifttt(are_equal(key_of("sort", $element), true), key_of("name", $element), key_of("sort", $element)) == $this->get_sort("key")
					)
					$cell = $this->core->html->add_class($cell,"active");

				if (
					not_empty(key_of("format", $element))
					)
					$cell = $this->core->html->add_class($cell,key_of("format", $element));

				$head[] = $cell;
			}

			$head = $this->core->html->tag("tr",$head);

			$search = false;

			foreach ($this->element as $key => $element) {

				if (key_of("search",$element) !== false) {
					$search = true;
					break;
				}

			}

			if ($search) {

				$search = [];

				// Create form fields

				$form = $this->core->form();
				$this->search_table = $this->core->session->get($this->identifier.'_search');

				foreach ($this->element as $key => $element) {

					if (key_of("search",$element)) {

						$name = either(
							key_of("var", key_of("search", $element)),
							key_of("var", key_of("search_config", $element)),
							$key
						);

						$search_config = remove_keys_of("var", key_of("search_config", $element));

						if (is_array($search_settings = key_of("search",$element))) {
							$element["search_config"] = $search_settings;
							$this->element[$key]["search_config"] = $search_settings;
							$search_config = $search_settings;
						}

						else if ($search_settings === true && is_array($search_config) && not_empty($search_config)){
							deprecated("search_config in tables is deprecated and should be set with search",2);
						}


						if (are_equal(key_of("type", $search_config), "select")) {

							if (is_empty($values = key_of("values", $search_config))) {
								$values = $object->column_values($key);
								$values = array_combine($values,$values);
							}

							$values = $this->core->array->union([""=>""],$values);
							$search_config["values"] = $values;
						}

						if (is_array($name))
							$name = key($name);

						$form->new_element($name,array(
							"class"	=> "search",
							"value" => key_of($name, $this->search_table),
						));

						$form->set_element($name, $search_config);
					}
				}

				$form->new_element('search', array(
					'type'		=> 'submit',
					'label'		=> $this->core->template->icon("zoom"),
					'markup'	=> 'class="small button"',
				));

				// Check form

				if ($form->submit()) {

					$this->search_table = $this->core->array->trim($form->values());

					// unset($this->search_table["search"]);

					$this->core->session->set(
						$this->identifier.'_search',
						$this->search_table
					);

					$this->core->refresh();

				}

				// Output form fields

				foreach ($this->element as $key => $element) {

					$cell_value = "";

					if (key_of("search",$element)) {

						$name = either(
							key_of("var", key_of("search_config", $element)),
							$key
						);

						if (is_array($name))
							$name = key($name);

						$cell_value = $form->field($name);

					}


					if ($key == current(array_slice(array_keys($this->element), -1)))
						$cell_value = $this->core->html->span([
							$cell_value,
							$form->field("search")
						],["class"=>"submit"]);

					$search[] = $this->core->html->tag("td",$cell_value);
				}

				$search = $this->core->html->tag("tr",$search, ["class"=>"search"]);
			}

			if (not_empty($this->search_table))

				foreach ($this->search_table as $column => $search_entry) {

					if (in_list(
						key_of("type", key_of($column, $form->element)),
						array("select", "radio")
					))
						$object->filter(array($column => $search_entry));

					else
						$object->search($column, $search_entry);
				}

			if ($this->get_sort()) {

				$object->order(array(
					either(
						ifttt(is_string($this->get_sort("sort")),$this->get_sort("sort"),""),
						$this->get_sort("key")
					)=>
					either(
						$this->get_sort("flag"),
						"ASC"
					)
				));
			}

			if ($object->group() && $object instanceof KB_data_group) {

				$entries = $object->read_structured();

				if ($this->pagination())
					$entries = $pagination->set($entries, $this->pagination());
			}

			else {

				if ($this->pagination())
					$object = $pagination->set($object, $this->pagination());

				$entries = $object->read();

			}

			$body = [];
			$keys = [];
			$even = 1;

			if (not_empty($entries))

				foreach ($entries as $row) {

					$even = 1 - $even;

					// Encode special chars which would confuse function mappings
					$row = preg_replace_array(
						array(
							"/\(/",
							"/\)/",
							"/,/"
						),
						array(
							"&#40;",
							"&#41;",
							"&#44;"
						),
						$row
					);

					$keys = array_keys($row);
					usort($keys, "rsort_by_length");
					foreach ($keys as $k => $v)
						$keys[$k] = '/(\$)('.$v.')(?!_)/';


					$return_cells = [];
					$first_column = true;
					$depth = either(int(key_of("depth", $row)), 0);

					foreach ($this->element as $key => $element) {

						$cell_value = "";
						$value = key_of("value", $element);
						$margin = "";

						if (($first_column) && ($key == "name" || key_of("name",$element) == "name")){
							$spacing = $this->config("spacing");
							if ($depth)
							$margin = str_repeat($spacing, $depth);
						}

						if (is_string($value)) {

							$cell_value = $this->replace_var($keys,$row,$value);

							if (preg_match($this->preg_function, $cell_value))
								$cell_value = $this->replace_function($cell_value);

							if (key_of("array", $element) && $json = $this->core->json->decode($cell_value)) {

								$cell_value = $json;

								if (is_indexed($cell_value) && count($cell_value) < 2) {
									$cell_value = implode($cell_value);
								}
							}

							if (not_empty($values = key_of("values", $element)) && is_array($values))
								$cell_value = key_of($cell_value, $values);

						}

						else if (is_array($value)) {

							$array = key_of(1,$value);
							$value = key_of(0,$value);

							$cell_value = $this->replace_var($keys,$row,$value);

							if (preg_match($this->preg_function, $cell_value))
								$cell_value = $this->replace_function($cell_value);

							$cell_value = either(fetch($cell_value, $array), $cell_value);

						}

						if (not_empty($cell_value)) {
							$cell_value = preg_replace($string,'',$cell_value);
							$cell_value = preg_replace(array("/&#40;/","/&#41;/","/&#44;/"), array("(",")",","),$cell_value);
						}

						if ($format = key_of("format", $element)) {

							$element["cell_markup"] = ["class" => $format];
							$options = NULL;

							if (in_list($format, ["file","file_icon"]))
								$options = $this->core->array->merge([
									"class" => true,
									"container" => [
										"span" => [
											"class" => "zoom"
										]
									]
								], key_of("format_options", $element)
							);

							if (not_empty($options))
								$cell_value = $this->core->format->$format($cell_value, $options);
							else
								$cell_value = $this->core->format->$format($cell_value);
						}

						$cell = $this->core->html->tag("td",$margin.$cell_value);
						$cell = $this->core->html->add_markup($cell, key_of("cell_markup",$element));
						$return_cells[] = $cell;
					}

					$return_row = $this->core->html->tr($return_cells, ["class"=>ifttt(are_equal($even,1),"even","odd")]);

					// Highlight active row
					if (
						not_empty($this->config("keygetname")) &&
						not_empty($this->core->get_var($this->config("keygetname"))) &&
						are_equal(key_of($object->key(), $row), $this->core->get_var($this->config("keygetname")))
						)
						$return_row = $this->core->html->add_class($return_row,"active");

					$body[] = $return_row;

					// Additional rows

					if (not_empty($this->row)) {

						if ($row_keys = array_intersect_key($this->row,$row)) {

							foreach($row_keys as $row_key => $row_config){

								if (not_empty($row[$row_key])) {

									$cell_value = $this->replace_var($keys,$row,key_of("value", $row_config));

									if (preg_match($this->preg_function, $cell_value))
										$cell_value = $this->replace_function($cell_value);

									$class = [];
									$class[] = ifttt(are_equal($even,1),"even","odd");
									$class[] = key_of("class",$row_config);

									$body[] = $this->core->html->tr(
										$this->core->html->td($cell_value,["colspan"=>count($this->element)]),
										["class"=>$class]
									);
								}
							}
						}

					}
				}

			$footer = [];

			if (not_empty($keys))
			foreach ($this->element as $key => $element) {

				$cell_value = "";
				$value = key_of("footer", $element);

				if (is_string($value)) {

					$cell_value = $this->replace_var($keys,$row,$value);

					if (preg_match($this->preg_function, $cell_value))
						$cell_value = $this->replace_function($cell_value);

					if (key_of("array", $element) && $this->core->json->decode($cell_value)) {

						$cell_value = $this->core->json->decode($cell_value);
						if (is_indexed($cell_value) && count($cell_value) < 2){
							$cell_value = implode($cell_value);
						}
					}
				}

				if ($format = key_of("format", $element))
					$cell_value = $this->core->format->$format($cell_value);


				$cell = $this->core->html->tag(
					"th",
					$cell_value
				);

				$cell = $this->core->html->property($cell,	key_of("markup", $element));

				if ($format = key_of("format", $element))
					$cell = $this->core->html->add_class($cell,	$format);

				$footer[] = $cell;
			}

			$footer = $this->core->html->tr($footer);

			$table = $this->core->html->tag(
				"table",
				array(
					$this->core->html->thead($head),
					$this->core->html->tfoot($footer),
					$this->core->html->tbody(array($search, $body))
				),
				array("class" => either(fetch("table",$this->config("style_classes")), $this->config("class")))
			);

			if (key_fetch("search", $this->element)) {
				$table = $this->core->html->tag("form",$table, array("method"=>"post"));
			}

			$return[] = $table;
		}

		if ($this->pagination())
			$return[] = $pagination->render();

		$return = $this->core->html->render($return);

		return $return;

	}

	protected function replace_var ($keys = [], $values = [], $value = "") {

		if (is_empty($value))
			return $value;

		// Add key containing all values

		if (is_null(key_of("values", $values))) {
			$values["values"] = $values;

			if (is_null(key_of("_values", $values)))
				$values["_values"] = &$values["values"];

		}

		else if (is_null(key_of("_values", $values))) {
			$values["_values"] = $values;
		}

		// Escape not found vars in data array

		array_push($keys, "/(\\$)([a-z_]+)(?![-a-z_(])/i");

		$var = preg_replace_callback(

			$keys, function($matches) use ($values, $value){

				// Dont replace previous replacements
				if (!preg_match(preg_mod($matches[2]), $value))
					return $matches[0];

				// Return match

				if (isset($values[$matches[2]]))
				 	return preg_replace("-,-","&#44;",$this->core->json->encode($values[$matches[2]]));

				// Return Fallback
				return "";
			},
			$value
		);

		// OLD
		// Escape not found vars in data array

		// if (preg_match("/([$]\w+)([\->(]*)/",$var,$matches))
			// if ($matches[2] != "->" && $matches[2] != "(")
				// $var = preg_replace("/[$]\w+/","",$var);

		return $var;
	}

	protected function replace_function ($string) {

		$function = '/([^\( ,]+)\((.*)\)(.*)/s';

		if (preg_match("/confirm\(/",$string))
			return stripslashes($string);

    preg_match($function,$string, $matches);

		$func = $matches[1];
		$var = $matches[2];
		$append = $matches[3];

		if (preg_match($function,$var,$sub_matches)) {

			$sub_func = $sub_matches[1];
			$sub_var = $sub_matches[2];
			$sub_append = $sub_matches[3];

			$return = $this->replace_function($sub_func."(".$sub_var.")");

			if (is_array($return)) {
				$return = $this->core->json->encode($return);
				$return = preg_replace_array(array(/*"/\(/","/\)/",*/"/,/"),array(/*"&#40;","&#41;",*/"&#44;"),$return);
			}
			else {
				if (matches(",",$return))
					$return = preg_replace("/,/","&#44;",$return);
			}

			$return = $return.$sub_append;
			$var = preg_replace(preg_mod(preg_quote($sub_matches[0])),$return,$var);
    }

		$object = false;

		if (preg_match('/((\$)(.*))->(.*)/',$func,$matches)){
			$object = $matches[3];
			// $func = preg_replace('/(\$(.*))->/',"",$func);
			$func = $matches[4];
		}

		if (is_array($array = $this->core->json->decode($var)))
			$var = $array;

    if (is_string($var)) {

			if (matches(",",$var)) {
        $var = explode(",",$var);
      }
      else{
        $var = array($var);
      }

      foreach($var as &$varpart) {

				$varpart = preg_replace(array("/&#40;/","/&#41;/","/&#44;/"), array("(",")",","),$varpart);

				// Escape not found vars in data array
				if (matches("$", $varpart))
					$varpart = preg_replace("/[$]\w+/","",$varpart);

				//JSON
				if (is_array($json_decode = $this->core->json->decode($varpart)))
    			$varpart = $json_decode;

      }
    }

		if (is_object($this->config($object))) {

			$call = $this->config($object);

		}

		else {

			$objects = preg_split("/->/",$object);

			$call = $this;

			foreach ((array) $objects as $object)
				if (not_empty($object) && is_object($call->$object))
					$call = $call->$object;

		}

    if ($call && !empty($call) && is_object($call) && method_exists($call,$func))
      return is_scalar($array = call_user_func_array(array($call, $func), $var)) ? $array.$append : $array;

	  // if ($object && is_object($this->config($object)) && method_exists($this->config($object),$func))
      // return is_scalar($array = call_user_func_array(array($this->config($object), $func), $var)) ? $array.$append : $array;

		if (function_exists($func) || is_callable($func))
			return is_scalar($array = call_user_func_array($func, $var)) ? $array.$append : $array;

		if (not_empty($this->config($func)) || is_callable($this->config($func)))
			return is_scalar($array = call_user_func_array($this->config($func), $var)) ? $array.$append : $array;

		if (is_array($var))
			$var = $this->core->array->implode($var);

		return $var;

	}

	protected function row_template ($element) {

		$elements = array(
			"name"			=> $element,
			"value"			=> preg_match('-\$-',$element) ? $element : '$'.$element,
			"markup"		=> "",
			"cell_markup"	=> ""
		);
		return $elements;
	}

	protected function search_begin (){
		$this->search_table = $this->core->session->get($this->identifier.'_search');
	}

	protected function search_row ($head) {

		$search = false;

		foreach ($head as $key => $value) {

			if (key_of("search",$value) !== false) {
				$search = true;
				break;
			}

		}

		if ($search === false)
			return "";

		if (is_empty($head) || !is_array($head))
			return "";

		$form = $this->core->form();

		foreach ($head as $h => $value) {

			if (!$head[$h]["search"])
				continue;

			$name = "search_".$h;
			$form->new_element($name,array(
				"class"	=> "search",
			));

			if (is_array($search_config = key_of("search",$head[$h])))
				$head[$h]["search_config"] = $search_config;

			if (not_empty($search_config = key_of("search_config", $head[$h]))) {

				if (are_equal(key_of("type", $search_config), "select")) {
					$values = key_of("values", $search_config);
					$values = $this->core->array->union([""=>""],$values);
					$search_config["values"] = $values;
				}

				if (isset($search_config["var"])) {

					if (is_array($search_config["var"]))
						$form->element[$name]["name"] = "search_".key($head[$h]['search_config']["var"]);

					else
						$form->element[$name]["name"] = "search_".$head[$h]['search_config']["var"];
				}

				foreach ($search_config as $key => $value)
					$form->element[$name][$key] = $value;

			}

			if (
				is_array($this->search_table) &&
				array_key_exists($h,$this->search_table)
			) {
				$form->element[$name]["value"] = $this->search_table[$h];
			}

		}

		$form->new_element('search_extended', array(
			'type'		=> 'submit',
			'label'		=> $this->core->template->icon("zoom"),
			'markup'	=> 'class="small button"',
		));

		if ($form->submit()) {

			$had_entries = false;

			if (!empty($this->search_table))
				$had_entries = true;

			$this->search_table = [];

			foreach ($head as $k => $t) {

				$name = "search_".$k;

				if (isset($t["search_config"]["var"])) {

					if (is_array($t["search_config"]["var"]))
						$name = "search_".key($t["search_config"]["var"]);

					else
						$name = "search_".$t["search_config"]["var"];
				}

				if (not_empty($form->value($name)))
					$this->search_table[$k] = $form->value($name);

			}

			if ($this->search_table || $had_entries){

				if ($this->search_table)
					$this->core->session->set($this->identifier.'_search',$this->search_table);

				else if ($had_entries)
					$this->core->session->set($this->identifier.'_search',"");
			}

		}

		$return = [];

		foreach($head as $h => $value) {

			$name = "search_".$h;

			$array = array(
				$form->field($name)
			);

			if ($h == key(array_slice($head, -1, 1, TRUE)))
				$array[] = $form->field("search_extended");

			$array = array(
				"td" => $array
			);

			$return[] = $array;
		}

		$return = array(
			"tr" => array(
				$return, "class"=>"search"
			)
		);

		return $this->core->html->render($return);

	}

	protected function sort_begin () {

		if ($sort = $this->core->get("sort_".$this->identifier, "get")){

			$sort = html_entity_decode($sort);

			if ($sort == "false")
				$this->core->session->unset_key($this->identifier.'_sort');

			else
				$this->core->session->set($this->identifier.'_sort',$sort);

			$this->core->refresh(array("sort_".$this->identifier=>""));
		}

	}



	// Aliases

	public function element ($name = false, $array = false) {
		return $this->column($name, $array);
	}


	public function new_element($element, $array = false){
		return $this->new_column($element, $array);
	}


	public function new_elements($elements = false) {
		return $this->new_columns($elements);
	}


	public function set_element($element = false, $array = false){
		return $this->set_column($element, $array);
	}


	public function render_csv($body, $head = false, $foot = ""){
		$csv = $this->core->view->csv();
		$csv->identifier = $this->identifier;
		$csv->element = $this->element;
		// return $csv->render($body, $head, $foot);
		return $csv->render($body, $head);
	}


}

// CSV

class KB_csv_table extends KB_tables {

	private
		$core,
		$search = false,
		$preg_function = '/([^(^ ]+)\((.*)\)(.*)/s',
		$search_table = [];


	function __construct ($core){
		parent::__construct($core);
		$this->core = $core;

	}

	// Public

	public function render ($body = false, $group = false, $foot = "") {
		// unset($this->calc->temp);
		$body = preg_replace_array(array("/\(/","/\)/"),array("&#40;","&#41;"),$body);

		if (isset($this->element))
			$head = array_intersect_key($this->element,array_flip(array_diff(array_keys($this->element),array_keys(array_flip($this->config("exclude"))))));

		if (is_empty($foot) && isset($this->element))
			$foot = array_intersect_key($this->element,array_flip(array_diff(array_keys($this->element),array_keys(array_flip($this->config("exclude"))))));

		$return  = "";
		$return	.= $this->begin($head);

		if (isset($this->search_table) && !empty($this->search_table)) {

			foreach($this->search_table as $key => $value){

				if (
					not_empty($element_1 = key_of($key,$this->element)) ||
					$element_2 = fetch(["var" => $key], $this->element)
				) {

					$element = either($element_1,$element_2);

					// $filter = either(key_of("var", $this->element[$key]["search_config"]), $key);
					$filter = either(key_of("var", key_of(["search","search_config"], $element)), $key);
					// $value = ifttt(are_equal(key_of("type",$this->element[$key]["search_config"]), "select"), $value, "(.*)".$value."(.*)");
					$value = ifttt(are_equal(key_of("type",key_of(["search","search_config"], $element)), "select"), $value, "(.*)".$value."(.*)");

					$body = $this->core->array->filter_by(
						$body,
						$filter,
						$value
					);
				}
			}
		}

		$return	.= $this->head($head);
		$return .= $this->body($body, $head);
		$return .= $this->foot($foot, $head);

		// $return .= $this->foot($foot, $head);
		return $return;
	}


	// Private

	private function begin($head=false){
		if (!empty($head) && is_array($head)){
			foreach($head AS $h){
				if ($h["search"]){
					$this->search = true;
					$this->search_column[] = array('Field' => $h["name"],'Type'=>'varchar');
				}
			}
		}

		$return = "";

		if ($this->search){
			// $this->search_begin();
			$this->search_table = $this->core->session->get($this->identifier.'_search');
		}
		return $return;
	}


	private function head($head=false){
		$return = "";

		if (!empty($head) && is_array($head)){

			foreach($head AS $h){

				$return .= $h['title'].";";
			}
			$return .= "\n";
		}
		return $return;
	}


	private function body($body, $head = false){

		$return = "";

		if (!empty($head) && is_array($head)) {

			foreach($body as $b) {

				//Special Chars Encode
				$b = preg_replace_array(array("/\(/","/\)/","-,-"),array("&#40;","&#41;","&#44;"),$b);

				//Search End
				if (!empty($b) && is_array($b)) {
					$var = $b;

					$keys = array_keys($var);
					usort($keys, "rsort_by_length");

					$string	= "/NULL/";
					foreach($keys as $k => $v)
						$keys[$k] = '/(\$)('.$v.')/';

					foreach($head as $key => $value) {

						if ($head[$key]['value'] && !is_array($head[$key]['value'])) {
							$value = $head[$key]['value'];
							$value = $this->replace_var($keys,$var,$value);//preg_replace($keys,'$this->replace_var("$2",$var)',$value);



							if (preg_match($this->preg_function,$value))
								$value = $this->replace_function($value);

							$value = preg_replace($string,'',$value);
							$value = strip_tags($value);
							$value = preg_replace_array(array("/&#40;/","/&#41;/","/&#44;/"),array("(",")",","),$value);

							if (!preg_match("/^[0-9,]+$/",$value))
								$value = '"'.$value.'"';

							$return .= $value;
						}

						else if ($head[$key]['value'] && is_array($head[$key]['value'])) {

							$value = $head[$key]['value'][0];
							$array = $head[$key]['value'][1];
							$value = $this->replace_var($keys,$var,$value);//$value = preg_replace($keys,'$this->replace_var("$2",$var)',$value);

							if (preg_match($this->preg_function,$value))
								$value = $this->replace_function($value/*,$this->preg_function*/);//$value = preg_replace($function,'$this->replace_function("$0",$function)',$value);

							$value = (isset($array[$value])) ? $array[$value] : "";
							$value = preg_replace($string,'',$value);
							$value = strip_tags($value);
							// $value = '"'.$value.'"';
							$return .= $value;
						}
						//else if (isset($var[$head[$key]['name']]))	$return .= preg_replace($string,'',$var[$head[$key]['name']]);
						$return .= ";";
					}
					next($body);
					$return .= "\n";
				}

			}
		}

		else if (!empty($body) && is_array($body)) {
			//$return .= (array_key_exists('tbody', $this->style_classes)) ? '<tbody class="'.$this->style_classes['tbody'].'">'."\n" : '<tbody>'."\n";

			$tr = 0;

			foreach ($body as $b)	{

				if (not_empty($b) && is_array($b)) {

					foreach ($b as $k => $v) {
						// $v = preg_replace("/\"/","\"\"",$v);
						// $v = preg_replace("/\“/","\"\"",$v);
						// $v = preg_replace("/\„/","\"\"",$v);
						if (in_array($k, $this->config("exclude")))
							continue;

						if (is_array($v))
							$v = $this->core->json->encode($v);

						if (!is_scalar($v))
							continue;

						if (!preg_match("/^[0-9,]+$/",$v))
							$v = '"'.$v.'"';

						$return .= $v;
						$return .= ";";

					}
				}
				else {

					if (!preg_match("/^[0-9,]+$/",$b))
						$b = '"'.$b.'"';

					$return .= $b;
					$return .= ";";
				}
				$return .= "\r\n";
			}
			//$return .= "</tbody>"."\n";
		}
		return $return;
	}


	private function foot($foot, $head = []){

		if (any_empty($foot, is_array($foot)))
			return "";

		if (is_empty(array_values((array) key_fetch("footer", $foot))))
			return "";

		$return = "";

		$keys = array_keys($foot);

		foreach ($keys as $k => $v)
			$keys[$k] = '/(\$)('.$v.')/';

		foreach ($head as $key => $value) {

			$footer = either(key_of("footer", key_of($key, $foot)), "");

			if (not_empty($footer))
				$footer = $this->replace_var($keys,$foot,$footer);

			if (not_empty($footer) && preg_match($this->preg_function,$footer))
				$footer = $this->replace_function($footer);

			// $return .= (!in_array($foot, $this->config("exclude"))) ? $footer: "";
			if (is_scalar($footer))
				$return .= $footer;

			$return .= ";";
		}

		$return .= "\n";

		return $return;
	}


}

// CALCULATIONS

class KB_table_calc extends KB_calc {

	private $core, $temp;


	function __construct ($core) {
		parent::__construct($core);
		$this->core = $core;
		$this->temp = new StdClass;

	}


	public function calc() {

		$args = func_get_args();

		if (empty($args))
			return 0;

		$calc = 0;
		$modus = "+";

		foreach($args as $arg){

			if (is_empty($arg))
				$arg = 0;

			if (in_array($arg,array("+","-","*","/"))){
				$modus = $arg;
				continue;
			}

			$arg = $this->intify($arg);
			$calc = $this->calculate($calc, $arg, $modus);
			$modus = "+";

		}

		return $calc;

	}

	function calculate($var1, $var2, $modus="+") {
		switch($modus){
			case "+":
				return $var1+$var2;

			case "-":
				return $var1-$var2;

			case "*":
				return $var1*$var2;

			case "/":
				return $var1/$var2;
		}
	}

	function prepare_sum($value, $column, $return = false) {

		if (!isset($this->temp->sum))
			$this->temp->sum = new StdClass;

		if (!isset($this->temp->sum->$column))
			$this->temp->sum->$column = 0;

		if (!empty($value))
			$this->temp->sum->$column += $this->intify($value);

		if ($return)
			return $this->temp->sum->$column;

		else
			return $value;
	}

	function prepare_intermediate_sum($value, $column,  $return = false) {
		return $this->prepare_sum($value, $column, true);
	}

	function output_sum($column) {
		if (isset($this->temp->sum->$column))
		return $this->temp->sum->$column;
	}

	function output_intermediate_sum($column) {
		return $this->output_sum($column);
	}

	function prepare_average($value, $column, $return = false) {

		if (!isset($this->temp->average))
			$this->temp->average = new StdClass;

		if (!isset($this->temp->average->$column))
		$this->temp->average->$column = 0;

		if (!empty($value)){

			$new_value = $value;
			$new_value = $this->intify($new_value);

			if (!isset($this->temp->average->$column))
			$this->temp->average->$column = 0;

			if (!isset($this->temp->average_count->$column))
			$this->temp->average_count->$column = 0;

			$this->temp->average->$column = (($this->temp->average->$column*$this->temp->average_count->$column)+$new_value)/++$this->temp->average_count->$column;
		}
		if ($return)
			return $this->temp->average->$column;
		else
			return $value;
	}

	function prepare_intermediate_average($value, $column, $return = false) {
		return $this->prepare_average($value, $column, true);
	}

	function output_average($column) {
		if (isset($this->temp->average->$column))
		return $this->temp->average->$column;
	}

	function output_intermediate_average($column) {
		return $this->output_average($column);
	}

	function prepare_max($value, $column, $return = false) {

		if (!isset($this->temp->max))
			$this->temp->max = new StdClass;

		if (!isset($this->temp->max->$column))
		$this->temp->max->$column = 0;

		if (!empty($value)){
			$new_value = $value;
			// Convert to integer if necessary
			if ($new_value != "".floatval($new_value)) {
				$new_value = preg_replace("-[.]-","",$new_value);
				$new_value = preg_replace("-&#44;-",".",$new_value);
			}
			$new_value = floatval($new_value);

			if (!isset($this->temp->max->$column))
			$this->temp->max->$column = 0;

			if ($new_value > $this->temp->max->$column)
			$this->temp->max->$column = $new_value;
		}
		if ($return)
		return $this->temp->max->$column;
		else
		return $value;
	}

	function prepare_intermediate_max($value, $column, $return = false) {
		return $this->prepare_max($value, $column, true);
	}

	function output_max($column) {
		if (isset($this->temp->max->$column))
		return $this->temp->max->$column;
	}

	function output_intermediate_max($column) {
		return $this->output_max($column);
	}
}


// CARDS

class KB_cards extends KB_class {


	public $identifier = "";

	private $core;


	function __construct($core){

		$this->core = $core;

		$this->config("link","href");
		$this->config("title",false);
		$this->config("url_options",false);

	}


	public function add($name = "", $options = []) {

		if (is_empty($name))
			return false;

		$options = $this->core->array->merge(
			array(
				"type" => ""
			),
			$options
		);

		$content = (array) $this->private_config("content");

		$content[$name] = $options;

		$this->private_config("content", $content);

		return true;

	}


	public function image($path = "", $options = []) {

		$options = $this->core->array->merge(
			array(
				"type" => "image"
			),
			$options
		);

		return $this->add($path, $options);

	}


	public function link($name = "", $options = []) {

		if (is_empty($name))
			return false;

		$this->config("link", $name);

		return true;

	}


	public function render($array = [], $options = []) {

		$class = "cards";
		$options_to_remove = ["tag","type","thumbnail", "size","values"];

		if (not_empty($options))
			$class .= " ".$this->identifier;

		if (not_empty($this->config("class")))
			$class .= " ".$this->config("class");

		$attr = $this->core->array->merge(
			array("class"=>$class),
			$this->config("attr")
		);

		$return = $this->core->html->tag("div","",$attr);

		if (is_empty($array))
			return $this->core->html->render($return);

		if (!is_indexed($array))
			return $this->core->html->render($return);

		$return = [];
		$content = (array) $this->private_config("content");

		foreach ($array as $entry) {

			if (!is_array($entry))
				continue;

			$main = [];

			foreach ($content as $name => $element_options) {

				if (not_empty($values = key_of("values", $element_options)))
					$entry[$name] = key_of(key_of($name,$entry), $values);

				switch (key_of("type",$element_options)) {

					case "image" :

						$src = key_of($name,$entry);

						if (is_empty($src) && $fallback = key_of("fallback", $element_options))
							$src = key_of($fallback,$entry);

						if (is_empty($src))
							break;

						$thumbnail_src = either(key_of(key_of("thumbnail_src", $element_options),$entry), $src);

						if ($size = key_of("size", $element_options))
							$src = thumbnail($src, $size);

						if ($thumbnail = key_of("thumbnail", $element_options))
							$thumbnail = thumbnail($thumbnail_src, $thumbnail);

						if ($file = base64_decode($src, true)) {

							$f = finfo_open();
							$mime_type = finfo_buffer($f, $file, FILEINFO_MIME_TYPE);

							if (fetch($mime_type, array("text/plain", "text/html", "image/svg")))
								$mime_type = "image/svg+xml";

							$src = "data:".$mime_type.";base64, ".$src;
						}

						if ($thumbnail)
							$main[] = $this->core->html->tag("a",
								$this->core->html->img($thumbnail),
								$this->core->array->merge(
									remove_keys_of($options_to_remove,$element_options),
									array("href"=>$src)
								)
							);

						else
							$main[] = $this->core->html->img($src,
								remove_keys_of($options_to_remove,$element_options)
							);

						break;

					default:
						$main[] = $this->core->html->tag(
							either(key_of("tag",$element_options),"p"),
							either(key_of("content",$element_options),key_of($name,$entry)),
							remove_keys_of(array("tag","type","content"),$element_options)
						);

				}
			}

			if (is_empty(key_of($this->config("link"), $entry)) && not_empty($url_options = $this->config("url_options"))) {
				foreach ($url_options as $key => $value) {
					if (!is_null(key_of($value,$entry)))
					 	$url_options[$key] = key_of($value,$entry);
				}
				$entry[$this->config("link")] = $this->core->generate_url($url_options);
			}

			$class = "cell";

			if (not_empty($compare = $this->config("active"))) {

				foreach($compare as $key => $value) {
					if (not_empty(key_of($value, $entry)) && are_equal($this->core->get_var($key), key_of($value, $entry)))
						$class = "cell active";
				}
			}

			$return[] = $this->core->html->tag(
				ifttt(not_empty(key_of($this->config("link"), $entry)),"a","span"),
				$main,
				array(
					"class"=>$class,
					"href" => ifttt($href = key_of($this->config("link"),$entry),$href, false),
					"title" => ifttt($title = key_of($this->config("title"),$entry),$title, $this->config("title")),
				)
			);
		}
		if (not_empty($options)) {

			$styles = "";

			if ($columns = int(key_of("columns",$options)))
				$styles .= "grid-template-columns: repeat($columns, 1fr);";

			$styles = "div.cards.".$this->identifier."{".$styles."}";

			$style_tag = $this->core->html->tag("style",$styles);

			$this->core->template->backend->head($style_tag);
		}


		$return = $this->core->html->tag("div",$return,$attr);
		return $this->core->html->render($return);

	}



	// Alias

	public function href($name = "", $options = []){
		return $this->link($name, $options);
	}
}
