<?php

$core->template->backend->title($core->language->translate("Authentication"));

$form = $core->form();

if ($type = $core->get("request")) {

	if ($type == "login"){


		$form->new_element("username", [
			"label"		=>	"Email address",
			"charset"	=>	"email",
			"essential"	=> true,
			"message"	=> "A user with this email address does not exist",
		]);

		$form->new_element("submit", [
			"type"		=>	"submit",
			"value"		=> "Send email",
			"class"		=> "green"
		]);

		if ($form->submit()) {

			$entry = $core->user->entry(["name" => $form->value("username")]);

			if (are_equal(key_of("active",$entry), 1)) {

				$first_key = $core->session->id();
				$second_key = sha1($entry["password"].$first_key);

				$core->user->update([
					"ID" => $entry["ID"],
					"session_".date("U") => $second_key.$first_key
				]);

				$link = either(ifttt(fetch("https",$_SERVER),"https"),"http")."://".$_SERVER["HTTP_HOST"]."/backend/?request=authenticate&key=".$second_key;
				$mail = $core->mail();
				$mail->to($form->value("username"));
				$mail->from($form->value("username"), "Kitbooth");

				$mail->body(
					$core->html->h1(
						$core->language->translate("Sign in request")
					)
				);

				$mail->body(
					$core->html->p(
						$core->language->translate("Click the following link to authenticate your sign in request:")
					)
				);

				$mail->body(
					$core->html->a($link, array("href"=>$link))
				);

				$mail->subject(
					$core->language->translate(
						"Sign in request"
					)
				);

				$mail->send();
				$core->session->set("user",$core->get("username"));
				$core->refresh(["request" => "awaiting"]);

			}

			else{
				$form->element["username"]["error"] = true;
			}
		}


		?>

		<div class="panel">
			<form method="post">
			<h2 class="section"><?=$core->language->translate("Request sign in")?></h2>
			<div class="section">
				<?=$form->render(false)?>
			</div>
			<div class="section flex fill">
				<?=$form->field("submit")?><a class="cancel button" href="<?=$core->generate_url(array("request"=>""))?>"><?=$core->language->translate("Back")?></a>
			</div>
			</form>
		</div>
		<?php
	}
	else if(($type == "authenticate") && ($key = $core->get("key"))) {

		$user = fetch($key."([0-9a-z]{32})",$core->user->read());

		if (not_empty($user)) {

			$user = current($user);
			$session = fetch($key."([0-9a-z]{32})",$user);
			$update = array(
				"ID" => $user["ID"],
				"session" => preg_replace(preg_mod($key."([0-9a-z]{32})"),"$1",current($session)),
				key($session) => "",
				"lastlogin" => date("Y-m-d H:i:s"),
			);

			$core->message->set(
				$core->user->update($update),
				"User login"
			);
		}

		echo $core->template->kit->panel(
			"Authentication complete",
			$core->html->tag("p",
			$core->language->translate("You can close this window now")
			)
		);
	}

	else {

		$core->template->backend->head(
			$core->html->meta(array("http-equiv"=>"refresh", "content"=>"10; URL=".$core->generate_url()))
		);

		echo $core->template->kit->panel(
			"Authentication requested",
			$core->html->p(
				$core->language->translate("An email has been send. It contains a link to authenticate your request")
			)
		);

	}

}

else 	if (is_empty($core->user->read())) {

	$form->new_elements(array(
		"username" => array(
			"label" => "Email Address",
			"charset" => "email",
			"type" => "email",
			"essential" => true,
		),
		"password" => array(
			"label" => "Password",
			"type" => "password",
			"essential" => true,
		),
		"usergroup" => array(
			"label" => "User Group",
			"placeholder" => "Administrator",
		),
		"submit" => array(
			"type" => "submit",
			"value" => "Accept",
		),
	));

	if ($form->check()) {

		$ID = key_of("ID",$core->user->group->entry(array("parent_ID"=>NULL)));

		if (is_empty($ID))
			$ID = $core->user->group->create(array(
				"parent_ID" => "NULL",
				"name" => either($form->value("usergroup"),"Administrator"),
				"backend" => "1",
			));

		$core->user->create(array(
			"group_ID" => $ID,
			"name" => $form->value("username"),
			"password" => $form->value("password"),
		));

		$core->user->check($form->value("username"),$form->value("password"));
		$core->refresh();

	}

	$core->template->kit->panel->title("Welcome");

	$core->template->kit->panel->body(
		$core->html->tag("h3", $core->language->translate("Create the first user account"))
	);
	$core->template->kit->panel->body(
		$form->render(false)
	);
	$core->template->kit->panel->body(
		$core->html->tag("h3", $core->language->translate("Licence Agreement"))
	);

	$licences = [];

	$fs = $core->filesystem();

	foreach ($fs->folder->read($fs->folder->path($core->path(),"core/licence")) as $file) {
		$key = preg_replace("/^(\w{2})(.*)/","$1",$file);
		$licences[$key] = $file;
	}

	$licence = either(key_of($core->language->accept(), $licences),key_of("en", $licences));
	$licence = $fs->file->read($fs->folder->path($core->path(),"core/licence").$licence);
	$licence = markdown_decode($licence, array("shift"=>3));

	$core->template->kit->panel->body(
		$core->html->tag("div",array($licence), array("style"=>"max-height: 20em; max-width: 60em; overflow: auto; font-size:.85em; line-height: 1.1em; padding:0 1em; border:solid 1px hsl(0, 0%, 79%);"))
	);

	$core->template->kit->panel->footer(
		$form->field("submit")
	);

	$core->template->kit->panel->options(array(
		"wrap_form" => true,
	));

	echo $core->template->kit->panel();

}

else {

	if ($core->config("username_is_email") == "0")
		$form->new_element("username", array(
			"label" => "Username",
			"essential" => true,
		));

	else
		$form->new_element("username", array(
			"label" => "Email Address",
			"charset" => "email",
			"type" => "email",
			"markup" => 'style="min-width: 15em"',
			"essential" => true,
		));

	$form->new_elements(array(
		"password" => array(
			"label" => "Password",
			"type" => "password",
			"essential" => true,
		),
		"submit" => array(
			"type" => "submit",
			"value" => "Sign in",
			"class"	=> "green",
		),
	));

	if($form->check()){

			$core->message->set(
				$core->user->check(
					$form->value("username"),
					$form->value("password")
				),
				"Check user credentials"
			);

			$core->message->info("","Invalid email or password");

			if ($core->message->status())
				$core->refresh();

	}

	echo $core->template->kit->panel(
		"Sign in",
		$form->render(false),
		array(
			$form->field("submit"),
			$core->template->kit->button("Sign in via email", array("href"=>$core->generate_url(array("request"=>"login"))))
		),
		array(
			"wrap_form"=>$form,
			"footer" => array("class" => "fill flex")
		)
	);

 }

