<?php

// Declaration of Variables

$additional_menu = true;
$entry = true;
$menu = true;
$form = true;
$table = true;

// Load Icons

$core->template->icon->set_folder($core->kit->current["url"]."backend/icons");

// Load  Config

if ($config = $core->file->read($core->kit->current["url"]."setup.json"))
	$config = $core->json->decode($config);

$config = ifttt(is_array($config), $config, []);

$config_keys = [
	"options",
	"rights",
	"overview_properties",
	"overview_group_properties",
	"edit_properties",
	"edit_group_properties",
	"edit_layout",
	"edit_group_layout",
	"group_overview_properties",
	"group_overview_group_properties",
	"group_edit_properties",
	"group_edit_group_properties",
	"group_edit_layout",
	"group_edit_group_layout",
];

foreach ($config_keys as $config_key)
	$config[$config_key] = either(
			$core->config($config_key),
			key_of($config_key, $config)
		);




$get_config = function ($key = "", $group = false) use ($core, $data, $config) {

	if (any_empty($key, is_string($key)))
		return [];

	if ($core->array->match($key, ["overview_properties","edit_properties","edit_layout"])) {

		if (are_equal($core->get("view"), "groups"))
			$key = "group_".$key;

		if (is_empty(key_of("group_properties", key_of("options", $config))))
			return key_of($key, $config);

		$ID = false;

		if (are_equal($core->get("view"), "groups")) {
			$ID = $core->get("ID");
		}

		else {

			if (!are_equal($key, "overview_properties") && ($ID = $core->get("ID")) &&  $data->group) {
				$entry = $data->entry($ID);
				$ID = key_of($data->group(), $entry);
			}

		}

		if ($ID || $ID = key_of($data->group(), $core->session->kit->get("filter"))) {

			$key_group = preg_replace("/_([a-z]+)$/", "_group_$1", $key);
			$group_config = key_of($ID, key_of($key_group, $config));

			if (not_empty($group_config))
				return $group_config;

			foreach ((array)key_fetch("ID", $data->group->parents($ID)) as $parent) {
				$group_config = key_of($parent, key_of($key_group, $config));

				if (not_empty($group_config))
					return $group_config;
			}

			return key_of($key, $config);

		}

		return key_of($key, $config);

	}

	return key_of($key, $config);

};

$set_config = function ($key = "", $set = NULL) use ($core, $data, $config) {

	if (is_empty($key) || is_null($set))
		return;

	if ($core->array->match($key, ["overview_properties","edit_properties","edit_layout"])) {

		if (are_equal($core->get("view"), "groups"))
			$key = "group_".$key;

		if (is_empty(key_of("group_properties", key_of("options", $config))))
			return $core->config($key, $set);

		$ID = false;

		if (are_equal($core->get("view"), "groups")) {
			$ID = $core->get("ID");
		}

		else {

			if (($ID = $core->get("ID")) && $data->group) {
				$entry = $data->entry($ID);
				$ID = key_of($data->group(), $entry);
			}

		}


		if ($ID || $ID = key_of($data->group(), $core->session->kit->get("filter"))) {

			$key_group = preg_replace("/_([a-z]+)$/", "_group_$1", $key);
			$config = key_of($key_group, $config);
			$config[$ID] = $set;
			return $core->config($key_group, $config);

		}

	}

	return $core->config($key, $set);

};

$user_can = function ($check_right) use($core, $get_config) {

	$rights = $get_config("rights");

	if (
		is_empty(key_of($check_right,$rights)) ||
		fetch($core->user->data("group_ID"), key_of($check_right, $rights))
		)
		return true;

	return false;

};

$options = function ($key = "") use ($get_config) {

	return key_of($key, $get_config("options"));

};

$display = function ($key = "") use ($get_config) {

	return in_list($key, key_of("display", $get_config("options")));

};


$correlation_o = $core->correlation;
$tables_o = $core->tables;

if (are_equal($core->get("view"), "groups")) {
	$view_groups = true;
	$group_folder_name = "_group";
	$object = $data->group;
}

else {
	$view_groups = false;
	$group_folder_name = "";
	$object = $data;
}

// Filesystem objects

$frontend_files = $core->filesystem(
	$core->path("files").key_of("unique_name", key_of("config", $kit)).$group_folder_name
);

$kit_files = $core->filesystem(
	$kit["path"]
);

$backend_files = $core->filesystem(
	$kit["path"]."backend"
);

// Extend Kit

$extend = $kit_files->file->read_php("extend.php", [
	"additional_menu" => &$additional_menu,
	"config" => &$config,
	"data" => &$data,
	"entry" => &$entry,
	"form" => &$form,
	"menu" => &$menu,
	"table" => &$table,
], [
	"backend_before_includes",
	"edit_before_check",
	"overview_before_output",
	"duplicate_settings",
	"toolbar_buttons",
]);


// Set static vars

$input_types = [
	[
		"type" => "text",
		"label" => "Text",
	],
	[
		"type" => "color",
		"label" => "Color",
	],
	[
		"type" => "number",
		"label" => "Number",
	],
	[
		"type" => "textarea",
		"label" => "Text Field",
	],
	[
		"type" => "select",
		"label" => "Choice Menu",
	],
	[
		"type" => "radio",
		"label" => "Choice",
	],
	[
		"type" => "checkbox",
		"label" => "Multiple Choice",
	],
	[
		"type" => "date",
		"label" => "Date",
	],
	[
		"type" => "datetime-local",
		"label" => "Date and Time",
	],
	[
		"type" => "time",
		"label" => "Time",
	],
	[
		"type" => "file",
		"label"	=> "File",
	],
	[
		"type" => "image",
		"field" => "image",
		"class" => "image",
		"label"	=> "Image",
	],
	[
		"type" => "video",
		"field" => "video",
		"class" => "video",
		"label"	=> "Video",
	],
	[
		"type" => "foreign",
		"label"	=> "Data source",
	],
	[
		"type" => "correlation",
		"label"	=> "Data correlation",
	],
	[
		"type" => "password",
		"label"	=> "Password",
	],
	[
		"type" => "code",
		"field" => "textarea",
		"class" => "code",
		"label"	=> "Code",
	],
	[
		"type" => "html",
		"field" => "html",
		"label"	=> "HTML",
	],
	[
		"type" => "markdown",
		"field" => "markdown",
		"label"	=> "Markdown",
	],
	[
		"type" => "hidden",
		"field" => "hidden",
		"label"	=> "hidden",
	],
];


if ($object->group)
	$input_types[] = [
		"type" => $object->group(),
		"label"	=> "Group",
	];

if ($object->position)
	$input_types[] = [
		"type" =>"position",
		"label"	=> "Position",
	];

$input_options["all"] = [
	"array"	=> "Multiple Values",
	"essential"	=> "Required",
];

$input_options["image"] = [
	"6mp"	=> "Reduce to 6MP",
];

$filter_options = [
	"children" => "Limit to",
	"descendant" => "All children of",
];

$calc_types = [
	"" => "",
	"sum" => "Sum",
	"intermediate_sum" => "Intermediate sum",
	"average" => "Average",
	"intermediate_average" => "Intermediate average",
	"max" => "Max value",
	"intermediate_max" => "Intermediate max value",
];

$format_types = [
	"" => "",
	"euro" => "Euro",
	"currency" => "Currency",
	"date" => "Datum",
	"datetime" => "Datum und Uhrzeit",
	"time" => "Uhrzeit",
	"file" => "Datei",
	"email" => "E-Mail",
	"website" => "Website",
	"phone" => "Telefonnummer",
	"toolbar" => "Toolbar",
];

$option_types = array(
	"filter"	=> "Filter",
);

if (is_object($data->group))
	$menu = $core->template->kit->menu(
		[
			"" => [
				"title" => either(key_of("data_title",$get_config("options")),$core->kit->title),
				"url_options" => ["group_ID" => ""],
			],
			"groups" => [
				"title" => either(key_of("group_title",$get_config("options")), "Groups"),
				"url_options" => ["set_group_ID" => ""],
			]
		]
	);

else
	$menu = false;


$backend_before_includes = key_of("backend_before_includes",$extend);
$backend_before_includes();


$content = [];
$shared_objects = [
	"additional_menu" => &$additional_menu,
	"calc_types" => $calc_types,
	"config" => $config,
	"data" => &$data,
	"display" => $display,
	"duplicate_settings" => key_of("duplicate_settings",$extend),
	"entry" => &$entry,
	"edit_before_check" => key_of("edit_before_check",$extend),
	"filter_options" => $filter_options,
	"format_types" => $format_types,
	"form" => &$form,
	"option_types" => $option_types,
	"get_config" => $get_config,
	"set_config" => $set_config,
	"input_options" => $input_options,
	"input_types" => $input_types,
	"menu" => &$menu,
	"object" => &$object,
	"options" => $options,
	"overview_before_output" => key_of("overview_before_output",$extend),
	"table" => &$table,
	"user_can" => $user_can,
	"view_groups" => $view_groups,
];

switch ($core->get("do")) {

	case "config":
	case "config_export":

		if ($user_can("config"))
			$content[] = $backend_files->file->read_php("config.php", $shared_objects);

		break;

	case "export":

		if ($user_can("export"))
			$content[] = [
				$backend_files->file->read_php("overview.php", $shared_objects),
				$backend_files->file->read_php("export.php", $shared_objects)
			];

		break;

	case "import":

		if ($user_can("edit"))
			$content[] = $backend_files->file->read_php("import.php", $shared_objects);

		break;

	case "view":

		$content[] = $backend_files->file->read_php("view.php", $shared_objects);
		break;

	case "edit":

		if ($user_can("edit"))
			$content[] = $backend_files->file->read_php("edit.php", $shared_objects);

		break;

	case "overview_properties":

		if ($user_can("overview_properties"))
			$content[] = $backend_files->file->read_php("overview_properties.php", $shared_objects);

		break;

	case "edit_properties":

		if ($user_can("edit"))
			$content[] = $backend_files->file->read_php("edit.php", $shared_objects);

		if ($user_can("edit_properties"))
			$content[] = $backend_files->file->read_php("edit_properties.php", $shared_objects);

		break;

	case "edit_layout":

		if ($user_can("edit"))
			$content[] = $backend_files->file->read_php("edit.php", $shared_objects);

		if ($user_can("edit_properties"))
			$content[] = $backend_files->file->read_php("edit_properties.php", $shared_objects);

		if ($user_can("edit_layout"))
			$content[] = $backend_files->file->read_php("edit_layout.php", $shared_objects);

		break;

	case "delete":

		if (is_empty($ID = $core->get_var("ID")))
			break;

		if (!$user_can("edit"))
			break;

		$frontend_files->folder->delete($ID);

		$core->message->set(
			$object->delete($ID),
			"Delete entry with ID $ID"
		);

		$core->message->info("Entry has been deleted");
		$core->refresh(["do"=>"", "ID"=>""]);

		break;

	case "archive":

		if (is_empty($ID = $core->get_var("ID")))
			break;

		if (!$user_can("edit"))
			break;

		$core->message->set(
			$object->update(["archived"=>"1"],$ID),
			"Archive entry with ID $ID"
		);

		$core->message->info("Entry has been archived");
		$core->refresh(["do"=>"", "ID"=>""]);

		break;

	case "unarchive":

		if (is_empty($ID = $core->get_var("ID")))
			break;

		if (!$user_can("edit"))
			break;

		$core->message->set(
			$object->update(["archived"=>"0"],$ID),
			"Restore entry with ID $ID from archive"
		);

		$core->message->info("Entry has been restored from archive");
		$core->refresh(["do"=>"", "ID"=>""]);

		break;

	case "toggle_archive":

		if ($core->session->kit->get("view_archive".ifttt($view_groups, "_group")))
			$core->session->kit->set("view_archive".ifttt($view_groups, "_group"),"");

		else
			$core->session->kit->set("view_archive".ifttt($view_groups, "_group"),"1");

		$core->refresh(["do"=>""]);

	case "update":

		if (!$user_can("edit"))
			$core->refresh(["do"=>"","ID"=>"","update"=>""]);

		if (any_empty($ID = $core->get("ID"), $update = $core->get("update")))
			$core->refresh(["do"=>"","ID"=>"","update"=>""]);

		$update = $core->json->decode($update);
		$update["ID"] = $ID;
		$object->update($update);

		$core->refresh(["do"=>"","ID"=>"","update"=>""]);

		break;


	case "duplicate":

		if ($user_can("edit"))
			$content[] = $backend_files->file->read_php("duplicate.php", $shared_objects);
		break;

	case "api":

		$content[] = $backend_files->file->read_php("api.php", $shared_objects);
		break;

}

switch ($core->get("view")) {

	case "groups":
	case "":
		$content[] = $backend_files->file->read_php("overview.php", $shared_objects);

}

// Load Blueprints Toolkit
$core->kit->load_toolkit("psia_blueprints", array("type"=>"data"));

$core->template->backend->head(
	$core->template->backend->js($kit["path"]."backend/forms.js")
);
$core->template->backend->head(
	$core->template->backend->css($kit["path"]."backend/forms.css")
);

echo $core->html->render($content);

